#include "broadcast.h"
#include <mpi.h>
#include <stdint.h>

#define _XOPEN_SOURCE 500
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

#include <assert.h>
#include <limits.h>

#include <math.h>
#include <string.h>

#include "timing.h"
#include "topo.h"

#include "utils.h"

#define BROADCAST_DEBUG 0

#if BROADCAST_DEBUG
#define BC_DEBUG(fmt, ...) LOG("[%d] BroadCast " fmt, rank, __VA_ARGS__);
#else
#define BC_DEBUG(fmt, ...)
#endif

#define ENUM_STR(ident) #ident,
const char *BCastAlgorithmStrings[BCastAlgorithmCount] = {
    BCAST_ALGORITHM_LIST(ENUM_STR)};
#undef ENUM_STR

/*
 * Topologically aware MPI broadcast operations
 */

extern int rank, size;

#define SEND_TAG 0

static BroadcastCtx *context = NULL;

void get_comm_name(MPI_Comm comm, char name[static MPI_MAX_OBJECT_NAME],
                   int *rlen) {
  if (comm == MPI_COMM_NULL) {
    strncpy(name, "(MPI_COMM_NULL)", MPI_MAX_OBJECT_NAME - 1);
  } else {
    MPI_Comm_get_name(comm, name, rlen);
  }
}

void get_comm_ranks(MPI_Comm comm, char out[1024]) {
  if (comm == MPI_COMM_NULL) {
    strncpy(out, "(MPI_COMM_NULL: no ranks)", 1024);
    return;
  }

  MPI_Group grp, world_grp;

  MPI_Comm_group(MPI_COMM_WORLD, &world_grp);
  MPI_Comm_group(comm, &grp);

  int grp_size;

  MPI_Group_size(grp, &grp_size);

  int *ranks = malloc(grp_size * sizeof(int));
  assert(ranks != NULL);
  int *world_ranks = malloc(grp_size * sizeof(int));
  assert(world_ranks != NULL);

  for (int i = 0; i < grp_size; i++)
    ranks[i] = i;

  MPI_Group_translate_ranks(grp, grp_size, ranks, world_grp, world_ranks);

  char name[MPI_MAX_OBJECT_NAME] = {0};
  int rlen = 0;
  get_comm_name(comm, name, &rlen);
  // MPI_Comm_get_name(comm, name, &rlen);

  memset(out, 0, 1024);
  for (int i = 0; i < grp_size; i++) {
    if (i > 0)
      sprintf(out + strlen(out), ",");
    sprintf(out + strlen(out), "%d", world_ranks[i]);
  }

  free(ranks);
  free(world_ranks);

  MPI_Group_free(&grp);
  MPI_Group_free(&world_grp);
}

static bool rank_in_group(NodeGroup *group, int rank) {
  for (int i = 0; i < group->n_ranks; i++)
    if (group->ranks[i] == rank)
      return true;
  return false;
}

static Cluster *find_rank_in_cluster(Cluster *cluster, int find_rank) {
  if (!cluster)
    return NULL;

  if (cluster->leaf) {
    if (rank_in_group(cluster->group, find_rank)) {
      return cluster;
    }
  } else {
    for (int i = 0; i < cluster->n_sub_clusters; i++) {
      Cluster *ret = find_rank_in_cluster(&cluster->sub_clusters[i], find_rank);
      if (ret != NULL)
        return ret;
    }
  }
  return NULL;
}

static Cluster *find_leader_node(Cluster *cluster) {
  assert(cluster->leaf);

  Cluster *work = cluster;
  // while (work->leader_rank == cluster->leader_rank && work->parent)
  //   work = work->parent;
  while (work->parent && work->parent->leader_rank == cluster->leader_rank)
    work = work->parent;

  // fprintf(stderr, "%d found leader node to #%d = #%d, subclusters=%d\n",
  // rank,
  //         cluster->id, work->id, work->n_sub_clusters);
  return work;
}

static void comm_from_node_group(const NodeGroup *group, MPI_Comm base,
                                 MPI_Comm *out) {
  MPI_Group base_group;
  MPI_Comm_group(base, &base_group);

  MPI_Group new_group;
  MPI_Group_incl(base_group, group->n_ranks, group->ranks, &new_group);

  MPI_Comm_create_group(base, new_group, 0, out);

  MPI_Group_free(&base_group);
  MPI_Group_free(&new_group);
}

void create_leader_comm(const char *debug, Cluster *cluster, MPI_Comm comm,
                        int my_rank, MPI_Comm *comm_out) {

  int n_ranks = 1;
  if (!cluster->leaf)
    n_ranks = cluster->n_sub_clusters;

  int leader_ranks[n_ranks];

  if (cluster->leaf) {
    leader_ranks[0] = my_rank;
  } else {
    for (int i = 0; i < cluster->n_sub_clusters; i++)
      leader_ranks[i] = cluster->sub_clusters[i].leader_rank;
  }

  NodeGroup parent_leader_group = {.n_ranks = n_ranks, .ranks = leader_ranks};
  comm_from_node_group(&parent_leader_group, comm, comm_out);
  char ranks[1024] = {0};
  get_comm_ranks(*comm_out, ranks);
  fprintf(stderr, "%d #%d Created leader comm %s: n=%d %s\n", cluster->id, rank,
          debug, n_ranks, ranks);
}

int topo_bcast_init(Topology *topology, MPI_Comm comm, BroadcastCtx *ctx) {

  Cluster *my_leaf = find_rank_in_cluster(&topology->cluster, rank);
  assert(my_leaf != NULL);
  assert(my_leaf->leaf);

  MPI_Comm my_group_comm;
  comm_from_node_group(my_leaf->group, comm, &my_group_comm);

  int my_rank, my_size;
  MPI_Comm_rank(my_group_comm, &my_rank);
  MPI_Comm_size(my_group_comm, &my_size);

  Cluster *my_node = find_leader_node(my_leaf);
  Cluster *my_parent = my_node->parent;

  context = malloc(sizeof(BroadcastCtx));
  assert(context != NULL);

  *context = (BroadcastCtx){
      .cluster = &topology->cluster,
      .my_leaf = my_leaf,
      .my_node = my_node,
      .my_parent = my_parent,
      .my_group_comm = my_group_comm,
      .my_parent_leader_comm = NULL,
      .my_children_leader_comm = NULL,
      .my_rank = my_rank,
      .my_size = my_size,
  };
  *ctx = *context;

  MPI_Barrier(comm);
  BC_DEBUG("topo_init(): my_rank=%d, my_size=%d, sub_clusters=%d\n", my_rank,
           my_size, my_node->n_sub_clusters);

  return 0;
}

int topo_bcast_finalize() {
  free(context);
  context = NULL;
  return 0;
}

// topo_broadcast does topological broadcast

int topo_broadcast_default(void *buffer, int count, MPI_Datatype datatype,
                           int root, MPI_Comm comm) {
  return MPI_Bcast(buffer, count, datatype, root, comm);
}

// http://cs.uh.edu/~gabriel/courses/cosc4397_s06/ParCo_21_ImplementationOptions.pdf
static inline int bcast_group_linear(void *buffer, int count,
                                     MPI_Datatype datatype, int tag,
                                     MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  const int sender_rank = 0;
  if (local_rank == sender_rank) {
    // we sent to all
    // MPI_Request *reqs = malloc(sizeof(MPI_Request) * (local_size - 1));
    MPI_Request reqs[local_size - 1];

    for (int dest = 1; dest < local_size; dest++) {
      MPI_Isend(buffer, count, datatype, dest, tag, comm, &reqs[dest - 1]);
    }
    int ret = MPI_Waitall(local_size - 1, reqs, MPI_STATUS_IGNORE);
    // free(reqs);
    return ret;
  } else {
    return MPI_Recv(buffer, count, datatype, sender_rank, tag, comm,
                    MPI_STATUS_IGNORE);
  }
}

static inline int bcast_group_chain(void *buffer, int count,
                                    MPI_Datatype datatype, int tag,
                                    MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  const int dest = local_rank + 1;
  const int src = local_rank - 1;

  // fprintf(stderr, "local_rank: %d, src:%d, dest:%d\n", local_rank, src,
  // dest);

  if (local_rank == 0) {
    // first in chain
    if (dest < local_size) {
      return MPI_Send(buffer, count, datatype, dest, tag, comm);
    } else {
      return MPI_SUCCESS;
    }
  } else if (local_rank + 1 >= local_size) {
    // last in chain
    return MPI_Recv(buffer, count, datatype, src, tag, comm, MPI_STATUS_IGNORE);
  } else {
    int ret =
        MPI_Recv(buffer, count, datatype, src, tag, comm, MPI_STATUS_IGNORE);
    if (ret != MPI_SUCCESS)
      return ret;
    return MPI_Send(buffer, count, datatype, dest, tag, comm);
  }
}

static inline int send_up_to_two(void *buffer, int count, MPI_Datatype datatype,
                                 MPI_Comm comm, int rank1, int rank2, int size,
                                 int tag) {
  MPI_Request reqs[2];
  int scount = 0;
  if (rank1 < size) {
    MPI_Isend(buffer, count, datatype, rank1, tag, comm, &reqs[0]);
    scount++;
  }
  if (rank2 < size) {
    MPI_Isend(buffer, count, datatype, rank2, tag, comm, &reqs[1]);
    scount++;
  }
  return MPI_Waitall(scount, reqs, MPI_STATUS_IGNORE);
}

static inline int bcast_group_btree(void *buffer, int count,
                                    MPI_Datatype datatype, int tag,
                                    MPI_Comm comm) {

  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  const int parent_rank = (local_rank - 1) / 2;
  const int child_rank_1 = 2 * local_rank + 1;
  const int child_rank_2 = 2 * (local_rank + 1);

  assert(local_rank == 0 || parent_rank >= 0);

  if (local_rank != 0) {
    assert(MPI_Recv(buffer, count, datatype, parent_rank, tag, comm,
                    MPI_STATUS_IGNORE) == MPI_SUCCESS);
  }

  return send_up_to_two(buffer, count, datatype, comm, child_rank_1,
                        child_rank_2, local_size, SEND_TAG);
}

static inline void create_split_comms(MPI_Comm comm, int local_rank,
                                      int is_even_rank, MPI_Comm *part1,
                                      MPI_Comm *part2) {
  // Caching mechanism because MPI_Comm_split takes a lot of time
  static int attr_key = INT_MIN;
  static uintptr_t id = -1;
  static MPI_Comm cache_part1, cache_part2;

  if (attr_key == INT_MIN) {
    MPI_Comm_create_keyval(MPI_NULL_COPY_FN, MPI_NULL_DELETE_FN, &attr_key,
                           (void *)0);
  }

  bool cache_valid = true;

  if (id == -1) {
    cache_valid = false;
  } else {
    int flag = 0;
    uintptr_t value = 0;
    MPI_Comm_get_attr(comm, attr_key, &value, &flag);

    if (!flag || value != id) {
      // no value, or mismatch
      cache_valid = false;
    }
  }

  if (!cache_valid) {
    MPI_Comm_split(comm, (!is_even_rank || local_rank == 0) ? 1 : MPI_UNDEFINED,
                   local_rank, &cache_part1);
    MPI_Comm_split(comm, (is_even_rank || local_rank == 0) ? 1 : MPI_UNDEFINED,
                   local_rank, &cache_part2);

    // set random id
    id = rand();
    MPI_Comm_set_attr(comm, attr_key, (void *)id);
  }

  *part1 = cache_part1;
  *part2 = cache_part2;
}

// https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=4536478
static inline int bcast_group_btree_with_swap(void *buffer, int count,
                                              MPI_Datatype datatype, int tag,
                                              MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  if (local_size <= 2) {
    return MPI_Bcast(buffer, count, datatype, tag, comm);
  }

  const bool is_even_rank = local_rank % 2 == 0;

  int datatype_size;
  MPI_Type_size(datatype, &datatype_size);

  MPI_Comm part1, part2;
  create_split_comms(comm, local_rank, is_even_rank, &part1, &part2);

  char *buffp = buffer;

  enum { PART1, PART2 };
  if (local_rank == 0) {
    bcast_group_btree(buffp, count / 2, datatype, 0, part1);
    bcast_group_btree(buffp + (count / 2) * datatype_size, count - count / 2,
                      datatype, 0, part2);
  } else if (!is_even_rank) {
    bcast_group_btree(buffp, count / 2, datatype, 0, part1);
  } else {
    bcast_group_btree(buffp + (count / 2) * datatype_size, count - count / 2,
                      datatype, 0, part2);
  }

  int dest;
  if (is_even_rank) {
    dest = local_rank + 1;
  } else {
    dest = local_rank - 1;
    if (dest < 0)
      dest = local_size - dest;
  }
  dest = dest % local_size;
  // fprintf(stderr, "%d is recieving from %d and sending to %d\n", local_rank,
  // dest, dest);

  if (!is_even_rank) {
    MPI_Sendrecv(buffp, count / 2, datatype, dest, PART1,
                 buffp + (count / 2) * datatype_size, count - count / 2,
                 datatype, dest, PART2, comm, MPI_STATUS_IGNORE);
  } else {
    MPI_Sendrecv(buffp + (count / 2) * datatype_size, count - count / 2,
                 datatype, dest, PART2, buffp, count / 2, datatype, dest, PART1,
                 comm, MPI_STATUS_IGNORE);
  }

  if (local_rank == 0 && local_size % 2 != 0) {
    // 0-rank sends twice
    dest = local_size - 1;
    // fprintf(stderr, "%d is recieving from %d and sending to (again)%d\n",
    // local_rank, dest, dest);
    if (dest % 2 == 0) {
      return MPI_Sendrecv(buffp, count / 2, datatype, dest, PART1,
                          buffp + (count / 2) * datatype_size,
                          count - count / 2, datatype, dest, PART2, comm,
                          MPI_STATUS_IGNORE);
    } else {
      return MPI_Sendrecv(buffp + (count / 2) * datatype_size,
                          count - count / 2, datatype, dest, PART2, buffp,
                          count / 2, datatype, dest, PART1, comm,
                          MPI_STATUS_IGNORE);
    }
  }
  // double end = MPI_Wtime();
  // if (local_rank == 0)
  // fprintf(stderr, "TIME TAKEN: WTIME=%lf\n", end-start);

  // then exchange
  return MPI_SUCCESS;
}

static double factor = 1.3;
void set_pipeline_chunk_factor(double f) { factor = f; }

static inline int bcast_group_pipeline(void *buffer, int count,
                                       MPI_Datatype datatype, int tag,
                                       MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  int dtsize;
  MPI_Type_size(datatype, &dtsize);

  int nchunks = local_size * factor;
  int chunk_size = count / nchunks;
  int rest_count = count - (chunk_size * nchunks);

  if (chunk_size < 1 || nchunks <= 0) {
    // default to builtin when not large enough for chunk size
    return MPI_Bcast(buffer, count, datatype, tag, comm);
  }

  assert(chunk_size != 0);
  assert(nchunks > 0);
  assert(local_size > 1);

#define CHUNK_PART(chunk) (buffer + (dtsize * chunk_size * (chunk)))

  if (local_rank == 0)
    BC_DEBUG(
        "Starting pipeline with nchunks=%d, chunk_size=%d and rest_count=%d\n",
        nchunks, chunk_size, rest_count);

  int prev_rank = local_rank - 1;
  int next_rank = local_rank + 1;

  if (local_rank + 1 == local_size)
    next_rank = -1;

  BC_DEBUG("has prev_rank %d next_rank %d\n", prev_rank, next_rank);

  if (local_rank == 0) {
    for (int chunk = 0; chunk < nchunks; chunk++) {
      // BC_DEBUG("sending %d of %d\n", chunk, nchunks);
      assert(MPI_Send(CHUNK_PART(chunk), chunk_size, datatype, next_rank, 0,
                      comm) == MPI_SUCCESS);
    }

    // send rest
    assert(MPI_Send(CHUNK_PART(nchunks), rest_count, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

  } else if (local_rank > 0 && local_rank < local_size - 1) {

    // first recieve
    assert(MPI_Recv(CHUNK_PART(0), chunk_size, datatype, prev_rank, 0, comm,
                    MPI_STATUSES_IGNORE) == MPI_SUCCESS);

    // then send and recieve
    for (int chunk = 1; chunk < nchunks; chunk++) {
      BC_DEBUG(
          "MID is sending chunk %d to %d, and recieving chunk %d from %d\n",
          chunk - 1, next_rank, chunk, prev_rank);

      assert(MPI_Sendrecv(CHUNK_PART(chunk - 1), chunk_size, datatype,
                          next_rank, 0, CHUNK_PART(chunk), chunk_size, datatype,
                          prev_rank, 0, comm,
                          MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    }

    // then send the last recieved one
    BC_DEBUG("MID is sending last chunk chunk %d to %d\n", nchunks - 1,
             next_rank);
    assert(MPI_Send(CHUNK_PART(nchunks - 1), chunk_size, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

    // recv rest
    assert(MPI_Recv(CHUNK_PART(nchunks), rest_count, datatype, prev_rank, 0,
                    comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    // send rest
    assert(MPI_Send(CHUNK_PART(nchunks), rest_count, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

  } else if (local_rank == local_size - 1) {
    // only recieve from previous
    for (int chunk = 0; chunk < nchunks; chunk++) {
      BC_DEBUG("LAST is recieving chunk %d from %d\n", chunk, prev_rank);
      assert(MPI_Recv(CHUNK_PART(chunk), chunk_size, datatype, prev_rank, 0,
                      comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    }

    assert(MPI_Recv(CHUNK_PART(nchunks), rest_count, datatype, prev_rank, 0,
                    comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
  }

  BC_DEBUG("finished all %d chunks\n", nchunks);

  return MPI_SUCCESS;
}

static inline int bcast_group_pipeline_nonblocking(void *buffer, int count,
                                                   MPI_Datatype datatype,
                                                   int tag, MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  int dtsize;
  MPI_Type_size(datatype, &dtsize);

  int nchunks = local_size * factor;
  int chunk_size = count / nchunks;
  int rest_count = count - (chunk_size * nchunks);

  if (chunk_size < 1) {
    // default to builtin when not large enough for chunk size
    return MPI_Bcast(buffer, count, datatype, tag, comm);
  }

  assert(chunk_size != 0);
  assert(nchunks > 0);
  assert(local_size > 1);

#define CHUNK_PART(chunk) (buffer + (dtsize * chunk_size * (chunk)))

  if (local_rank == 0)
    BC_DEBUG(
        "Starting pipeline with nchunks=%d, chunk_size=%d and rest_count=%d\n",
        nchunks, chunk_size, rest_count);

  int prev_rank = local_rank - 1;
  int next_rank = local_rank + 1;

  if (local_rank + 1 == local_size)
    next_rank = -1;

  BC_DEBUG("has prev_rank %d next_rank %d\n", prev_rank, next_rank);

  if (local_rank == 0) {
    MPI_Request reqs[nchunks + 1];

    for (int chunk = 0; chunk < nchunks; chunk++) {
      // BC_DEBUG("sending %d of %d\n", chunk, nchunks);
      // assert(MPI_Send(CHUNK_PART(chunk), chunk_size, datatype, next_rank, 0,
      //                 comm) == MPI_SUCCESS);
      ASSERT_MPI_OK(MPI_Isend(CHUNK_PART(chunk), chunk_size, datatype,
                              next_rank, 0, comm, &reqs[chunk]));
    }

    // send rest
    // assert(MPI_Send(CHUNK_PART(nchunks), rest_count, datatype, next_rank, 0,
    //                 comm) == MPI_SUCCESS);
    ASSERT_MPI_OK(MPI_Isend(CHUNK_PART(nchunks), rest_count, datatype,
                            next_rank, 0, comm, &reqs[nchunks]));

    ASSERT_MPI_OK(MPI_Waitall(nchunks + 1, reqs, MPI_STATUSES_IGNORE));

  } else if (local_rank > 0 && local_rank < local_size - 1) {

    // first recieve
    assert(MPI_Recv(CHUNK_PART(0), chunk_size, datatype, prev_rank, 0, comm,
                    MPI_STATUSES_IGNORE) == MPI_SUCCESS);

    // then send and recieve
    for (int chunk = 1; chunk < nchunks; chunk++) {
      BC_DEBUG(
          "MID is sending chunk %d to %d, and recieving chunk %d from %d\n",
          chunk - 1, next_rank, chunk, prev_rank);

      assert(MPI_Sendrecv(CHUNK_PART(chunk - 1), chunk_size, datatype,
                          next_rank, 0, CHUNK_PART(chunk), chunk_size, datatype,
                          prev_rank, 0, comm,
                          MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    }

    // then send the last recieved one
    BC_DEBUG("MID is sending last chunk chunk %d to %d\n", nchunks - 1,
             next_rank);
    assert(MPI_Send(CHUNK_PART(nchunks - 1), chunk_size, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

    // recv rest
    assert(MPI_Recv(CHUNK_PART(nchunks), rest_count, datatype, prev_rank, 0,
                    comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    // send rest
    assert(MPI_Send(CHUNK_PART(nchunks), rest_count, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

  } else if (local_rank == local_size - 1) {
    // only recieve from previous
    MPI_Request reqs[nchunks + 1];
    for (int chunk = 0; chunk < nchunks; chunk++) {
      BC_DEBUG("LAST is recieving chunk %d from %d\n", chunk, prev_rank);
      // assert(MPI_Recv(CHUNK_PART(chunk), chunk_size, datatype, prev_rank, 0,
      //                 comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
      ASSERT_MPI_OK(MPI_Irecv(CHUNK_PART(chunk), chunk_size, datatype,
                              prev_rank, 0, comm, &reqs[chunk]));
    }

    // assert(MPI_Recv(CHUNK_PART(nchunks), rest_count, datatype, prev_rank, 0,
    //                 comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    ASSERT_MPI_OK(MPI_Irecv(CHUNK_PART(nchunks), rest_count, datatype,
                            prev_rank, 0, comm, &reqs[nchunks]));

    ASSERT_MPI_OK(MPI_Waitall(nchunks + 1, reqs, MPI_STATUSES_IGNORE));
  }

  BC_DEBUG("finished all %d chunks\n", nchunks);

  return MPI_SUCCESS;
}

int topo_broadcast_pipelined(void *buffer, int count, MPI_Datatype datatype,
                             int root, MPI_Comm _comm) {

#define CHUNK_PART(chunk) (buffer + (dtsize * chunk_size * (chunk)))

  bool is_leader = context->my_rank == 0;
  bool is_root = rank == root;

  int dtsize;
  MPI_Type_size(datatype, &dtsize);

  int nchunks = size * factor;
  int chunk_size = count / nchunks;
  int rest_count = count - (chunk_size * nchunks);

  // REQUIREMENT: root must be a rank 0 (root leader)
  assert(root == 0);

  if (chunk_size < 1) {
    return topo_broadcast_default(buffer, count, datatype, root, _comm);
  }

  for (int chunk = 0; chunk < nchunks; chunk++) {
    // fprintf(stderr, "sending chunk %d/%d\n", chunk, nchunks);
    ASSERT_MPI_OK(topo_broadcast_default(CHUNK_PART(chunk), chunk_size,
                                         datatype, root, _comm));
  }

  return MPI_SUCCESS;
}

int topo_broadcast_magpie_pipelined(void *buffer, int count,
                                    MPI_Datatype datatype, int root,
                                    MPI_Comm _comm) {

#define CHUNK_PART(chunk) (buffer + (dtsize * chunk_size * (chunk)))

  bool is_leader = context->my_rank == 0;
  bool is_root = rank == root;

  int dtsize;
  MPI_Type_size(datatype, &dtsize);

  int nchunks = size * factor;
  int chunk_size = count / nchunks;
  int rest_count = count - (chunk_size * nchunks);

  // REQUIREMENT: root must be a rank 0 (root leader)
  assert(root == 0);

  if (chunk_size < 1) {
    return topo_broadcast_magpie(buffer, count, datatype, root, _comm);
  }

  for (int chunk = 0; chunk < nchunks; chunk++) {
    // fprintf(stderr, "sending chunk %d/%d\n", chunk, nchunks);
    ASSERT_MPI_OK(topo_broadcast_magpie(CHUNK_PART(chunk), chunk_size, datatype,
                                        root, _comm));
  }

  return MPI_SUCCESS;
}

// TODO: not implemented
static inline int bcast_group_multisite(void *buffer, int count,
                                        MPI_Datatype datatype, int tag,
                                        MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  int dtsize;
  MPI_Type_size(datatype, &dtsize);

  int nchunks = local_size * factor;
  int chunk_size = count / nchunks;
  int rest_count = count - (chunk_size * nchunks);

  if (chunk_size < 1) {
    // default to builtin when not large enough for chunk size
    return MPI_Bcast(buffer, count, datatype, tag, comm);
  }

  assert(chunk_size != 0);
  assert(nchunks > 0);
  assert(local_size > 1);

#define CHUNK_PART(chunk) (buffer + (dtsize * chunk_size * (chunk)))

  if (local_rank == 0)
    BC_DEBUG(
        "Starting pipeline with nchunks=%d, chunk_size=%d and rest_count=%d\n",
        nchunks, chunk_size, rest_count);

  int prev_rank = local_rank - 1;
  int next_rank = local_rank + 1;

  if (local_rank + 1 == local_size)
    next_rank = -1;

  BC_DEBUG("has prev_rank %d next_rank %d\n", prev_rank, next_rank);

  if (local_rank == 0) {
    MPI_Request reqs[nchunks + 1];

    for (int chunk = 0; chunk < nchunks; chunk++) {
      // BC_DEBUG("sending %d of %d\n", chunk, nchunks);
      // assert(MPI_Send(CHUNK_PART(chunk), chunk_size, datatype, next_rank, 0,
      //                 comm) == MPI_SUCCESS);
      ASSERT_MPI_OK(MPI_Isend(CHUNK_PART(chunk), chunk_size, datatype,
                              next_rank, 0, comm, &reqs[chunk]));
    }

    // send rest
    // assert(MPI_Send(CHUNK_PART(nchunks), rest_count, datatype, next_rank, 0,
    //                 comm) == MPI_SUCCESS);
    ASSERT_MPI_OK(MPI_Isend(CHUNK_PART(nchunks), rest_count, datatype,
                            next_rank, 0, comm, &reqs[nchunks]));

    ASSERT_MPI_OK(MPI_Waitall(nchunks + 1, reqs, MPI_STATUSES_IGNORE));

  } else if (local_rank > 0 && local_rank < local_size - 1) {

    // first recieve
    assert(MPI_Recv(CHUNK_PART(0), chunk_size, datatype, prev_rank, 0, comm,
                    MPI_STATUSES_IGNORE) == MPI_SUCCESS);

    // then send and recieve
    for (int chunk = 1; chunk < nchunks; chunk++) {
      BC_DEBUG(
          "MID is sending chunk %d to %d, and recieving chunk %d from %d\n",
          chunk - 1, next_rank, chunk, prev_rank);

      assert(MPI_Sendrecv(CHUNK_PART(chunk - 1), chunk_size, datatype,
                          next_rank, 0, CHUNK_PART(chunk), chunk_size, datatype,
                          prev_rank, 0, comm,
                          MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    }

    // then send the last recieved one
    BC_DEBUG("MID is sending last chunk chunk %d to %d\n", nchunks - 1,
             next_rank);
    assert(MPI_Send(CHUNK_PART(nchunks - 1), chunk_size, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

    // recv rest
    assert(MPI_Recv(CHUNK_PART(nchunks), rest_count, datatype, prev_rank, 0,
                    comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    // send rest
    assert(MPI_Send(CHUNK_PART(nchunks), rest_count, datatype, next_rank, 0,
                    comm) == MPI_SUCCESS);

  } else if (local_rank == local_size - 1) {
    // only recieve from previous
    MPI_Request reqs[nchunks + 1];
    for (int chunk = 0; chunk < nchunks; chunk++) {
      BC_DEBUG("LAST is recieving chunk %d from %d\n", chunk, prev_rank);
      // assert(MPI_Recv(CHUNK_PART(chunk), chunk_size, datatype, prev_rank, 0,
      //                 comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
      ASSERT_MPI_OK(MPI_Irecv(CHUNK_PART(chunk), chunk_size, datatype,
                              prev_rank, 0, comm, &reqs[chunk]));
    }

    // assert(MPI_Recv(CHUNK_PART(nchunks), rest_count, datatype, prev_rank, 0,
    //                 comm, MPI_STATUSES_IGNORE) == MPI_SUCCESS);
    ASSERT_MPI_OK(MPI_Irecv(CHUNK_PART(nchunks), rest_count, datatype,
                            prev_rank, 0, comm, &reqs[nchunks]));

    ASSERT_MPI_OK(MPI_Waitall(nchunks, reqs, MPI_STATUSES_IGNORE));
  }

  BC_DEBUG("finished all %d chunks\n", nchunks);

  return MPI_SUCCESS;
}

typedef int BCast_group(void *buffer, int count, MPI_Datatype datatype, int tag,
                        MPI_Comm comm);

BCast_group *bcast_group_func = bcast_group_btree;
enum BCastAlgorithm bcast_algo = BCastDefault;

enum BCastAlgorithm get_bcast_algorithm(void) { return bcast_algo; }

void set_bcast_algorithm(enum BCastAlgorithm choice) {
  bcast_algo = choice;
  assert(choice >= BCastDefault && choice < BCastAlgorithmCount);
  bcast_group_func = BCastFuncs[choice];
}

static inline int bcast_group(void *buffer, int count, MPI_Datatype datatype,
                              int tag) {
  const MPI_Comm comm = context->my_group_comm;

  int local_size;
  MPI_Comm_size(comm, &local_size);

  if (local_size <= 1)
    return MPI_SUCCESS;

  int ret = bcast_group_func(buffer, count, datatype, tag, comm);
  // fprintf(stderr, "rank %d finished bcast_group\n", rank);
  return ret;
}

// Broadcast function optimized for intra communication
// members of comm should be on different nodes
int topo_bcast_inter(void *buffer, int count, MPI_Datatype datatype, int tag,
                     MPI_Comm comm) {
  return bcast_group_func(buffer, count, datatype, tag, comm);
}

// Broadcast function for communication within nodes
int topo_bcast_intra(void *buffer, int count, MPI_Datatype datatype, int tag) {
  return bcast_group(buffer, count, datatype, tag);
}

int topo_broadcast_magpie(void *buffer, int count, MPI_Datatype datatype,
                          int root, MPI_Comm _comm) {
  bool is_leader = context->my_rank == 0;
  bool is_root = rank == root;

  // REQUIREMENT: root must be a rank 0 (root leader)
  assert(root == 0);

  const int tag = SEND_TAG;

  if (context->cluster->leaf) {
    return bcast_group(buffer, count, datatype, tag);
  }

  if (is_leader && !is_root) {
    Cluster *parent = context->my_parent;
    assert(parent != NULL);
    const int leader_rank = parent->leader_rank;
    assert(MPI_Recv(buffer, count, datatype, leader_rank, SEND_TAG, _comm,
                    MPI_STATUS_IGNORE) == MPI_SUCCESS);
  }

  // linear
  if (is_leader || is_root) {
    Cluster *my_node = context->my_node;

    MPI_Request reqs[my_node->n_sub_clusters];
    int reqs_count = 0;

    for (int i = 0; i < my_node->n_sub_clusters; i++) {
      int leader_rank = my_node->sub_clusters[i].leader_rank;
      if (rank == leader_rank)
        continue;

      assert(MPI_Isend(buffer, count, datatype, leader_rank, SEND_TAG, _comm,
                       &reqs[reqs_count]) == MPI_SUCCESS);
      reqs_count++;
    }
    int ret = bcast_group(buffer, count, datatype, tag);
    MPI_Waitall(reqs_count, reqs, MPI_STATUSES_IGNORE);
    return ret;
  }

  return bcast_group(buffer, count, datatype, tag);
}

int topo_broadcast_magpie2(void *buffer, int count, MPI_Datatype datatype,
                           int root, MPI_Comm _comm) {
  assert(false && "not implemented");
  return -1;
}

#if 0
int topo_broadcast_magpie2(void *buffer, int count, MPI_Datatype datatype,
                           int root, MPI_Comm _comm) {
  bool is_leader = context->my_rank == 0;
  bool is_root = rank == root;

  // REQUIREMENT: root must be a rank 0 (root leader)
  assert(root == 0);

  const int tag = SEND_TAG;

  if (context->cluster->leaf) {
    return topo_bcast_intra(buffer, count, datatype, tag);
  }

  if (is_leader && !is_root) {
    assert(context->my_parent_leader_comm != MPI_COMM_NULL);
    BC_DEBUG("is leader and bcasting recieve from leaders\n", 0);
    assert(topo_bcast_inter(buffer, count, datatype, SEND_TAG,
                            context->my_parent_leader_comm) == MPI_SUCCESS);
    return topo_bcast_intra(buffer, count, datatype, tag);
  }

  if (is_leader || is_root) {
    if (context->my_children_leader_comm == MPI_COMM_NULL) {
      return MPI_SUCCESS;
    }
    BC_DEBUG("is leader or root and bcasting send to leaders\n", 0);
    assert(context->my_children_leader_comm != MPI_COMM_NULL);
    assert(topo_bcast_inter(buffer, count, datatype, SEND_TAG,
                            context->my_children_leader_comm) == MPI_SUCCESS);
    return topo_bcast_intra(buffer, count, datatype, tag);
  }

  return topo_bcast_intra(buffer, count, datatype, tag);
}
#endif

// These two swap functions are from
// https://www.geeksforgeeks.org/block-swap-algorithm-for-array-rotation/
/*This function swaps d elements starting at index fi
with d elements starting at index si */
static void swap_arr(char *arr, size_t fi, size_t si, size_t d) {
  for (size_t i = 0; i < d; i++) {
    char temp = arr[fi + i];
    arr[fi + i] = arr[si + i];
    arr[si + i] = temp;
  }
}

// This code is contributed by rathbhupendra

static void left_rotate(char *arr, size_t d, size_t n) {
  size_t i, j;
  if (d == 0 || d == n)
    return;

  /* If number of elements to be rotated is more than
   * array size*/
  if (d > n)
    d = d % n;

  i = d;
  j = n - d;
  while (i != j) {
    if (i < j) /*A is shorter*/
    {
      swap_arr(arr, d - i, d + j - i, i);
      j -= i;
    } else /*B is shorter*/
    {
      swap_arr(arr, d - i, d, j);
      i -= j;
    }
  }
  /*Finally, block swap A and B*/
  swap_arr(arr, d - i, d, i);
}

int bcast_group_scatter_gather_overlap(void *buffer, int count,
                                       MPI_Datatype datatype, int root,
                                       MPI_Comm _comm) {
  int my_rank, my_size;
  MPI_Comm_rank(_comm, &my_rank);
  MPI_Comm_size(_comm, &my_size);

  const int send_count = count / my_size;
  const int rest_count = count - send_count * my_size;
  BC_DEBUG("send_count: %d, rest_count: %d, total_count: %d\n", send_count,
           rest_count, count);

  const int tag = 808123;
  const int rest_tag = 909923;

  if (send_count > 0) {
    // Step 1: Scatter
    MPI_Request reqs[2 * (my_size - 1)]; // only used for root
    MPI_Request rest_req;                // not used by root

    void *rest_part = buffer + (send_count * my_size);

    if (my_rank == 0) {
      // scatter part:
      for (int to_rank = 1; to_rank < my_size; to_rank++) {
        const void *send_part = buffer + send_count * to_rank;
        ASSERT_MPI_OK(MPI_Isend(send_part, send_count, datatype, to_rank,
                                tag + to_rank, _comm, &reqs[to_rank - 1]));

        // send rest part as well, but with other tag
        if (rest_count > 0) {
          ASSERT_MPI_OK(MPI_Isend(rest_part, rest_count, datatype, to_rank,
                                  rest_tag + to_rank, _comm,
                                  &reqs[(my_size - 1) + (to_rank - 1)]));
        }
      }
    } else {
      // scatter part: recieve our first chunk from root
      MPI_Request req;
      ASSERT_MPI_OK(MPI_Recv(buffer, send_count, datatype, 0, tag + my_rank,
                             _comm, MPI_STATUS_IGNORE));

      // also recieve rest part, but non-blocking because it isn't part of
      // the brucks algorithm step
      if (rest_count > 0) {
        ASSERT_MPI_OK(MPI_Irecv(rest_part, rest_count, datatype, 0,
                                rest_tag + my_rank, _comm, &rest_req));
      }
    }

    // Step 2: Allgather using brucks algorithm
    int count = send_count;
    int pow2 = 1;

    while (pow2 < my_size) {
      int to_rank = (my_rank - pow2 + my_size) % my_size;
      int from_rank = (my_rank + pow2) % my_size;

      BC_DEBUG("step=%d, to_rank=%d,from_rank=%d\n", pow2, to_rank, from_rank);

      const int sendrecv_tag = pow2;

      void *recv_part = buffer + count;
      void *send_part = buffer;

      ASSERT_MPI_OK(MPI_Sendrecv(
          send_part, count, datatype, to_rank, sendrecv_tag, recv_part, count,
          datatype, from_rank, sendrecv_tag, _comm, MPI_STATUSES_IGNORE));

      pow2 *= 2;
      count *= 2;
    }

    // we have to reorder data - shift by my_rank steps
    left_rotate(buffer, count - (send_count * my_rank), count);

    // Root should wait for all send requests to be complete
    if (my_rank == 0) {
      ASSERT_MPI_OK(MPI_Waitall(my_size - 1, reqs, MPI_STATUSES_IGNORE));
    } else if (rest_count > 0) {
      ASSERT_MPI_OK(MPI_Wait(&rest_req, MPI_STATUS_IGNORE));
    }
  } else {
    // Just use bcast because otherwise it would be a linear send
    return MPI_Bcast(buffer, rest_count, datatype, root, _comm);
  }
  return MPI_SUCCESS;
}

int bcast_group_scatter_gather(void *buffer, int count, MPI_Datatype datatype,
                               int root, MPI_Comm _comm) {
  int my_rank, my_size;
  MPI_Comm_rank(_comm, &my_rank);
  MPI_Comm_size(_comm, &my_size);

  const int send_count = count / my_size;
  void *my_part = buffer + send_count * my_rank;
  if (my_rank == 0)
    my_part = MPI_IN_PLACE;

  const int rest_count = count - send_count * my_size;
  BC_DEBUG("send_count: %d, rest_count: %d, total_count: %d\n", send_count,
           rest_count, count);

  if (send_count > 0) {
    assert(MPI_Scatter(buffer, send_count, datatype, my_part, send_count,
                       datatype, root, _comm) == MPI_SUCCESS);

    assert(MPI_Allgather(MPI_IN_PLACE, send_count, datatype, buffer, send_count,
                         datatype, _comm) == MPI_SUCCESS);
  }

  if (rest_count == 0)
    return MPI_SUCCESS;

  void *rest_part = buffer + (send_count * my_size);
  return MPI_Bcast(rest_part, rest_count, datatype, root, _comm);
}

int topo_broadcast_normal(void *buffer, int count, MPI_Datatype datatype,
                          int root, MPI_Comm comm) {
  return bcast_group_func(buffer, count, datatype, SEND_TAG, comm);
}

Topo_broadcast *BCastFuncs[BCastAlgorithmCount] = {
    MPI_Bcast, bcast_group_btree,
    // bcast_group_linear,
    // bcast_group_chain,
    bcast_group_btree_with_swap, bcast_group_pipeline,
    bcast_group_pipeline_nonblocking, bcast_group_scatter_gather,
    // bcast_group_scatter_gather_overlap,
};
