#define _XOPEN_SOURCE 500

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <mpi.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "broadcast_bench.h"
#include "optimized_bcast.h"

/* Globals */
// static int rank, size;
// static int N = 100;
// static int WARMUP = 5;
// static bool verbose = false;
// static char *msgsize_str = NULL;
// static char *types_str = NULL;
// static Topology topology;
// static BroadcastCtx broadcastCtx;
// static char *bench_name = NULL;
// static bool VERIFY_BCAST_EVERY_ITER = false;
// static bool VERIFY_BCAST = false;

// static bool options_set = false;

static void statistics(double *nums, int count, double *sum, double *mean,
                       double *min, double *max, double *sum_sq, double *var) {
  double sum_0;
  double mean_0;
  double min_0;
  double max_0;
  double sum_sq0;
  double var0;

  if (!sum)
    sum = &sum_0;
  if (!mean)
    mean = &mean_0;
  if (!min)
    min = &min_0;
  if (!max)
    max = &max_0;
  if (!sum_sq)
    sum_sq = &sum_sq0;
  if (!var)
    var = &var0;

  *sum = 0.0;
  *sum_sq = 0.0;
  *min = 9e99;
  *max = 0.0;

  for (int i = 0; i < count; i++) {
    const double num = nums[i];
    *sum += num;
    *sum_sq = *sum_sq + (num * num);
    if (*min > num)
      *min = num;
    if (*max < num)
      *max = num;
  }
  *mean = *sum / count;

  if (count > 1) {
    *var = (*sum_sq - (*sum * *sum) / count) / (count - 1);
    if (*var < 0.0)
      *var = 0.0;
  } else {
    *var = 0.0;
  }
}

static void reset_buffer(const BroadcastBenchOpts *opts, char *buffer,
                         const int msgsize) {
  if (opts->rank == 0) {
    for (int i = 0; i < msgsize; i++) {
      buffer[i] = (i % 26) + 'A';
    }
  } else {
    for (int i = 0; i < msgsize; i++) {
      buffer[i] = '.';
    }
  }
}

static int verify_buffer(const BroadcastBenchOpts *opts, char *buffer,
                         const int msgsize, const char *name, int i) {
  // verify buffer
  int valid = 1;
  for (int i = 0; i < msgsize; i++) {
    if (buffer[i] != (i % 26) + 'A')
      valid = 0;
  }

  // printf("TEST %d got: %.*s\n", rank, msgsize, buffer);

  int gatherbuf[opts->size];
  memset(gatherbuf, 0, opts->size * sizeof(int));

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Gather(&valid, 1, MPI_INT, gatherbuf, 1, MPI_INT, 0, MPI_COMM_WORLD);

  enum BCastAlgorithm bcast_algo = get_bcast_algorithm();
  int failed = 0;

  if (opts->rank == 0) {
    bool allvalid = true;
    for (int i = 0; i < opts->size; i++)
      if (!gatherbuf[i])
        allvalid = false;
    if (allvalid) {
      //        fprintf(stderr, ">> %s passed test\n", name);
    } else {
      fprintf(stderr, "> Benchmarking broadcast %s: %d/%d (algo: %s)\n", name,
              i, opts->iterations, BCastAlgorithmStrings[bcast_algo]);
      fprintf(stderr, ">> %s FAILED test: ", name);
      for (int i = 0; i < opts->size; i++)
        if (!gatherbuf[i])
          fprintf(stderr, "%d,", i);
      fprintf(stderr, "\n");

      failed = true;
    }
  }

  MPI_Bcast(&failed, 1, MPI_INT, 0, MPI_COMM_WORLD);
  return failed;
}

static int test_broadcast_impl_bytes(const BroadcastBenchOpts *opts,
                                     const int msgsize, char *buffer,
                                     const char *name,
                                     Topo_broadcast broadcast_func,
                                     enum BCastAlgorithm bcast_algo) {
  double *statsdata = calloc(opts->iterations, sizeof(double));
  assert(statsdata != NULL);

  set_bcast_algorithm(bcast_algo);

  int failed = false;

  for (int i = 0; i < opts->warmup; i++) {
    reset_buffer(opts, buffer, msgsize);
    MPI_Barrier(MPI_COMM_WORLD);
    int ret = broadcast_func(buffer, msgsize, MPI_BYTE, 0, MPI_COMM_WORLD);
  }

  for (int i = 0; i < opts->iterations; i++) {
    if (opts->rank == 0 && opts->verbose)
      fprintf(stderr, "> Benchmarking broadcast %s: %d/%d\n", name, i,
              opts->iterations);

    reset_buffer(opts, buffer, msgsize);

    MPI_Barrier(MPI_COMM_WORLD);

    const double start = MPI_Wtime();
    int ret = broadcast_func(buffer, msgsize, MPI_BYTE, 0, MPI_COMM_WORLD);
    const double end = MPI_Wtime();
    assert(ret == MPI_SUCCESS);

    if (opts->verify_bcast_every_iter &&
        verify_buffer(opts, buffer, msgsize, name, i) != 0) {
      // fprintf(stderr, "verify_buffer failed: %s\n", buffer)
      fprintf(stderr, "%d verify_buffer failed, we got: %.*s\n", opts->rank,
              msgsize, buffer);
      failed = true;
      break;
    }

    statsdata[i] = 1000 * (end - start); // milliseconds
  }

  if (opts->verify_bcast &&
      verify_buffer(opts, buffer, msgsize, name, opts->iterations - 1) != 0) {
    // fprintf(stderr, "verify_buffer failed: %s\n", buffer)
    fprintf(stderr, "%d verify_buffer failed, we got: %.*s\n", opts->rank,
            msgsize, buffer);
    failed = true;
  }

  double mean = 0.0, min = 0.0, max = 0.0, var = 0.0;
  if (!failed) {
    // Collect stats from all ranks and use them to generate statistics
    double *all_statsdata =
        calloc(opts->iterations * opts->size, sizeof(double));
    assert(statsdata != NULL);

    MPI_Gather(statsdata, opts->iterations, MPI_DOUBLE, all_statsdata,
               opts->iterations, MPI_DOUBLE, 0, MPI_COMM_WORLD);

    statistics(all_statsdata, opts->iterations * opts->size, NULL, &mean, &min,
               &max, NULL, &var);
    free(all_statsdata);
  }
  free(statsdata);

  if (opts->rank == 0) {
    if (failed) {
      fprintf(stderr, "Broadcast FAILED N=%d, %s (%s) msgsize=%d\n",
              opts->iterations, name,
              BCastAlgorithmStrings[get_bcast_algorithm()], msgsize);
    } else {
      const char *algo = BCastAlgorithmStrings[bcast_algo];

      if (strcmp(name, "optimized") == 0)
        algo = "BCastDefault";

      fprintf(
          stdout,
          "Broadcast BYTES "
          "N=%d\t%s\t(%-15s)\tmsgsize=%-7d\tmean=%-7.4lf\tvar=%-7.4lf\tmin=%-"
          "7.4lf\tmax=%-7.4lf\tP=%d "
          "(ms)\tname=%s\n",
          opts->iterations, name, algo, msgsize, mean, var, min, max,
          opts->size, opts->bench_name);
    }
  }
  return failed;
}

typedef struct {
  int a;
  float b;
  char message[10];
  int sum_message;
  double doubles[3];
  double sum_doubles;
} TestStruct;

static int test_broadcast_impl_struct(const BroadcastBenchOpts *opts,
                                      const int msgsize, const char *name,
                                      Topo_broadcast broadcast_func,
                                      enum BCastAlgorithm bcast_algo) {

  int blocklengths[6] = {1, 1, 10, 1, 3, 1};
  MPI_Datatype types[6] = {MPI_INT, MPI_FLOAT,  MPI_BYTE,
                           MPI_INT, MPI_DOUBLE, MPI_DOUBLE};
  MPI_Aint offsets[6] = {
      offsetof(TestStruct, a),       offsetof(TestStruct, b),
      offsetof(TestStruct, message), offsetof(TestStruct, sum_message),
      offsetof(TestStruct, doubles), offsetof(TestStruct, sum_doubles)};
  MPI_Datatype test_type;
  MPI_Type_create_struct(6, blocklengths, offsets, types, &test_type);
  assert(MPI_Type_commit(&test_type) == MPI_SUCCESS);

  double *data = calloc(opts->iterations, sizeof(double));
  assert(data != NULL);

  const int structs_size = ceil((double)msgsize / sizeof(TestStruct));

  TestStruct *buffer = calloc(structs_size, sizeof(TestStruct));
  assert(buffer != NULL);

  int *gatherbuf = calloc(opts->size, sizeof(int));

  set_bcast_algorithm(bcast_algo);

  bool failed = false;

  const char msg[10] = {'h', 'e', 'l', 'l', 'o', ' ', 'w', 'o', 'r', '\0'};
  double dbls[3] = {0.123, 0.9812, 0.784};

  for (int i = 0; i < opts->iterations; i++) {
    for (int j = 0; j < structs_size; j++) {
      if (opts->rank == 0) {
        buffer[j] = (TestStruct){
            .a = j,
            .b = j + 0.123f,
            // .message = msg,
            // .doubles
            .sum_message = 908,
            .sum_doubles = 1.8882,
        };
        memcpy(buffer[j].message, msg, sizeof(msg));
        memcpy(buffer[j].doubles, dbls, sizeof(dbls));
      } else {
        memset(&buffer[j], 0, sizeof(buffer[j]));
      }
    }

    MPI_Barrier(MPI_COMM_WORLD);

    const double start = MPI_Wtime();
    int ret =
        broadcast_func(buffer, structs_size, test_type, 0, MPI_COMM_WORLD);
    const double end = MPI_Wtime();
    assert(ret == MPI_SUCCESS);

    // verify buffer
    int valid = 1;
    for (int i = 0; i < structs_size; i++) {
      if (buffer[i].a != i || buffer[i].b != i + 0.123f ||
          strcmp(buffer[i].message, "hello wor") != 0 ||
          buffer[i].sum_message != 908 || buffer[i].doubles[0] != 0.123 ||
          buffer[i].doubles[1] != 0.9812 || buffer[i].doubles[2] != 0.784 ||
          buffer[i].sum_doubles != 1.8882) {
        valid = 0;
        // fprintf(stderr, "rank: %d failed test\n", rank);
      }
    }

    // printf("TEST %d got: %.*s\n", rank, msgsize, buffer);

    // MPI_Barrier(MPI_COMM_WORLD);
    MPI_Gather(&valid, 1, MPI_INT, gatherbuf, 1, MPI_INT, 0, MPI_COMM_WORLD);

    if (opts->rank == 0) {
      bool allvalid = true;
      for (int i = 0; i < opts->size; i++)
        if (!gatherbuf[i])
          allvalid = false;
      if (allvalid) {
        //        fprintf(stderr, ">> %s passed test\n", name);
      } else {
        fprintf(stderr,
                "> Benchmarking broadcast struct %s: %d/%d (algo: %s)\n", name,
                i, opts->iterations, BCastAlgorithmStrings[bcast_algo]);
        fprintf(stderr, ">> %s FAILED test: ", name);
        for (int i = 0; i < opts->size; i++)
          if (!gatherbuf[i])
            fprintf(stderr, "%d,", i);
        fprintf(stderr, "\n");

        failed = true;
        break;
      }
    }

    MPI_Barrier(MPI_COMM_WORLD);
    data[i] = 1000 * (end - start); // milliseconds
  }

  double mean = 0.0, min = 0.0, max = 0.0, var = 0.0;
  statistics(data, opts->iterations, NULL, &mean, &min, &max, NULL, &var);
  free(data);

  free(buffer);

  MPI_Type_free(&test_type);

  if (opts->rank == 0) {
    if (failed) {
      fprintf(stderr, "Broadcast struct FAILED N=%d, %s (%s) msgsize=%d\n",
              opts->iterations, name,
              BCastAlgorithmStrings[get_bcast_algorithm()], msgsize);
    } else {
      const char *algo = BCastAlgorithmStrings[bcast_algo];

      if (strcmp(name, "optimized") == 0)
        algo = "BCastDefault";

      fprintf(stdout,
              "Broadcast struct "
              "N=%d\t%s\t(%s)\tmsgsize=%-7d\tmean=%-7.4lf\tvar=%-7.4lf\tmin=%-"
              "7.4lf\tmax=%-7.4lf "
              "(ms)\n",
              opts->iterations, name, algo, msgsize, mean, var, min, max);
    }
  }
  return failed;
}

static void test_broadcast_impl(const BroadcastBenchOpts *opts,
                                const int msgsize, char *buffer,
                                const char *name, Topo_broadcast broadcast_func,
                                enum BCastAlgorithm bcast_algo) {
  assert(broadcast_func != NULL);
  if (test_broadcast_impl_bytes(opts, msgsize, buffer, name, broadcast_func,
                                bcast_algo) != 0) {
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  // if (test_broadcast_impl_struct(msgsize, name, broadcast_func, bcast_algo)
  // !=
  //     0) {
  //   MPI_Abort(MPI_COMM_WORLD, 1);
  // };
}

static void test_broadcast(const BroadcastBenchOpts *opts, int msgsize) {
  MPI_Barrier(MPI_COMM_WORLD);
  char *buffer = malloc(msgsize);
  assert(buffer != NULL);

  const char *commasep = ",";
  const char *colonsep = ":";

  char *str = strdup(opts->types_str);
  char *outer_save;

  char *type = strtok_r(str, commasep, &outer_save);

  while (type != NULL) {
    char *inner_save;

    char *func_name = strtok_r(type, colonsep, &inner_save);
    assert(func_name != NULL);
    char *alg_name = strtok_r(NULL, colonsep, &inner_save);
    assert(alg_name != NULL);

    Topo_broadcast *func = topo_broadcast_default;
    if (strcmp(func_name, "mpi") == 0) {
      func = topo_broadcast_default;
    } else if (strcmp(func_name, "normal") == 0) {
      func = topo_broadcast_normal;
    } else if (strcmp(func_name, "optimized") == 0) {
      func = topo_bcast_optimized;
    } else if (strcmp(func_name, "magpie") == 0) {
      func = topo_broadcast_magpie;
    } else if (strcmp(func_name, "magpie2") == 0) {
      func = topo_broadcast_magpie2;
    } else if (strcmp(func_name, "interpipe") == 0) {
      func = topo_broadcast_magpie_pipelined;
    } else if (strcmp(func_name, "pipeline") == 0) {
      func = topo_broadcast_pipelined;
    } else {
      fprintf(stderr, "got unrecognized type: '%s\n'", func_name);
      assert(false);
    }

    enum BCastAlgorithm bcast_algo = BCastDefault;
    bool found_algo = false;
    for (int i = BCastDefault; i < BCastAlgorithmCount; i++) {
      if (strcmp(alg_name, BCastAlgorithmStrings[i]) == 0) {
        found_algo = true;
        bcast_algo = i;
        break;
      }
    }
    assert(found_algo == true);

    test_broadcast_impl(opts, msgsize, buffer, func_name, func, bcast_algo);

    type = strtok_r(NULL, commasep, &outer_save);
  }

  free(str);
  free(buffer);

  MPI_Barrier(MPI_COMM_WORLD);
}

void broadcast_bench(BroadcastBenchOpts opts) {
  fprintf(stderr, "opts.msgsizes_str: %s\n", opts.msgsizes_str);
  bool free_types_str = false;
  if (opts.types_str == NULL) {
    char all_types[2048] = {0};
    int offset = 0;
    offset += sprintf(all_types + offset, "mpi:BCastDefault");
    // char *types[] = {"normal", "magpie", "magpie2"};
    char *types[] = {"normal", "magpie", "interpipe"};
    for (int typei = 0; typei < sizeof(types) / sizeof(*types); typei++) {
      for (int i = BCastDefault; i < BCastAlgorithmCount; i++) {
        offset += sprintf(all_types + offset, ",%s:%s", types[typei],
                          BCastAlgorithmStrings[i]);
      }
    }
    opts.types_str = strdup(all_types);
    free_types_str = true;
  }

  if (opts.msgsizes_str == NULL) {
    test_broadcast(&opts, 1);
    test_broadcast(&opts, 1024);
  } else {
    char *str = strdup(opts.msgsizes_str);
    const char *delim = ",";
    char *save;

    char *token = strtok_r(str, delim, &save);
    while (token != NULL) {
      int msgsize = atoi(token);
      test_broadcast(&opts, msgsize);
      token = strtok_r(NULL, delim, &save);
    }
    free(str);
  }

  if (free_types_str)
    free(opts.types_str);
}
