#define _XOPEN_SOURCE 500

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <mpi.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "test.h"
#include "topo.h"

#include "broadcast.h"
#include "broadcast_bench.h"
// #include "tree.h"

#include "optimized_bcast.h"

/* Globals */
int rank, size;
bool quiet = false, broadcast = false;
int N = 100;
int WARMUP = 5;
bool verbose = false;
char *msgsize_str = NULL;
char *types_str = NULL;
Topology topology;
BroadcastCtx broadcastCtx;
bool repeat_without_topo = false;
int V = 0;
bool VERIFY_BCAST = false;
bool VERIFY_BCAST_EVERY_ITER = false;
bool include_without = false;

char **topology_files = NULL;
int n_topology_files;

void show_help(int argc, char **argv) {
  if (rank == 0) {
    printf("Usage: %s [-v] [-N NUM_ITERS] [-W WARMUPS] [-M MSGSIZES...] [-T "
           "TYPES...] [-V1|-V2] [-a] TOPOS...\n\n",
           argv[0]);
    printf(" -v\tenable verbose output\n");
    printf(" -V1\tverify bcast on last iteration\n");
    printf(" -V2\tverify bcast on every iteration\n");
    printf(" -a ALL\tdo the benchmark without topology as well\n");
    printf(" -W WARMUPS\tset the number of warmup iterations\n");
    printf(" -M MSGSIZES\tcomma seperated list of message sizes to test\n");
    printf(" -T TYPES\tcomma seperated list of broadcast types to test\n");
  }
}

void options(int argc, char **argv) {
  int o;
  while ((o = getopt(argc, argv, "V:N:W:M:T:va")) != -1) {
    switch (o) {
    case 'V':
      V = atoi(optarg);
      if (V == 1)
        VERIFY_BCAST = true;
      else if (V == 2)
        VERIFY_BCAST_EVERY_ITER = true;
      else
        assert(false);

      break;
    case 'N':
      N = atoi(optarg);
      break;
    case 'W':
      WARMUP = atoi(optarg);
      break;
    case 'M':
      msgsize_str = strdup(optarg);
      break;
    case 'T':
      types_str = strdup(optarg);
      break;
    case 'v':
      verbose = true;
      break;
    case 'a':
      include_without = true;
      break;
    default:
      show_help(argc, argv);
      exit(1);
    }
  }

  n_topology_files = argc - optind;
  topology_files = malloc(sizeof(char *) * n_topology_files);
  assert(topology_files != NULL);

  // fprintf(stderr, "msgsize_str=%s\n", msgsize_str);

  for (int i = 0, index = optind; index < argc; index++, i++)
    topology_files[i] = strdup(argv[index]);

  for (int i = 0; i < n_topology_files; i++)
    if (rank == 0)
      fprintf(stderr, "topology file %d: %s\n", i, topology_files[i]);
}

int read_file(FILE *f, char **out) {
  int buffer_cap = 2 * BUFSIZ + 1;
  int buffer_len = 0;
  char *buffer = malloc(buffer_cap);

  do {
    size_t read_bytes = fread(buffer + buffer_len, 1, BUFSIZ, f);
    buffer_len += read_bytes;
    if (read_bytes < BUFSIZ)
      break;

    if (buffer_len + BUFSIZ >= buffer_cap) {
      buffer_cap += BUFSIZ;
      buffer = realloc(buffer, buffer_cap);
    }
  } while (true);
  buffer[buffer_len] = '\0';

  *out = buffer;
  return buffer_len - 1;
}

void load_topology(const char *filename) {

  char *buffer = NULL;
  int size = 0;

  if (rank == 0) {
    FILE *file = fopen(filename, "r");
    assert(file != NULL);
    size = read_file(file, &buffer);
    assert(buffer != NULL);
    fclose(file);
  }

  MPI_Bcast(&size, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if (rank != 0) {
    buffer = malloc(size + 1);
    assert(buffer);
    memset(buffer, 0, size + 1);
  }

  MPI_Bcast(buffer, size, MPI_BYTE, 0, MPI_COMM_WORLD);

  parse_topology(buffer, &topology);

  memset(buffer, 0, size);
  free(buffer);
}

int main(int argc, char **argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  options(argc, argv);

  if (include_without) {
    char topology_def[1024];
    sprintf(topology_def, "cluster { group { 0-%d }}", size - 1);
    parse_topology(topology_def, &topology);

    topo_bcast_init(&topology, MPI_COMM_WORLD, &broadcastCtx);

    BroadcastBenchOpts opts = {
        .bench_name = "NONE",
        .rank = rank,
        .size = size,
        .verbose = verbose,
        .warmup = WARMUP,
        .iterations = N,
        .verify_bcast = VERIFY_BCAST,
        .verify_bcast_every_iter = VERIFY_BCAST_EVERY_ITER,

        .msgsizes_str = msgsize_str,
        .types_str = types_str,

        .topology = topology,
        .ctx = broadcastCtx,
    };
    broadcast_bench(opts);

    topo_bcast_finalize();
    topology_finalize(&topology);
  }

  for (int i = 0; i < n_topology_files; i++) {
    const char *topo_file = topology_files[i];
    if (rank == 0)
      fprintf(stderr, "Running benchmark for topology file %d/%d: %s\n", i,
              n_topology_files, topo_file);
    load_topology(topo_file);
    if (rank == 0)
      print_topology(stderr, &topology);

    topo_bcast_init(&topology, MPI_COMM_WORLD, &broadcastCtx);

    BroadcastBenchOpts opts = {
        .bench_name = topology_files[i],
        .rank = rank,
        .size = size,
        .verbose = verbose,
        .warmup = WARMUP,
        .iterations = N,
        .verify_bcast = VERIFY_BCAST,
        .verify_bcast_every_iter = VERIFY_BCAST_EVERY_ITER,

        .msgsizes_str = msgsize_str,
        .types_str = types_str,

        .topology = topology,
        .ctx = broadcastCtx,
    };
    broadcast_bench(opts);

    topo_bcast_finalize();
    topology_finalize(&topology);
  }

  for (int i = 0; i < n_topology_files; i++)
    free(topology_files[i]);

  if (msgsize_str)
    free(msgsize_str);
  if (types_str)
    free(types_str);

  MPI_Finalize();
}
