#!/usr/bin/env python3

import numpy as np
from numpy import where, unique
import networkx as nx
from matplotlib import pyplot as plt
import matplotlib as mpl
import hdbscan
import os
import time

from sklearn.cluster import (
    AgglomerativeClustering,
    SpectralClustering,
    KMeans,
    AffinityPropagation,
    DBSCAN,
)
from sklearn.manifold import TSNE

from graph_based_clustering import (
    ConnectedComponentsClustering,
    SpanTreeConnectedComponentsClustering,
)

from scipy.sparse.csgraph import minimum_spanning_tree

from ssstree import SSSTree


from scipy.cluster.hierarchy import fcluster

import kmedoids

SPAN_TREE_CLUSTERING = "span_tree"
SPECTRAL_CLUSTERING = "spectral"
SSSTREE_CLUSTERING = "ssstree"
AGGLOMATIVE_CLUSTERING = "agglomative"
BIPARTITE_SPECTRAL_CLUSTERING = "bipartite_spectral"
NJW_SPECTRAL_CLUSTERING = "njw"
K_MEDOIDS_CLUSTERING = "k_medoids"
AFFINITY_PROPAGATION_CLUSTERING = "affinity_propagation"
MST_CLUSTERING = "mst"
DBSCAN_CLUSTERING = "dbscan"
HDBSCAN_CLUSTERING = "hdbscan"
CONNECTED_COMPONENTS_CLUSTERING = "connected_components"

CLUSTER_METHODS = [
    AFFINITY_PROPAGATION_CLUSTERING,
    AGGLOMATIVE_CLUSTERING,
    BIPARTITE_SPECTRAL_CLUSTERING,
    DBSCAN_CLUSTERING,
    HDBSCAN_CLUSTERING,
    K_MEDOIDS_CLUSTERING,
    MST_CLUSTERING,
    NJW_SPECTRAL_CLUSTERING,
    SPAN_TREE_CLUSTERING,
    SPECTRAL_CLUSTERING,
    SSSTREE_CLUSTERING,
    CONNECTED_COMPONENTS_CLUSTERING,
]

DISTMATRIX_CLUSTER_METHODS = [
    DBSCAN_CLUSTERING,
    HDBSCAN_CLUSTERING,
    K_MEDOIDS_CLUSTERING,
    SSSTREE_CLUSTERING,
    CONNECTED_COMPONENTS_CLUSTERING,
]


STATIC_METHODS = set(
    [
        K_MEDOIDS_CLUSTERING,
        MST_CLUSTERING,
        SPECTRAL_CLUSTERING,
        AGGLOMATIVE_CLUSTERING,
    ]
)

DYNAMIC_METHODS = set(
    [
        SSSTREE_CLUSTERING,
        DBSCAN_CLUSTERING,
        HDBSCAN_CLUSTERING,
        CONNECTED_COMPONENTS_CLUSTERING,
    ]
)


def evaulate(A, f, ths, parameter_name="threshold"):
    prediction = f(A, precomputed=True, parameters={parameter_name: ths})
    return prediction.max()


def binary_search_for_clusters(
    A,
    f,
    low=0.000000001,
    high=1.0,
    incr=0.0000001,
    maxval=None,
    minval=1,
    discrete=False,
    parameter_name="threshold",
):
    if maxval is None:
        maxval = A.shape[0]
    ret = -1
    while low <= high:
        mid = (high + low) / 2
        # print('low', low, 'high', high, 'mid', mid)

        if discrete:
            mid = int(np.floor(mid))

        x = evaulate(A, f, mid, parameter_name=parameter_name)
        if x < minval:
            high = mid - incr
        elif x >= maxval:
            low = mid + incr
        else:
            ret = mid
            break

    print("BSEARCH FOUND", ret)
    return ret


def linear_search(A, f, parameter_name, low, high, incr, discrete=False):
    predictions = []
    seen = set()
    v = low
    while v <= high:
        if discrete:
            v = int(np.floor(v))

        try:
            prediction = f(A, precomputed=True, parameters={parameter_name: v})
        except:
            v += incr
            continue

        hashval = hash(prediction.data.tobytes())

        if hashval in seen:
            v += incr
            continue
        seen.add(hashval)

        groups = _get_clusters_from_labels(prediction)
        pred = {}
        for group in groups.values():
            id_ = min(group)
            for elem in group:
                pred[elem] = id_

        predictions.append((groups, pred))

        v += incr
    return predictions


def find_possible_clusterings(
    A,
    f,
    parameter_name,
    n=1000,
    c=100,
    low=0.000000000001,
    high=1.0,
    incr=0.000000001,
    discrete=False,
    linear=False,
):
    if linear:
        return linear_search(A, f, parameter_name, low, high, incr, discrete=discrete)

    predictions = []
    seen = set()
    ths = None
    maxval = A.shape[0]
    minval = 0

    def explore(ths):
        if ths == -1:
            return

        nonlocal predictions
        nonlocal maxval, minval
        prediction = f(A, precomputed=True, parameters={parameter_name: ths})

        # prediction = _connected_components_clustering(
        #     A, precomputed=True, parameters={"threshold": ths}
        # )
        # prediction = _ssstree_clustering(A, precomputed=True, parameters={"alpha": ths})

        prediction.flags.writeable = False
        hashval = hash(prediction.data.tobytes())

        if hashval in seen:
            return
        seen.add(hashval)

        groups = _get_clusters_from_labels(prediction)
        pred = {}
        for group in groups.values():
            id_ = min(group)
            for elem in group:
                pred[elem] = id_
        predictions.append((groups, pred))
        minval = max(minval, len(groups))
        maxval = min(maxval, len(groups))

    for i in range(c):
        if ths is not None:
            ths = binary_search_for_clusters(
                A,
                f,
                low=ths,
                high=high,
                maxval=maxval,
                incr=incr,
                discrete=discrete,
                parameter_name=parameter_name,
            )
            explore(ths)
            ths = binary_search_for_clusters(
                A,
                f,
                low=low,
                high=ths,
                minval=minval,
                incr=incr,
                discrete=discrete,
                parameter_name=parameter_name,
            )
            explore(ths)
        else:
            ths = binary_search_for_clusters(
                A,
                f,
                low=low,
                high=high,
                incr=incr,
                discrete=discrete,
                parameter_name=parameter_name,
            )
            explore(ths)
        if ths == -1:
            break

    return predictions

    # ths_min = A.min()
    # ths_max = A.max()

    # lspace = np.linspace(ths_min, ths_max, num=n)
    # unique_preds = {}
    # for ths in lspace:
    #     prediction = _connected_components_clustering(A, precomputed=True, parameters={"threshold": ths})
    #     prediction.flags.writeable = False
    #     hashval = hash(prediction.data.tobytes())
    #     if not hashval in unique_preds:
    #         unique_preds[hashval] = prediction
    # return unique_preds


class ClusteringSearchDef:
    def __init__(
        self, name, f, parameter_name, low, high, incr, discrete=False, linear=False
    ):
        self.name = name
        self.f = f
        self.parameter_name = parameter_name
        self.low = low
        self.high = high
        self.incr = incr
        self.discrete = discrete
        self.linear = linear

    def run_search(self, A):
        return find_possible_clusterings(
            A,
            self.f,
            self.parameter_name,
            low=self.low,
            high=self.high,
            incr=self.incr,
            discrete=self.discrete,
            linear=self.linear,
        )


def create_cluster_defs(A, n_nodes, n_cores):
    Amin = A.min()
    Amax = A.max()

    static_samples = n_cores
    dynamic_samples = 50

    def dynamic_inc(low, high):
        return (high - low) / dynamic_samples

    definitions = [
        ClusteringSearchDef(
            "Kmedoids",
            k_medoids_clustering,
            "n_clusters",
            1,
            n_cores,
            1,
            discrete=True,
            linear=True,
        ),
        ClusteringSearchDef(
            "Agglomative",
            _agglomative_clustering,
            "n_clusters",
            1,
            n_cores,
            1,
            discrete=True,
            linear=True,
        ),
        ClusteringSearchDef(
            "Spectral",
            _spectral_clustering,
            "n_clusters",
            1,
            n_cores,
            1,
            discrete=True,
            linear=True,
        ),
        ClusteringSearchDef(
            "DBSCAN",
            _dbscan_clustering,
            "eps",
            Amin,
            Amax,
            dynamic_inc(Amin, Amax),
            linear=True,
        ),
        ClusteringSearchDef(
            "HDBSCAN",
            hdbscan_clustering,
            "min_cluster_size",
            2,
            n_cores - 1,
            1,
            linear=True,
        ),
        ClusteringSearchDef(
            "SSSTree",
            _ssstree_clustering,
            "alpha",
            0.001,
            5.0,
            dynamic_inc(0.001, 5.0),
            linear=True,
        ),
        ClusteringSearchDef(
            "CComponents",
            _connected_components_clustering,
            "threshold",
            Amin,
            Amax,
            dynamic_inc(Amin, Amax),
            linear=True,
            # 0.000000001,
            # 0.00001,
            # 0.000000001,
        ),
    ]
    return definitions


def run_cluster_search(A, n_nodes, n_cores):
    definitions = create_cluster_defs(A, n_nodes, n_cores)

    results = {}
    timings = {}
    for d in definitions:
        print("Clustering using", d.name)
        start = time.time()
        search_results = d.run_search(A)
        end = time.time()
        took = end - start
        if not search_results:
            print("Failed", d.name)
            assert False
        timings[d.name] = took
        results[d.name] = search_results

    return results, timings


def _mst_clustering(n_clusters, data):
    clustering = minimum_spanning_tree(data)
    print("clustering", clustering)
    return False


def _span_tree_clustering(n_clusters, data):
    clustering = SpanTreeConnectedComponentsClustering(
        n_clusters=n_clusters,
        metric="euclidean",
        n_jobs=-1,
    )
    return clustering.fit_predict(data)


def _connected_components_clustering(data, precomputed=False, parameters=None):
    ths = parameters.get("threshold", 0.01) if parameters else 0.01
    print("ths", ths)
    ths = float(os.environ.get("CCC_THS", ths))

    clustering = ConnectedComponentsClustering(
        metric="euclidean",
        threshold=ths,
        n_jobs=-1,
    )
    if precomputed:
        clustering.set_params(metric="precomputed")
    pred = clustering.fit_predict(data)
    print("PRED", pred)
    return pred


def _spectral_clustering(data, precomputed=False, parameters=None):
    n_clusters = parameters.get("n_clusters", 1)
    clustering = SpectralClustering(n_clusters=n_clusters)
    if precomputed:
        clustering.set_params(affinity="precomputed")
    return clustering.fit_predict(data)


def _dbscan_clustering(data, precomputed=False, parameters=None):
    eps = parameters.get("eps", 0.1) if parameters else 0.1

    # clustering = DBSCAN(eps=0.1, n_jobs=-1)
    clustering = DBSCAN(eps=eps, n_jobs=-1)
    if precomputed:
        A = data
        matmax, matmin = A.max(), A.min()
        A = (A - matmin) / (matmax - matmin)
        print(A)
        clustering.set_params(metric="precomputed")
        return clustering.fit_predict(A)
    return clustering.fit_predict(data)


def hdbscan_clustering_levels(A, parameters=None, levels=3):
    min_cluster_size = 2
    min_samples = 1

    if parameters is not None:
        min_cluster_size = parameters["min_cluster_size"]
        min_samples = parameters.get("min_samples", min_cluster_size)

    clustering = hdbscan.HDBSCAN(
        min_cluster_size=min_cluster_size, min_samples=min_samples, metric="precomputed"
    )

    pred = clustering.fit_predict(A)
    # tree = clustering.condensed_tree_.to_pandas()
    # network = clustering.condensed_tree_.to_networkx()
    # print(network)

    clustering.single_linkage_tree_.plot()
    plt.show()

    tree = clustering.single_linkage_tree_

    hierarchy = tree.to_numpy()
    print(fcluster(hierarchy, 4, "maxclust"))

    clusters = tree.get_clusters(0.8 * 1e-6)
    print(clusters)

    # root = next(nx.topological_sort(network))
    # print("root", root)
    # print("Level 1---")

    # results = []
    # nodes = [root]
    # for level in range(levels):
    #     result = []

    #     for node in nodes:
    #         for ch in sorted(nx.descendants_at_distance(network, node, level)):
    #             desc = sorted(nx.descendants(network, ch))
    #             print("LEVEL", level, node, len(desc), desc)

    # print("Level 2---")
    # for ch in sorted(nx.descendants_at_distance(network, root, 2)):
    #     desc = sorted(nx.descendants(network, ch))
    #     print(len(desc), desc)

    # print("Level 3---")
    # for ch in sorted(nx.descendants_at_distance(network, root, 2)):
    #     desc = sorted(nx.descendants(network, ch))
    #     print(len(desc), desc)

    # print(nx.descendants_at_distance(network, root, 2))
    # print(nx.descendants_at_distance(network, root, 3))
    # nx.draw(network)
    # plt.show()

    # _bfs_condensed_tree(tree)

    # root = tree.parent.min()
    # print("ROOT", root)

    # unique_lambda_vals = tree.lambda_val.drop_duplicates()

    # root_val = tree.lambda_val.head(1)
    # print("root_val", root_val)

    # clusters = tree[tree.child_size > 1]

    # print("CLUSTERS")
    # print(clusters)
    # print("LAST", clusters[0])


def hdbscan_clustering(data, precomputed=False, parameters=None):
    min_cluster_size = 2
    # min_samples = 1

    if parameters is not None:
        min_cluster_size = parameters.get("min_cluster_size", min_cluster_size)
        # min_samples = parameters.get("min_samples", min_cluster_size)

    clustering = hdbscan.HDBSCAN(
        min_cluster_size=min_cluster_size  # , min_samples=min_samples
    )
    # clustering = hdbscan.HDBSCAN(min_cluster_size=16)
    if precomputed:
        A = data
        # matmax, matmin = A.max(), A.min()
        # A = (A - matmin) / (matmax - matmin)
        # print(A)
        clustering.set_params(metric="precomputed")
        # clustering.set_params(cluster_selection_method="leaf")

        pred = clustering.fit_predict(A)
        # tree = clustering.condensed_tree_.to_pandas()
        # clusters = tree[tree.child_size > 1]
        # print("clusters", clusters)
        # ct = clustering.condensed_tree_
        # _hdbscan_create_tree(ct)
        # print("CONDENSED TREE", ct)

        # clustering.condensed_tree_.plot()
        # plt.show()
        return pred
    return clustering.fit_predict(data)


def _create_dist_matrix(data, normalized=False):
    # print("DATA", data)
    n = data.shape[0]
    A = np.zeros((n, n))
    for i in range(n):
        for j in range(i, n):
            A[i][j] = data[i][j]  # first field
            A[j][i] = data[i][j]  # first field

    if normalized:
        matmax, matmin = A.max(), A.min()
        A = (A - matmin) / (matmax - matmin)

    return A


def _laplacian(A):
    n = A.shape[0]
    L = np.zeros_like(A)
    np.fill_diagonal(L, n - 1)
    L += A
    # L = np.negative(A)
    return L


def _plot_dist_matrix(dist_matrix):
    from matplotlib import pyplot as plt

    plt.matshow(dist_matrix)
    plt.colorbar()
    plt.show()


def _plot_dist_matrices(As):
    from matplotlib import pyplot as plt

    plt.figure(1)
    for i, A in enumerate(As):
        plt.subplot(131 + i)
        plt.imshow(A)
        plt.xticks(np.arange(A.shape[0]))
        plt.yticks(np.arange(A.shape[0]))
        plt.colorbar()

    # plt.matshow(A)
    plt.show()


def _eigen_n(mat, values, vectors, d: int):
    n = mat.shape[0]
    B = np.zeros_like(mat)
    for i in range(n):
        for j in range(n):
            if j < d:
                B[i][j] = vectors[i][j] * values[j]

    C = np.zeros_like(B)

    for i in range(n):
        for j in range(n):
            for k in range(n):
                C[i][j] += B[i][k] * vectors[j][k]

    return C


def _get_eigenvalues(A, standardize=False):
    (vals, vecs) = np.linalg.eigh(A)
    ind = vals.argsort()[::-1]
    vals = vals[ind]
    vecs = vecs[:, ind]
    cum_var_exp = np.cumsum(100 * abs(vals) / sum(abs(vals)))
    if standardize:
        vecs = vecs * np.sqrt(abs(vals))
    return vals, vecs, cum_var_exp


def k_medoids_clustering(data, precomputed=False, parameters=None):
    n_clusters = parameters.get("n_clusters", 1)
    if precomputed:
        A = data
    else:
        A = _create_dist_matrix(data, normalized=True)
    km = kmedoids.KMedoids(n_clusters, method="fasterpam")
    return km.fit_predict(A)


def _bipartite_spectral_clustering(data):
    # take lap
    A = _create_dist_matrix(data, normalized=True)
    L = _laplacian(A)
    # take eigenvalue decomposition of laplacian
    values, vectors, var = _get_eigenvalues(L)  # np.linalg.eigh(L)
    # bipartite split
    a = set()
    b = set()
    for i, value in enumerate(vectors[1]):
        if value < 0:
            a.add(i)
        else:
            b.add(i)

    labels = np.zeros(data.shape[0])
    for i in range(A.shape[0]):
        if i in a:
            labels[i] = 0
        else:
            labels[i] = 1
    return labels


def _njw_create_coords(n_clusters, data, precomputed=True):
    # take lap
    if precomputed:
        A = data
    else:
        A = _create_dist_matrix(data, normalized=True)
    n = A.shape[0]

    sigma = np.median(A)
    A = np.exp(-(A**2) / (2 * sigma))  # preprocess step as described in algorithm

    A.flat[:: n + 1] = 0.0
    # print("A", A)
    D = np.diag(A.sum(axis=1))
    invRootD = np.sqrt(np.linalg.inv(D))
    L = np.dot(invRootD, np.dot(A, invRootD))

    # take eigenvalue decomposition of laplacian
    values, vectors, var = _get_eigenvalues(L)  # np.linalg.eigh(L)
    # print("values", values)
    # print("vectors", vectors)
    # print("shape", vectors.shape)
    # print("var", var)
    X = np.transpose(vectors[:n_clusters, :])
    norm = np.linalg.norm(X, axis=1, keepdims=True)
    Y = X / norm
    # print("Y", Y)

    # dimensions = 3
    # coords_matrix = Y[:, :dimensions]
    # varexp = var[dimensions - 1]
    coords_matrix = Y

    def normalize_coords(coords_matrix):
        sqsum = np.sum(coords_matrix**2, axis=1).reshape(coords_matrix.shape[0], -1)
        return coords_matrix / np.sqrt(sqsum)

    normalize_coords(coords_matrix)
    coords = coords_matrix
    return coords


def _njw_spectral_clustering(n_clusters, data, precomputed=False):

    # ssstree = SSSTree(coords, ndims=coords.shape[1], distance=distance)
    # clusters = ssstree.clusters(min_cluster_size=2)
    # labels = np.zeros(coords.shape[0])
    # for i, cluster in enumerate(clusters.values()):
    #     for node in cluster:
    #         labels[node.index] = i
    # return labels

    # return _spectral_clustering(n_clusters, coords)

    coords = _njw_create_coords(n_clusters, data, precomputed=precomputed)
    km = KMeans(n_clusters=n_clusters)
    labels = km.fit_predict(coords)
    return labels


def _affinity_propagation_clustering(n_clusters, data, precomputed=False):
    if precomputed:
        model = AffinityPropagation(damping=0.7)
        model.set_params(affinity="precomputed")
        return model.fit_predict(data)
    else:
        model = AffinityPropagation(damping=0.7)
        return model.fit_predict(data)


def _ssstree_clustering(data, precomputed=False, alpha=None, parameters=None):
    d = parameters.get("eigen_d", 0) if parameters else 0
    alpha = parameters.get("alpha", alpha) if parameters else alpha

    if not precomputed:
        # take lap
        A = _create_dist_matrix(data, normalized=True)
    else:
        A = data

    if d > 0:
        values, vectors = np.linalg.eigh(A)
        E = _eigen_n(A, values, vectors, d)

        matmax, matmin = E.max(), E.min()
        D = (E - matmin) / (matmax - matmin)
        np.fill_diagonal(D, 0)
        # _plot_dist_matrices([A, E, D])
        A = D

    # this ssstree implementation takes a list of datapoints
    # but we can imitate a matrix by using indexes and a custom
    # distance function
    def distance(p, q):
        return A[p[0]][q[0]]

    new_data = np.zeros((A.shape[0], 1)).astype(int)
    for i in range(A.shape[0]):
        new_data[i][0] = i

    # ssstree = SSSTree(data, ndims=data.shape[1])
    if alpha is None:
        alpha = 0.35

    ssstree = SSSTree(new_data, ndims=new_data.shape[1], distance=distance, alpha=alpha)
    clusters = ssstree.clusters(min_cluster_size=4)

    labels = np.zeros(data.shape[0])
    for i, cluster in enumerate(clusters.values()):
        for node in cluster:

            labels[node.index] = i
    return labels


def _agglomative_clustering(data, precomputed=False, parameters=None):
    n_clusters = parameters.get("n_clusters", 1)

    if precomputed:
        A = data
    else:
        A = _create_dist_matrix(data, normalized=False)
        # L = _laplacian(A)
        # values, vectors = np.linalg.eigh(L)
        # E = _eigen_n(L, values, vectors, n_clusters)

    # _plot_dist_matrices([A, L, E])

    cluster = AgglomerativeClustering(n_clusters=n_clusters, linkage="ward")
    return cluster.fit_predict(A)


def _fit_predict(method, data, n_clusters=None):
    labels = None
    if method == SPAN_TREE_CLUSTERING:
        assert n_clusters is not None
        labels = _span_tree_clustering(n_clusters, data)
    elif method == SPECTRAL_CLUSTERING:
        assert n_clusters is not None
        labels = _spectral_clustering(n_clusters, data)
    elif method == SSSTREE_CLUSTERING:
        labels = ssstree_clustering(n_clusters, data)
    elif method == AGGLOMATIVE_CLUSTERING:
        labels = _agglomative_clustering(n_clusters, data)
    elif method == BIPARTITE_SPECTRAL_CLUSTERING:
        labels = _bipartite_spectral_clustering(data)
    elif method == NJW_SPECTRAL_CLUSTERING:
        labels = _njw_spectral_clustering(n_clusters, data)
    elif method == K_MEDOIDS_CLUSTERING:
        labels = k_medoids_clustering(n_clusters, data)
    elif method == AFFINITY_PROPAGATION_CLUSTERING:
        labels = _affinity_propagation_clustering(n_clusters, data)
    elif method == DBSCAN_CLUSTERING:
        labels = _dbscan_clustering(data)
    elif method == HDBSCAN_CLUSTERING:
        labels = hdbscan_clustering(data)
    elif method == CONNECTED_COMPONENTS_CLUSTERING:
        labels = _connected_components_clustering(n_clusters, data)
    else:
        return None

    return labels


def _get_clusters_from_labels(labels):
    clusters = {}
    for i, v in enumerate(labels):
        if not v in clusters:
            clusters[v] = set()
        clusters[v].add(i)
    return clusters


def predict_clusters(method, data, n_clusters=None, show_graph=False):
    labels = _fit_predict(method, data, n_clusters)

    groups = _get_clusters_from_labels(labels)

    # print("labels", labels)
    # print("groups", groups)

    pred = {}
    for group in groups.values():
        id_ = min(group)
        for elem in group:
            pred[elem] = id_

    if show_graph:
        projection = TSNE().fit_transform(data)
        plt.scatter(*projection.T)
        plt.show()

    return groups, pred


def _fit_predict_distmatrix(method, A, parameters):
    labels = None

    if method in STATIC_METHODS:
        n_clusters = parameters["n_clusters"]
        if method == AFFINITY_PROPAGATION_CLUSTERING:
            labels = _affinity_propagation_clustering(n_clusters, A, precomputed=True)
        elif method == AGGLOMATIVE_CLUSTERING:
            labels = _agglomative_clustering(n_clusters, A, precomputed=True)
        elif method == K_MEDOIDS_CLUSTERING:
            labels = k_medoids_clustering(A, parameters=parameters)
        elif method == MST_CLUSTERING:
            labels = _mst_clustering(n_clusters, A)
        elif method == NJW_SPECTRAL_CLUSTERING:
            labels = _njw_spectral_clustering(n_clusters, A, precomputed=True)
        elif method == SPECTRAL_CLUSTERING:
            labels = _spectral_clustering(n_clusters, A, precomputed=True)
    elif method in DYNAMIC_METHODS:
        if method == SSSTREE_CLUSTERING:
            labels = _ssstree_clustering(
                A, precomputed=True, alpha=parameters.get("alpha", None)
            )
        elif method == DBSCAN_CLUSTERING:
            labels = _dbscan_clustering(A, precomputed=True, parameters=parameters)
        elif method == HDBSCAN_CLUSTERING:
            labels = hdbscan_clustering(A, precomputed=True, parameters=parameters)
        elif method == CONNECTED_COMPONENTS_CLUSTERING:
            labels = _connected_components_clustering(
                A, precomputed=True, parameters=parameters
            )
    else:
        assert "Invalid method"
    # elif method == AGGLOMATIVE_CLUSTERING:
    #    labels = _agglomative_clustering(n_clusters, data, precomputed=True)
    # elif method == BIPARTITE_SPECTRAL_CLUSTERING:
    #    labels = _bipartite_spectral_clustering(data, precomputed=True)
    # elif method == K_MEDIODS_CLUSTERING:
    #    labels = _k_mediods_clustering(n_clusters, data, precomputed=True)
    # elif method == AFFINITY_PROPAGATION_CLUSTERING:
    #    labels = _affinity_propagation_clustering(n_clusters, data, precomputed=True)

    return labels


def _draw_weighted_network(G, clusters):
    # elarge = [(u, v) for (u, v, d) in G.edges(data=True) if d["weight"] < 0.3]
    elarge = []
    esmall = [(u, v) for (u, v, d) in G.edges(data=True)]

    pos = nx.spring_layout(
        G, seed=7, iterations=1000
    )  # positions for all nodes - seed for reproducibility

    # nodes
    for color, nodes in clusters:
        nodes = list(nodes)
        print(color, nodes)
        nx.draw_networkx_nodes(G, pos, nodelist=nodes, node_size=700, node_color=color)

    # edges
    nx.draw_networkx_edges(G, pos, edgelist=elarge, width=5)
    # nx.draw_networkx_edges(
    #     G, pos, edgelist=esmall, width=6, alpha=0.5, edge_color="b", style="dashed"
    # )
    nx.draw_networkx_edges(
        G, pos, edgelist=esmall, width=1, alpha=0.5, edge_color="b", style="dashed"
    )

    # node labels
    nx.draw_networkx_labels(G, pos, font_size=20, font_family="sans-serif")
    # edge weight labels
    edge_labels = nx.get_edge_attributes(G, "weight")
    nx.draw_networkx_edge_labels(G, pos, edge_labels)

    ax = plt.gca()
    ax.margins(0.08)
    plt.axis("off")
    plt.tight_layout()
    plt.show()


def _plot_distance_graph_with_clusters(A, clusters):
    from sklearn.preprocessing import normalize

    norm_A = 1 / normalize(A)

    def get_cmap(n, name="plasma"):
        """Returns a function that maps each index in 0, 1, ..., n-1 to a distinct
        RGB color; the keyword argument name must be a standard mpl colormap name."""
        return plt.cm.get_cmap(name, n)

    n = A.shape[0]
    d = len(clusters)

    def avg_distance(cluster_a, cluster_b):
        sum_distances = 0.0
        count = 0
        for a in cluster_a:
            for b in cluster_b:
                sum_distances += norm_A[a][b]
                count += 1
        print("avg dist", sum_distances / count)
        return sum_distances / count

    def min_distance(cluster_a, cluster_b):
        min_dist = 99999999
        for a in cluster_a:
            for b in cluster_b:
                min_dist = min(min_dist, norm_A[a][b])
        return min_dist

    def max_distance(cluster_a, cluster_b):
        max_dist = 0
        for a in cluster_a:
            for b in cluster_b:
                max_dist = max(max_dist, norm_A[a][b])
        return max_dist

    # G = nx.complete_graph(n)
    G = nx.empty_graph()
    # print('drawing graph', G)
    for i in range(d):
        G.add_node(i)

    for i in range(d):
        for j in range(i + 1, d):
            G.add_edge(i, j, weight=avg_distance(clusters[i], clusters[j]))

    colormap = get_cmap(len(clusters))
    colored_clusters = []
    for i, cluster in enumerate(clusters.values()):
        colored_clusters.append((colormap(i), [i]))

    # for i, cluster in enumerate(clusters.values()):
    #     colored_clusters.append((colormap(i), cluster))

    # nx.draw(G)
    _draw_weighted_network(G, colored_clusters)


def predict_clusters_distmatrix(
    method, A, n_clusters=None, show_graph=False, parameters=None
):
    # _plot_dist_matrix(A)

    if parameters is None:
        parameters = {}
    parameters["n_clusters"] = n_clusters
    labels = _fit_predict_distmatrix(method, A, parameters)
    groups = _get_clusters_from_labels(labels)

    # print("labels", labels)
    # print("groups", groups)

    pred = {}
    for group in groups.values():
        id_ = min(group)
        for elem in group:
            pred[elem] = id_

    if show_graph:
        if -1 in groups:
            n1 = max(groups.keys()) + 1
            if n1 not in groups:
                groups[n1] = set()
            for elem in groups[-1]:
                groups[n1].add(elem)
            del groups[-1]
        # _plot_distance_graph_with_clusters(A, groups)

        tsne = TSNE(init="random", metric="precomputed", n_jobs=-1)
        projection = tsne.fit_transform(A)
        print("projection", projection)
        plt.scatter(*projection.T)
        plt.show()

    return groups, pred
