import numpy as np
from numpy import linalg

import os

import itertools


def read_mat(filename):
    with open(filename, "r") as f:
        lines = f.readlines()
        rows = []
        for line in lines:
            row = list(map(float, line.split()))
            rows.append(row)

        return np.array(rows)


def make_symmetric(mat):
    return np.maximum(mat, mat.transpose())


def write_mat(mat, filename):
    with open(filename, "w") as f:
        for row in mat:
            for col in row:
                f.write(f"{col:.2e}")
                f.write(" ")
            f.write("\n")


def write_mat_str(mat, filename):
    with open(filename, "w") as f:
        for row in mat:
            for col in row:
                f.write(f'"{col}"')
                f.write(" ")
            f.write("\n")


def write_vec(vec, filename):
    with open(filename, "w") as f:
        for col in vec:
            f.write(f"{col:.2e}")
            f.write(" ")
        f.write("\n")


def read_metadata(dirpath):
    with open(f"{dirpath}/metadata", "r") as metadata_file:
        lines = metadata_file.readlines()
        mapping = {}
        hostnames = {}
        pairs = []
        for line in lines:
            parts = line.split()
            from_rank, from_host, to_rank, to_host, filename = parts
            to_rank = int(to_rank)
            from_rank = int(from_rank)
            hostnames[to_rank] = to_host
            hostnames[from_rank] = from_host

            if from_rank not in mapping:
                mapping[from_rank] = {}
            if to_rank not in mapping[from_rank]:
                mapping[from_rank][to_rank] = filename

            pairs.append((from_rank, to_rank, filename))

    return pairs, mapping, hostnames


def parse_nodes(lines):
    nodes = {}
    out = []
    max_rank = 0
    for line in lines:
        split = line.split()
        from_ = int(split[0])
        to = int(split[2])
        file = split[-1]

        from_node = split[1]
        to_node = split[3]

        if from_node not in nodes:
            nodes[from_node] = set()
        if to_node not in nodes:
            nodes[to_node] = set()

        nodes[from_node].add(from_)
        nodes[to_node].add(to)

        max_rank = max(max_rank, from_)
        max_rank = max(max_rank, to)
    return nodes, max_rank


def _read_meta_file2(lines, resultsfile):
    with open(resultsfile) as rfile:

        nodes, max_rank = parse_nodes(lines)

        rlines = rfile.readlines()

        v2 = False
        index = 0
        if rlines[0].startswith("VERSION 2"):
            v2 = True
            index = 1

        out = []
        while index < len(rlines):
            assert rlines[index].startswith("START")
            parts = rlines[index].split()
            from_ = int(parts[2])
            to = int(parts[4])

            rows = []
            index += 1
            while index < len(rlines) and not rlines[index].startswith("START"):
                row = list(map(float, rlines[index].split()))
                rows.append(row)
                index += 1

            mat = np.array(rows)
            if mat[0][0] != 0.0:
                print('skip')
                continue # skip
            if v2:
                # print('is v2')
                L = mat[0][1]
                np.delete(mat, 1, 1)  # remove latency column
                mat[0][1] = L
            out.append((from_, to, "NULL", mat))
        return out, max_rank + 1, nodes


def read_meta_file(directory):
    dir_prefix = os.path.dirname(directory)
    if dir_prefix != "":
        dir_prefix += "/"

    filepath = f"{directory}/metadata"
    assert os.path.isfile(filepath)
    with open(filepath, "r") as f:
        lines = f.readlines()

    resultsfile = f"{directory}/results"
    if os.path.isfile(resultsfile):
        # new file format
        # print("USING NEW FORMAT")
        return _read_meta_file2(lines, resultsfile)

    nodes = {}
    out = []
    max_rank = 0
    for line in lines:
        split = line.split()
        from_ = int(split[0])
        to = int(split[2])
        file = split[-1]

        from_node = split[1]
        to_node = split[3]

        if from_node not in nodes:
            nodes[from_node] = set()
        if to_node not in nodes:
            nodes[to_node] = set()

        nodes[from_node].add(from_)
        nodes[to_node].add(to)

        max_rank = max(max_rank, from_)
        max_rank = max(max_rank, to)

        mat = read_mat(file)
        if mat[0][0] != 0.0:
            print('skip')
            continue # skip
        out.append((from_, to, file, mat))
    return out, max_rank + 1, nodes


def load_logp_matrices(pairs, mapping, hostnames):
    size = len(hostnames)

    # Open matrix to determine dimensions
    tmpmat = read_mat(pairs[0][2])
    n = tmpmat.shape[0]
    MSGs = tmpmat[:, 0]  # first column is the message sizse

    # The values are indexed like RTTs[messagesize][sender][reciever]
    RTTs = np.zeros((n, size, size))
    OSs = np.zeros((n, size, size))
    ORs = np.zeros((n, size, size))
    Gs = np.zeros((n, size, size))

    all_matrices = []
    for (sender, reciever, filename) in pairs:
        mat = read_mat(filename)
        all_matrices.append(mat)

        for i, row in enumerate(mat):
            rtt = row[1]
            os = row[2]
            or_ = row[3]
            g = row[4]

            RTTs[i][sender][reciever] = rtt
            OSs[i][sender][reciever] = os
            ORs[i][sender][reciever] = or_
            Gs[i][sender][reciever] = g
    return MSGs, RTTs, OSs, ORs, Gs
