#!/usr/bin/env python3
import argparse

import subprocess

import numpy as np

from common import *

from sys import argv, stdin
from broadcast import extract_bcast_lines


def extract(line, keys, floats=None, ints=None):
    out = []
    parts = line.split()
    if not parts:
        return []
    out.append(parts[3])

    name = parts[4][1:].strip("() ")
    # out.append(parts[4][1:])
    out.append(name)

    for part in parts:
        kv = part.split("=")
        if kv[0] in keys:
            v = kv[1]
            if floats and kv[0] in floats:
                v = float(v)
            if ints and kv[0] in ints:
                v = int(v)
            out.append(v)
    return out


def create_optimized_func(best_msgsizes):
    out = """
#include <mpi.h>
#include "broadcast.h"

int topo_bcast_optimized(void *buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm) {
\tint dt_size = 0;
\tMPI_Type_size(datatype, &dt_size);

\tsize_t msgsize = (size_t) dt_size * (size_t) count;
    """

    max_best_msgsizes = []

    for msgsize, (name, algo, mean, var) in sorted(
        best_msgsizes.items(), key=lambda x: x[0]
    ):
        if len(max_best_msgsizes) == 0:
            max_best_msgsizes.append([msgsize, name, algo, mean, var])
            continue

        prev = max_best_msgsizes[-1]
        if prev[1] == name and prev[2] == algo:
            max_best_msgsizes[-1][0] = msgsize
            continue
        else:
            max_best_msgsizes.append([msgsize, name, algo, mean, var])

    # print(max_best_msgsizes)

    def add_call(msgsize, name, algo, indent=True):
        nonlocal out
        func_name = ""
        if name == "normal":
            func_name = f"BCastFuncs[{algo}]"
        elif name == "mpi":
            func_name = f"BCastFuncs[BCastDefault]"
        elif name == "magpie":
            func_name = f"topo_broadcast_magpie"
        elif name == "magpie2":
            func_name = f"topo_broadcast_magpie2"
        else:
            assert False

        if indent:
            out += "\t"
        out += f"\tset_bcast_algorithm({algo});\n"

        if indent:
            out += "\t"
        out += f"\treturn {func_name}(buffer, count, datatype, root, comm);\n"

    for (msgsize, name, algo, mean, var) in max_best_msgsizes:
        out += f"\tif (msgsize <= {msgsize}) {{\n"
        add_call(msgsize, name, algo)
        out += "\t}\n"

    add_call(msgsize, name, algo, indent=False)

    out += "}\n"
    print("OUT", out)
    return out


def show_optimized_func_latex(best_msgsizes):
    max_best_msgsizes = []

    for msgsize, (name, algo, mean, var) in sorted(
        best_msgsizes.items(), key=lambda x: x[0]
    ):
        if len(max_best_msgsizes) == 0:
            max_best_msgsizes.append([msgsize, name, algo, mean, var])
            continue

        prev = max_best_msgsizes[-1]
        if prev[1] == name and prev[2] == algo:
            max_best_msgsizes[-1][0] = msgsize
            continue
        else:
            max_best_msgsizes.append([msgsize, name, algo, mean, var])

    print(f"def choose_best_bcast(msgsize):")
    first = True
    for (msgsize, name, algo, mean, var) in max_best_msgsizes:
        if first:
            first = False
            print(f"\tif msgsize <= {msgsize}: ", end="")
        else:
            print(f"\telif msgsize <= {msgsize}: ", end="")
        if name == "magpie":
            print(f"use {name} with {algo}")
        else:
            print(f"use {algo}")
    print("\telse: ", end="")
    last_name = max_best_msgsizes[-1][1]
    last_algo = max_best_msgsizes[-1][2]
    if last_name == "magpie":
        print(f"use {last_name} with {last_algo}")
    else:
        print(f"use {last_algo}")


parser = argparse.ArgumentParser(prog="Create optimized bcast")
parser.add_argument("filename")
parser.add_argument("--outfile", default="optimized_bcast.c")
parser.add_argument("--database", type=str, default=None)
parser.add_argument("--show-latex", default=False, action="store_true")

args = parser.parse_args()

lines = []
if args.filename == "-":
    lines = stdin.readlines()
else:
    with open(args.filename) as f:
        lines = f.readlines()

lines = extract_bcast_lines(lines)

fields = ["N", "msgsize", "mean", "var"]
ints = set(["N", "msgsize"])
floats = set(["mean", "var"])
data = list(map(lambda x: extract(x, set(fields), ints=ints, floats=floats), lines))

msgsizes = {}
for (name, algo, N, msgsize, mean, var) in data:
    if name == "optimized":  # or name == "mpi":
        # we cant base our optimized version on the previous optimized version
        # also skip mpi
        continue
    # print(name, algo, msgsize, time)
    if not msgsize in msgsizes:
        msgsizes[msgsize] = []
    msgsizes[msgsize].append((name, algo, mean, var))

best_msgsizes = {}
for msgsize, entries in msgsizes.items():
    best = None
    best_perc = None
    for (name, algo, mean, var) in entries:
        percentile_time = mean + 1.65 * var
        print(name, algo, mean, var, "95th percentile", percentile_time)

        if best is None or percentile_time < best_perc:
            best = (name, algo, mean, var)
            best_perc = percentile_time

    best_msgsizes[msgsize] = best

print(best_msgsizes)
func_out = create_optimized_func(best_msgsizes)

if args.show_latex:
    show_optimized_func_latex(best_msgsizes)


with open(args.outfile, "w") as f:
    f.write(func_out)
print(f"Wrote file to '{args.outfile}'")
