#define _XOPEN_SOURCE 500
#include <assert.h>
#include <limits.h>
#include <math.h>
#include <mpi.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "test.h"
#include "topo.h"

#include "broadcast.h"
#include "broadcast_bench.h"
// #include "tree.h"

#define USE_OPTIMIZED 1
#define TUNE_PIPELINE 0

#if USE_OPTIMIZED
#include "optimized_bcast.h"
#endif

/* Globals */
int rank, size;
bool quiet = false, broadcast = false;
char *filename = NULL, *output = NULL;
int N = 100;
int WARMUP = 5;
bool verbose = false;
char *msgsize_str = NULL;
char *types_str = NULL;
char *topo_file = NULL;
Topology topology;
BroadcastCtx broadcastCtx;
bool repeat_without_topo = false;

void options(int argc, char **argv) {
  int o;
  while ((o = getopt(argc, argv, "f:o:N:W:M:T:qbvRt:")) != -1) {
    switch (o) {
    case 'f':
      filename = strdup(optarg);
      break;
    case 't':
      topo_file = strdup(optarg);
      break;
    case 'o':
      output = strdup(optarg);
      break;
    case 'N':
      N = atoi(optarg);
      break;
    case 'W':
      WARMUP = atoi(optarg);
      break;
    case 'M':
      msgsize_str = strdup(optarg);
      break;
    case 'T':
      types_str = strdup(optarg);
      break;
    case 'v':
      verbose = true;
      break;
    case 'R':
      repeat_without_topo = true;
      break;
    case 'q':
      quiet = true;
      break;
    case 'b':
      broadcast = true;
      break;
    }
  }
}

int read_file(FILE *f, char **out) {
  int buffer_cap = 2 * BUFSIZ + 1;
  int buffer_len = 0;
  char *buffer = malloc(buffer_cap);

  do {
    size_t read_bytes = fread(buffer + buffer_len, 1, BUFSIZ, f);
    buffer_len += read_bytes;
    if (read_bytes < BUFSIZ)
      break;

    if (buffer_len + BUFSIZ >= buffer_cap) {
      buffer_cap += BUFSIZ;
      buffer = realloc(buffer, buffer_cap);
    }
  } while (true);
  buffer[buffer_len] = '\0';

  *out = buffer;
  return buffer_len - 1;
}

void load_topology(const char *filename) {

  char *buffer = NULL;
  int size = 0;

  if (rank == 0) {
    FILE *file = fopen(filename, "r");
    assert(file != NULL);
    size = read_file(file, &buffer);
    assert(buffer != NULL);
    fclose(file);
  }

  MPI_Bcast(&size, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if (rank != 0) {
    buffer = malloc(size + 1);
    assert(buffer);
    memset(buffer, 0, size + 1);
  }

  MPI_Bcast(buffer, size, MPI_BYTE, 0, MPI_COMM_WORLD);

  parse_topology(buffer, &topology);

  memset(buffer, 0, size);
  free(buffer);
}

int cluster_n_leaves(Cluster *cluster) {
  if (cluster->leaf)
    return 1;
  int sum = 0;
  for (int i = 0; i < cluster->n_sub_clusters; i++)
    sum += cluster_n_leaves(&cluster->sub_clusters[i]);
  return sum;
}

void lgate_main(void) {
  test_init(&topology.cluster); // test_init should not modify cluster, and
                                // should copy all data it needs
  test_leaves(&topology.cluster);
  while (cluster_n_leaves(&topology.cluster) > 1)
    join_leaves(&topology.cluster);
  test_finalize(output);
}

#if 0
static double tune_pipeline_round(char *buffer, int msgsize, int iters,
                                  double *range_start, double *range_end,
                                  MPI_Comm comm) {
  int local_rank, local_size;
  MPI_Comm_rank(comm, &local_rank);
  MPI_Comm_size(comm, &local_size);

  double step = (*range_end - *range_start) / iters;
  double timings[iters];
  memset(timings, 0, sizeof(timings));

  if (step < 0.001)
    return *range_start;

  const int n = 1000;

  double iter_timings[n];

  // Collect stats from all ranks and use them to generate statistics
  double *all_statsdata = calloc(n * size, sizeof(double));
  assert(all_statsdata != NULL);

  if (rank == 0)
    printf("searching between %lf and %f, step=%lf\n", *range_start, *range_end,
           step);

  set_bcast_algorithm(BCastPipeline);

  for (int i = 0; i < iters; i++) {
    double factor = 0;
    if (local_rank == 0)
      factor = *range_start + i * step;

    MPI_Bcast(&factor, 1, MPI_DOUBLE, 0, comm);

    set_pipeline_chunk_factor(factor);
    MPI_Barrier(comm);

    int nchunks = local_size * factor;
    int chunk_size = msgsize / nchunks;
    int rest_count = msgsize - (chunk_size * nchunks);

    if (nchunks < 1 || chunk_size < 1)
      /* not a valid factor */
      continue;

    memset(iter_timings, 0, sizeof(timings));

    for (int j = 0; j < n; j++) {
      MPI_Barrier(comm);
      const double start = MPI_Wtime();
      int ret = topo_broadcast_normal(buffer, msgsize, MPI_BYTE, 0, comm);
      const double end = MPI_Wtime();
      assert(ret == MPI_SUCCESS);

      const double time = (end - start) * 1000;
      iter_timings[j] = time;
      MPI_Barrier(comm);
    }

    double mean = 0.0, min = 0.0, max = 0.0, var = 0.0;
    MPI_Gather(iter_timings, n, MPI_DOUBLE, all_statsdata, n, MPI_DOUBLE, 0,
               MPI_COMM_WORLD);

    statistics(all_statsdata, n * local_size, NULL, &mean, &min, &max, NULL,
               &var);

    timings[i] = mean;

    if (rank == 0)
      printf("factor %lf: mean=%.4lf var=%.4lf min=%.4lf max=%.4lf\n", factor,
             mean, var, min, max);
  }

  free(all_statsdata);

  // find the best increasing pair
  int best_iter = 0;
  for (int i = 1; i < iters; i++) {
    if (timings[i] < timings[i - 1] && timings[i - 1] < timings[best_iter])
      best_iter = i - 1;
  }

  double best_factor = *range_start + best_iter * step;
  if (rank == 0)
    printf("found best of round: timing=%lf, factor=%lf\n", timings[best_iter],
           best_factor);

  MPI_Bcast(&best_factor, 1, MPI_DOUBLE, 0, comm);

  *range_start = best_factor;
  *range_end = best_factor + step;
  return best_factor;
}

void tune_pipeline(void) {

  int msgsize = 1024 * 1024; // 1 megabyte
  char *buffer = malloc(msgsize);
  assert(buffer != NULL);

  MPI_Barrier(MPI_COMM_WORLD);

  int rounds = 5;
  int iters = 50;

  double range_start = 1.0;
  double range_end = 4.0;

  double factor = 2.0;
  for (int i = 0; i < rounds; i++) {
    if (rank == 0) {
      for (int j = 0; j < msgsize; j++) {
        buffer[j] = j % ('Z' - 'A' + 1) + 'A';
      }
    }
    factor = tune_pipeline_round(buffer, msgsize, iters, &range_start,
                                 &range_end, MPI_COMM_WORLD);
    MPI_Barrier(MPI_COMM_WORLD);
  }

  printf("chose pipeline chunk factor: %lf\n", factor);
  set_pipeline_chunk_factor(factor);
}
#endif

void run(char *topo_file) {
  if (topo_file != NULL) {
    load_topology(topo_file);
  } else {
    char topology_def[1024];
    sprintf(topology_def, "cluster { group { 0-%d }}", size - 1);
    parse_topology(topology_def, &topology);
  }

  if (rank == 0)
    print_topology(stderr, &topology);

  topo_bcast_init(&topology, MPI_COMM_WORLD, &broadcastCtx);

  if (broadcast) {
    assert(false);
    // set_broadcast_bench_options(topo_file, rank, size, verbose, msgsize_str,
    //                             types_str, N, WARMUP, true, false, topology,
    //                             broadcastCtx);
    // broadcast_bench();
  } else {
    lgate_main();
  }

  topo_bcast_finalize();
  topology_finalize(&topology);
}

int main(int argc, char **argv) {
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  options(argc, argv);

  if (repeat_without_topo && filename) {
    if (rank == 0)
      fprintf(stderr, "===== Running with topo file =====\n");
    run(filename);
    MPI_Barrier(MPI_COMM_WORLD);
    if (rank == 0)
      fprintf(stderr, "===== Running without topo file =====\n");
    run(NULL);
  } else {
    run(filename);
  }

  if (msgsize_str)
    free(msgsize_str);
  if (types_str)
    free(types_str);

  MPI_Finalize();
}
