#!/usr/bin/env python3
#
import os
import argparse
import numpy as np
import time

from common import read_meta_file

import sqlite3

# TODO: not completely done
# METADAT ISN't added
METADATA_TABLE_FIELDS = ["id", "name", "time", "n_ranks", "n_nodes"]

NODE_RANKS_TABLE_FIELDS = ["name", "rank"]
NODE_RANKS_FIELDS = ", ".join(NODE_RANKS_TABLE_FIELDS)
LOGP_TABLE_FIELDS = [
    "from_rank",
    "to_rank",
    "msgsize",
    "rtt",
    "o_r",
    "o_s",
    "g",
]
LOGP_FIELDS = ", ".join(LOGP_TABLE_FIELDS)


def create_tables(table_name, con):
    logp_table_name = f"{table_name}_logp"
    nodes_table_name = f"{table_name}_nodes"

    cur = con.execute(f"CREATE TABLE {nodes_table_name}({NODE_RANKS_FIELDS})")
    cur = con.execute(f"CREATE TABLE {logp_table_name}({LOGP_FIELDS})")
    con.commit()
    print("[=] Created tables")


def create_indexes(table_name, con):
    logp_table_name = f"{table_name}_logp"
    index = f"CREATE INDEX {logp_table_name}_index ON {logp_table_name}(from_rank,to_rank,msgsize)"
    con.execute(index)
    con.commit()
    print("[=] Created indexes")


def load_into_database(directory, table_name, con):
    metadata, ranks, nodes = read_meta_file(directory)

    logp_table_name = f"{table_name}_logp"
    nodes_table_name = f"{table_name}_nodes"

    create_tables(table_name, con)
    create_indexes(table_name, con)

    # insert nodes
    print("Inserting nodes...")
    node_data = []
    for name, ranks in nodes.items():
        for rank in ranks:
            node_data.append({"name": name, "rank": rank})
    cur = con.executemany(
        f"INSERT INTO {nodes_table_name} VALUES (:name, :rank)", node_data
    )
    con.commit()

    # list table
    cur.execute(f"SELECT * FROM {nodes_table_name}")
    print("[=] Inserted", len(cur.fetchall()), "elements")

    print("Inserting logp_table...", logp_table_name)
    logp_data = []
    for (from_, to, _, mat) in metadata:
        for row in mat:
            data = {
                "from_rank": int(from_),
                "to_rank": int(to),
                "msgsize": int(row[0]),
                "rtt": float(row[1]),
                "o_r": float(row[2]),
                "o_s": float(row[3]),
                "g": float(row[4]),
            }
            logp_data.append(data)
    cur = con.executemany(
        f"INSERT INTO {logp_table_name} VALUES (:from_rank, :to_rank, :msgsize, :rtt, :o_r, :o_s, :g)",
        logp_data,
    )
    con.commit()

    # list table
    cur.execute(f"SELECT * FROM {logp_table_name}")
    con.commit()
    print("[=] Inserted", len(cur.fetchall()), "elements")


def _load_nodes_from_database(table_name, con):
    cur = con.execute(f"SELECT name, rank FROM {table_name}")
    results = cur.fetchall()
    return results


def _load_matrix_from_database(table_name, con, from_rank, to_rank):
    cur = con.execute(
        f"SELECT msgsize,rtt,o_s,o_r,g FROM {table_name} WHERE from_rank=:from_rank AND to_rank=:to_rank ORDER BY msgsize",
        {"from_rank": from_rank, "to_rank": to_rank},
    )
    results = cur.fetchall()
    nrows = len(results)
    ncols = 5
    mat = np.zeros((nrows, ncols))
    for i, (msgsize, rtt, or_, os, g) in enumerate(results):
        mat[i] = np.array([msgsize, rtt, or_, os, g])
    return mat


def load_from_database(table_name, con):
    # read
    logp_table_name = f"{table_name}_logp"
    nodes_table_name = f"{table_name}_nodes"

    # read nodes to get all pairs
    ranks = set()
    node_rows = _load_nodes_from_database(nodes_table_name, con)
    nodes = {}
    for name, rank in node_rows:
        if not name in nodes:
            nodes[name] = set()
        nodes[name].add(rank)
        ranks.add(rank)

    # print('nodes', nodes)
    out = []
    for from_ in ranks:
        for to in ranks:
            if to <= from_:
                continue

            matrix = _load_matrix_from_database(logp_table_name, con, from_, to)
            out.append((from_, to, "db", matrix))

    # read matrices
    return out, max(ranks) + 1, nodes


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="Tool to load logp data into sqlite3 database"
    )
    parser.add_argument("directory")
    parser.add_argument("database", default="db.sqlite3")
    parser.add_argument("--table_name", default="default")
    parser.add_argument("--load-only", default=False, action="store_true")
    args = parser.parse_args()

    print("[=] Connecting to database", args.database)
    con = sqlite3.connect(args.database)
    if not args.load_only:
        load_into_database(args.directory, args.table_name, con)

    print("Loading from database")
    stime = time.time()
    data = load_from_database(args.table_name, con)
    etime = time.time()
    print("Loaded", len(data), "in", etime - stime, "seconds")

    con.close()
