/*
 * Description: PLogP measurements for LGate
 * Author: Jørgen Bele Reinfjell
 * Date: 21.06.2023
 *
 * The implementation is based on the paper:
 *   "Fast Measurement of LogP Parameters for Message Passing Platforms" by Kielmann et al.
 * Paper is available from here: https://link.springer.com/chapter/10.1007/3-540-45591-4_162
 *
 * Several techniques are based on their implementation "mpi_logp".
 */

/*
 * What it does:
 *   1. It uses the LGate test harness to do pairwise measuremenents of PLogP parameters.
 *      It only does it one way, so for a pair of ranks (i,j) we only test from i to j, not from j to i
 *      We do this for EVERY pair (i,j) where i < j
 *
 *   2. Sometimes we need additional measurements. They are stored in 'additional_results', and combined
 *      in the save_results step. This results in better quality results, but takes more time.
 *
 *   3. We save results to a directory choosen by 'output'. We create a file 'results' that consists of
 *      the PLogP parameters for a given message size, for a pair of ranks
 *
 *   4. We save a file with metadata to 'metadata'. This file contains ranks and assignment to nodes.
 */

#define _POSIX_C_SOURCE 200809L
#define _XOPEN_SOURCE 500

#include <assert.h>
#include <math.h>
#include <mpi.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "test.h"

#include "include/fini.h"
#include "topo.h"

#define MAX(a, b) ((a) < (b) ? (b) : (a))
#define MIN(a, b) ((a) < (b)) ? (a) : (b)

#define CLOCK_TIMINGS 1000
#define MAX_DIFF 0.20
#define HOSTNAME_SIZE 128
#define USE_ADDITIONAL_RESULTS 1
#define ALWAYS_SATURATE_G 1

#define T_EPSILON (1e-8)


#define CALCULATE_L(rtt0, gap0) (fmax(0.0, (rtt0 - 2*(gap0))/2.0))

typedef struct {
  double L;
  double rtt;
  double o_s;
  double o_r;
  double g;

  int32_t message_size;
} Logp_result;

typedef struct {
  bool verbose;

  int64_t repetitions;
  double epsilon;
  double rtt_factor;
  int64_t max_message_size;
  bool additional_measurements;
  bool g_saturation;
  int64_t g_saturation_limit;

  /* fast mode */
  bool fast;
  double fast_group_perc;

  /* computed from values above */
  int64_t num_datapoints;
  int *selected; // selected[rank] == 0 or 1
  NodeGroup *leaves;
  int leaves_count;
} Config;

enum ConfigKey {
  CK_None,
  CK_Verbose,
  CK_Repetitions,
  CK_Epsilon,
  CK_RTT_Factor,
  CK_Max_Message_Size,
  CK_Additional_Measurements,
  CK_G_Saturation,
  CK_G_Saturation_Limit,
  CK_Fast,
  CK_Fast_Group_Perc,
  ConfigKeys_Count,
};

const char *ConfigKeysStr[ConfigKeys_Count] = {
    NULL,   "verbose",         "repetitions", "epsilon", "rtt_factor", "max_message_size", "additional_measurements", "g_saturation", "g_saturation_limit",
    "fast", "fast_group_perc",
};

/*
 * Global variables
 */
static Config config;

extern int rank, size;

// array of all our results node_results[<rank>*size + i] is the i'th datapoint
// for communication with rank <rank>
static Logp_result *node_results;

#if USE_ADDITIONAL_RESULTS

typedef struct {
  Logp_result result;
  int from_rank;
  int to_rank;
} Logp_additional_result;

// sometimes we need more measurements to get accurate results
// store this in this dynamic array
typedef struct {
  Logp_additional_result *results;
  int size;
  int cap;
} Logp_additional_results;

// XXX GLOBAL:
static Logp_additional_results additional_results = {0};

static void add_to_additional_results(int from_rank, int to_rank,
                                      Logp_result result) {
  Logp_additional_result additional_result = {
      .result = result, .from_rank = from_rank, .to_rank = to_rank};
  if (additional_results.size + 1 >= additional_results.cap) {
    additional_results.cap *= 2;
    if (additional_results.cap == 0)
      additional_results.cap = 1;

    additional_results.results =
        realloc(additional_results.results,
                additional_results.cap * sizeof(Logp_additional_result));
    assert(additional_results.results != NULL);
  }
  additional_results.results[additional_results.size++] = additional_result;
}
#endif

static double logp_start_time;
static double logp_end_time;

static uint8_t *messages[2];
static MPI_Request *reqs;
static double clock_overhead;

static void statistics(double *nums, int count, double *sum, double *mean,
                       double *min, double *max, double *sum_sq, double *var) {
  double sum_0;
  double mean_0;
  double min_0;
  double max_0;
  double sum_sq0;
  double var0;

  if (!sum)
    sum = &sum_0;
  if (!mean)
    mean = &mean_0;
  if (!min)
    min = &min_0;
  if (!max)
    max = &max_0;
  if (!sum_sq)
    sum_sq = &sum_sq0;
  if (!var)
    var = &var0;

  *sum = 0.0;
  *sum_sq = 0.0;
  *min = 9e99;
  *max = 0.0;

  for (int i = 0; i < count; i++) {
    *sum += nums[i];
    *sum_sq += nums[i] * nums[i];
    if (*min > nums[i])
      *min = nums[i];
    if (*max < nums[i])
      *max = nums[i];
  }
  *mean = *sum / count;

  if (count > 1) {
    *var = (*sum_sq - (*sum * *sum) / count) / (count - 1);
    if (*var < 0.0)
      *var = 0.0;
  } else {
    *var = 0.0;
  }
}

static void load_config(char *config_path) {
  fini_element_t fini_config;
  fini_init(config_path, '#', &fini_config);
  int n_keys = fini_n_keys(&fini_config);
  fini_element_t *keys[n_keys];
  fini_keys(&fini_config, keys);

  config = (Config){
      .verbose = false,
      .repetitions = 10000,
      .epsilon = 0.01,
      .rtt_factor = 1.3,
      .max_message_size = 67108864,
      .additional_measurements = false,
      .g_saturation = false,
      .g_saturation_limit = 1<<16,

      .fast = false,
      .fast_group_perc = 0.5,

      .selected = NULL,
      .leaves = NULL,
      .leaves_count = 0,
  };

  for (int k = 0; k < n_keys; k++) {
    enum ConfigKey key = CK_None;
    for (int i = CK_None + 1; i < ConfigKeys_Count; i++) {
      if (strcmp(keys[k]->name, ConfigKeysStr[i]) == 0) {
        key = i;
        break;
      }
    }
    const char *value = keys[k]->value;

    switch (key) {
    case CK_Verbose:
      config.verbose = strcmp(value, "true") == 0;
      break;

    case CK_Additional_Measurements:
      config.additional_measurements = strcmp(value, "true") == 0;
      break;

    case CK_G_Saturation:
      config.g_saturation = strcmp(value, "true") == 0;
      break;

    case CK_G_Saturation_Limit:
      config.g_saturation_limit = strtol(value, NULL, 10);
      assert(config.g_saturation_limit >= 0);
      break;

    case CK_Repetitions:
      config.repetitions = strtol(value, NULL, 10);
      assert(config.repetitions >= 0);
      break;

    case CK_Max_Message_Size:
      config.max_message_size = strtol(value, NULL, 10);
      assert(config.max_message_size >= 0);
      break;

    case CK_Fast:
      config.fast = strcmp(value, "true") == 0;
      break;

    case CK_Fast_Group_Perc:
      config.fast_group_perc = strtod(value, NULL);
      assert(config.fast_group_perc > 0);
      break;

    case CK_Epsilon:
      config.epsilon = strtod(value, NULL);
      assert(config.epsilon > 0);
      break;

    case CK_RTT_Factor:
      config.rtt_factor = strtod(value, NULL);
      assert(config.rtt_factor > 0);
      break;

    case CK_None:
      /* fallthrough */
    default:
      fprintf(stderr, "ERR load_config: invalid key: %s\n", keys[k]->name);
      exit(1);
    }
  }

  int message_size_count = log2(config.max_message_size) + 1;
  config.num_datapoints = 1 + message_size_count; // add 1 for size 0

  fini_finalize(&fini_config);
}

// https://stackoverflow.com/questions/33058848/generate-a-random-double-between-1-and-1
static double randfrom(double min, double max) {
  double range = (max - min);
  double div = RAND_MAX / range;
  return min + (rand() / div);
}

static void select_ranks(Cluster *cluster) {
  config.selected = malloc(sizeof(int) * size);
  assert(config.selected != NULL);

  NodeGroup *leaves = NULL;
  int leaves_count = 0;
  // Store the leaves for later
  cluster_leaves(cluster, &leaves, &leaves_count);
  config.leaves = leaves;
  config.leaves_count = leaves_count;

  if (rank == 0) {
    for (int i = 0; i < leaves_count; i++) {
      bool added_from_group = false;
      for (int j = 0; j < leaves[i].n_ranks; j++) {
        double d = randfrom(0, 1.0);
        int l_rank = leaves[i].ranks[j];
        if (d < config.fast_group_perc) {
          config.selected[l_rank] = 1;
          fprintf(stderr, "selected rank: %d\n", l_rank);
          added_from_group = true;
        } else {
          config.selected[l_rank] = 0;
        }
      }

      if (!added_from_group) {
        assert(leaves[i].n_ranks > 0);
        // add the first rank in the group
        config.selected[leaves[i].ranks[0]] = 1;
      }
    }
  }

  MPI_Bcast(config.selected, size, MPI_INT, 0, MPI_COMM_WORLD);
}

// from logp_mpi.c
static void busy(void) {
  volatile int dummy;
  volatile int i; /* Huh?!: gcc optimization bug/oddity otherwise?! */

  dummy = 42;
  for (i = 0; i < 300; i++) {
    dummy++;
  }
}

// inspired by measure_clock_ovhd in logp_mpi
static double measure_clock_overhead() {
  // we first repeat a function that should have known performance
  // characteristics the busy() function
  // then we surround it by MPI_Wtime(), and subtract the value we got in the
  // previous step this is the overhead
  //

  double min_loop0 = INFINITY, min_loop1 = INFINITY;

  for (int i = 0; i < 5; i++) {
    // repeat 5 times for warmup

    double start = MPI_Wtime();
    for (int j = 0; j < CLOCK_TIMINGS; j++)
      busy();
    double stop = MPI_Wtime();
    double busy_avg = (stop - start) / CLOCK_TIMINGS;

    double t0[CLOCK_TIMINGS] = {0.0};
    double t1[CLOCK_TIMINGS] = {0.0};
    start = MPI_Wtime();
    for (int j = 0; j < CLOCK_TIMINGS; j++) {
      t0[j] = MPI_Wtime();
      busy();
      t1[j] = MPI_Wtime();
    }
    stop = MPI_Wtime();
    double sum = 0.0;
    for (int j = 0; j < CLOCK_TIMINGS; j++)
      sum += t1[j] - t0[j];
    double avg = sum / CLOCK_TIMINGS;

    min_loop0 = MIN(min_loop0, busy_avg);
    min_loop1 = MIN(min_loop1, avg);
  }
  return min_loop1 - min_loop0;
}

void test_init(Cluster *cluster) {
  logp_start_time = MPI_Wtime();

  char *config_path = getenv("LOGP_CONFIG");
  if (!config_path)
    config_path = "logp.ini";
  if (rank == 0)
    fprintf(stderr, "Loading config from: '%s'\n", config_path);

  load_config(config_path);

  if (config.fast) {
    // TODO: should not be done here
    srand(time(NULL));

    select_ranks(cluster);
  }

  if (rank == 0) {
    printf("[==] LogP init\n");
    printf("[==] repetitions: %ld\n", config.repetitions);
    printf("[==] epsilon: %g\n", config.epsilon);
    printf("[==] rtt_factor: %g\n", config.rtt_factor);
    printf("[==] max_message_size: %ld\n", config.max_message_size);
    printf("[==] additional_measurements: %d\n", config.additional_measurements);
    printf("[==] g_saturation: %d\n", config.g_saturation);
    printf("[==] g_saturation_limit: %ld\n", config.g_saturation_limit);
    printf("[==] fast: %d\n", config.fast);
    printf("[==] fast_group_perc: %lf\n", config.fast_group_perc);
  }

  // Allocate memory for messages
  messages[0] = malloc(config.max_message_size);
  messages[1] = malloc(config.max_message_size);

  reqs = calloc(config.max_message_size, sizeof(MPI_Request));

  node_results = calloc(size * config.num_datapoints, sizeof(Logp_result));
  clock_overhead = measure_clock_overhead();
  if (rank == 0) {
    fprintf(stderr, "clock overhead: %.10e\n", clock_overhead);
  }
}

static inline bool double_equal(double a, double b, double eps) {
  if (fabs(a - b) <= eps)
    return true;
  return false;
}

// Send one byte in both directions and return average time to send/recv
static double measure_pair_rtt_0(int rank_i, int rank_j) {
  int count = 0;
  char dummy[2] = {1, 0};

  double start = MPI_Wtime();
  if (rank == rank_i) {
    MPI_Ssend(&dummy[0], count, MPI_BYTE, rank_j, 0, MPI_COMM_WORLD);
    MPI_Recv(&dummy[1], count, MPI_BYTE, rank_j, 0, MPI_COMM_WORLD,
             MPI_STATUS_IGNORE);
  } else {
    MPI_Recv(&dummy[1], count, MPI_BYTE, rank_i, 0, MPI_COMM_WORLD,
             MPI_STATUS_IGNORE);
    MPI_Ssend(&dummy[0], count, MPI_BYTE, rank_i, 0, MPI_COMM_WORLD);
  }
  double end = MPI_Wtime();

  return (end - start) / 2.0;
}

static inline void share_int(int from, int to, int *value) {
  if (rank == from)
    MPI_Send(value, 1, MPI_INT, to, 0, MPI_COMM_WORLD);
  else
    MPI_Recv(value, 1, MPI_INT, from, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
}


static inline void share_byte(int from, int to, uint8_t *value) {
  if (rank == from)
    MPI_Send(&value, 1, MPI_BYTE, to, 0, MPI_COMM_WORLD);
  else
    MPI_Recv(&value, 1, MPI_BYTE, from, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
}

static inline void share_empty(int from, int to) {
  int value = from + to;
  if (rank == from)
    MPI_Send(&value, 1, MPI_BYTE, to, 0, MPI_COMM_WORLD);
  else
    MPI_Recv(&value, 1, MPI_BYTE, from, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
}

static void synchronize(int rank_i, int rank_j) {
  uint8_t val = rank_i + rank_j;
  share_byte(rank_i, rank_j, &val);
}


// measure rtt of a given message count
static double measure_pair_rtt_msg(int rank_i, int rank_j, int msgsize, int count,
                                   char *buffer) {

  assert(rank == rank_i || rank == rank_j);
  assert(count > 0);
  assert(msgsize >= 0);

  double start = MPI_Wtime();
  if (rank == rank_i) {
    // The measurer sends message_count messages in a row
    for (int i = 0; i < count; i++) {
      MPI_Send(buffer, msgsize, MPI_BYTE, rank_j, 0, MPI_COMM_WORLD);
    }
  } else if (rank == rank_j) {
    // The mirror recieves message_count messages in a row
    for (int i = 0; i < count; i++) {
      MPI_Recv(buffer, msgsize, MPI_BYTE, rank_i, 0, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE
      );
    }
  }
  share_empty(rank_j, rank_i);

  double end = MPI_Wtime();
  return (end - start) / count;
}

static double measure_latency(int rank_i, int rank_j) {
  return measure_pair_rtt_0(rank_i, rank_j);
}

static double measure_gap(int sender_rank, int reciever_rank, double rtt0,
                          int msgsize, char *message) {

  // First we want to congest the pipeline
  int n = 10;
  int stable = 0;
  double rtt = rtt0;

  double rtt_sum = 0;
  int iters = 0;

  const int min_runs = 250;
  const int max_runs = 200 * min_runs;

  // Measure gap by link saturation
  while ((!stable || n < min_runs) && n <= max_runs) {
    iters++;
    n *= 2;

    synchronize(sender_rank, reciever_rank);

    double start_time = MPI_Wtime();

    // send n messages and measure time it takes
    double rtt_ = measure_pair_rtt_msg(sender_rank, reciever_rank, msgsize, n, message);
    double total_time = MPI_Wtime() - start_time;

    if (rtt < (1 + config.epsilon) * rtt_ && rtt / total_time < config.epsilon) {
      stable = 1;
    }
    share_int(sender_rank, reciever_rank, &stable);

    rtt = rtt_;
  }

  return rtt;
}

static void measure_overhead(int sender_rank, int reciever_rank, double rtt0,
                             double gap0, int msgsize, double *timings0,
                             double *timings1, int maxcount, double *os_mean,
                             double *os_var, double *or_mean, double *or_var,
                             double *rtt_mean, double *rtt_min,
                             double *rtt_var) {
  synchronize(sender_rank, reciever_rank);

  int iter = 0;

  // Timing of sending messages
  while (iter < maxcount) {
    double start = MPI_Wtime();

    double send_time = -9e99;

    if (rank == sender_rank) {
      MPI_Ssend(messages[0], msgsize, MPI_BYTE, reciever_rank, 0,
                MPI_COMM_WORLD);
      send_time = MPI_Wtime() - start;
      MPI_Recv(messages[0], 0, MPI_BYTE, reciever_rank, 0, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
    } else {
      MPI_Recv(messages[0], msgsize, MPI_BYTE, sender_rank, 0, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
      MPI_Ssend(messages[0], 0, MPI_BYTE, sender_rank, 0, MPI_COMM_WORLD);
    }

    double end = MPI_Wtime();
    timings0[iter] = end - clock_overhead - start;
    timings1[iter] = send_time;
    iter++;
  }

  // find the mean and save it as rtt
  double rtt_max = 0.0;
  statistics(timings0, iter, NULL, rtt_mean, rtt_min, &rtt_max, NULL, rtt_var);
  statistics(timings1, iter, NULL, os_mean, NULL, NULL, NULL, os_var);

  // Timing of recieving messages
  iter = 0;
  while (iter < maxcount) {
    double recieve_time = -9e99;
    if (rank == sender_rank) {
      MPI_Send(messages[0], 0, MPI_BYTE, reciever_rank, 0, MPI_COMM_WORLD);
      double end_at = MPI_Wtime() + config.rtt_factor * *rtt_mean;
      // spin wait
      volatile int i = 0;
      while (MPI_Wtime() < end_at) {
        // busy wait
        i = !i;
      }
      double start = MPI_Wtime();
      MPI_Recv(messages[0], msgsize, MPI_BYTE, reciever_rank, 0, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
      recieve_time = MPI_Wtime() - start - 2*clock_overhead;
    } else {
      MPI_Recv(messages[0], 0, MPI_BYTE, sender_rank, 0, MPI_COMM_WORLD,
               MPI_STATUS_IGNORE);
      MPI_Send(messages[0], msgsize, MPI_BYTE, sender_rank, 0, MPI_COMM_WORLD);
    }
    double end = MPI_Wtime();
    timings0[iter] = recieve_time;
    iter++;
  }

  statistics(timings0, iter, NULL, or_mean, NULL, NULL, NULL, or_var);
}

static inline bool has_gap_stabilized(const Logp_result *largest,
                                      const Logp_result *prev) {
  double gap_per_byte_0 = largest->g / (double)largest->message_size;
  double gap_per_byte_1 = prev->g / (double)prev->message_size;
  return (double_equal(gap_per_byte_0, gap_per_byte_1, T_EPSILON));
}

static inline bool has_nonlinear_behaviour(const Logp_result *curr,
                                           const Logp_result *prev,
                                           const Logp_result *prev2) {
  // extrapolate o_s, o_r, and g for curr ba// sed on previous two messages

  int32_t size_diff = prev->message_size - prev2->message_size;

  double y_o_s = (prev->o_s + prev2->o_s) / size_diff;
  double y_o_r = (prev->o_r + prev2->o_r) / size_diff;
  double y_g = (prev->g + prev2->g) / size_diff;

  double o_s = prev->o_s + size_diff * y_o_s;
  double o_r = prev->o_r + size_diff * y_o_r;
  double g = prev->g + size_diff * y_g;

  // fprintf(stderr, "expected o_s=%12g\to_r=%12g\tg=%12g\n", o_s, o_r, g);
  // fprintf(stderr, "got      o_s=%12g\to_r=%12g\tg=%12g\n", curr->o_s,
  // curr->o_r, curr->g);

  double o_s_err = 0.0, o_r_err = 0.0, g_err = 0.0;
  if (o_s > T_EPSILON)
    o_s_err = fabs((o_s - curr->o_s) / o_s);
  if (o_r > T_EPSILON)
    o_r_err = fabs((o_r - curr->o_r) / o_r);
  if (g > T_EPSILON)
    g_err = fabs((g - curr->g) / g);

  if (o_s_err > MAX_DIFF || o_r_err > MAX_DIFF || g_err > MAX_DIFF) {
    // fprintf(stderr, "error    o_s=%12g\to_r=%12g\tg=%12g\n", o_s_err,
    // o_r_err, g_err);
    return false;
  }

  return true;
}

static inline double measure_or_estimate_g(int rank_i, int rank_j, double rtt, double rtt0, double rtt0_min, double gap0, int msgsize) {
    double g = rtt - rtt0 + gap0;
    if (rtt <= rtt0)
      g = rtt - rtt0_min + gap0;
    if (g < 0)
      g = T_EPSILON;

    if (config.g_saturation && msgsize <= config.g_saturation_limit) {
      double estimate = g;
      g = measure_gap(rank_i, rank_j, rtt0, msgsize, messages[0]);
      if (config.verbose && rank == rank_i) {
        fprintf(stderr, "Used saturation for g(m) m=%d, real=%.10e, estimate=%.10e, rtt <= rtt0: %d, g(0)=%g, rtt(m)=%g, rtt(0)=%g\n", msgsize, g, estimate, rtt <= rtt0, gap0, rtt, rtt0);
      }
    }

    return g;
}

static inline void measure_until_stabilized(
    const int rank_i, const int rank_j, double rtt0, double rtt0_min, double gap0,
    double *timings0, double *timings1, int timings_count,
    const Logp_result *curr, const Logp_result *prev, const Logp_result *prev2,
    Logp_result **measurements_out, int *measurements_count_out,
    int *measurements_cap) {
  int mcap = MAX(*measurements_cap, 64);
  int msize = 0;
  Logp_result *measurements = *measurements_out;

  if (!measurements)
    measurements = malloc(mcap * sizeof(Logp_result));

  assert(measurements != NULL);

  measurements[msize++] = *prev2;
  measurements[msize++] = *prev;

  int done = 0;

  // stop condition: m_k - m_{k-1} <= max(32, config. * m_k), or linear
  for (;;) {
    if (measurements[msize - 1].message_size -
            measurements[msize - 2].message_size <=
        MAX(32, config.epsilon * curr->message_size))
      done = true;

    share_int(rank_i, rank_j, &done);

    if (done)
      /* reached end condition */
      break;

    // measure m = (m_{k-1} + m_k) /2
    int32_t message_size =
        (curr->message_size + measurements[msize - 1].message_size) / 2.0;

    fprintf(
        stdout,
        "measuring size %d to fix non-linear problems, prev: %d, curr : % d\n ",
        message_size, measurements[msize - 1].message_size, curr->message_size);

    double rtt = 0.0, rtt_var = 0.0, rtt_min = 0.0, o_s = 0.0, o_s_var = 0.0,
           o_r = 0.0, o_r_var = 0.0;
    measure_overhead(rank_i, rank_j, rtt0, gap0, message_size, timings0, timings1,
                     timings_count, &o_s, &o_s_var, &o_r, &o_r_var, &rtt,
                     &rtt_min, &rtt_var);

    // check if we need to reallocate
    if (msize + 1 >= mcap) {
      mcap *= 2;
      measurements = realloc(measurements, mcap * sizeof(Logp_result));
      assert(measurements != NULL);
    }

    double g = measure_or_estimate_g(rank_i, rank_j, rtt, rtt0, rtt0_min, gap0, message_size);

    measurements[msize] = (Logp_result){
        .message_size = message_size,
        .L = CALCULATE_L(rtt0, gap0),
        .rtt = rtt,
        .o_s = o_s,
        .o_r = o_r,
        .g = g,
    };
    msize++;

    if (!has_nonlinear_behaviour(&measurements[msize - 1],
                                 &measurements[msize - 2],
                                 &measurements[msize - 3]))
      /* no non-linear bahaviour */
      done = true;

    share_int(rank_i, rank_j, &done);

    if (done)
      /* reached end condition */
      break;
  }
  // fprintf(stderr, "STABILIZED after: %d repeated measurements\n", size-2);

  *measurements_out = measurements;
  *measurements_count_out = msize;
}

void test_pair(int rank_i, int rank_j) {
  Logp_result *results = &node_results[rank_j * config.num_datapoints];

  if (config.fast && (!config.selected[rank_i] || !config.selected[rank_j])) {
    // don't include this pair
    if (rank == rank_i)
      fprintf(stderr, "skipping (%d %d) because of fast flag\n", rank_i,
              rank_j);

    // fill results data with recognizable values
    int iter = 0;
    for (int msgsize = 1; msgsize <= config.max_message_size; msgsize *= 2) {
      results[iter] = (Logp_result){
          .message_size = msgsize,
          .L = INFINITY,
          .rtt = INFINITY,
          .o_s = INFINITY,
          .o_r = INFINITY,
          .g = INFINITY,
      };
      iter++;
    }

    return;
  }

  // Logp_result *results = calloc(config.max_message_size + 1,
  // sizeof(Logp_result));

  const int timings_count = config.repetitions;
  double *timings0 = calloc(timings_count, sizeof(double));
  double *timings1 = calloc(timings_count, sizeof(double));

  // First we measure RTT(0)
  // Warmup
  for (int i = 0; i < 200; i++) {
    measure_pair_rtt_0(rank_i, rank_j);
  }

  double rtt0 = 0.0, rtt0_min = 0.0;
  for (int i = 0; i < timings_count; i++) {
    timings0[i] = measure_pair_rtt_0(rank_i, rank_j);
  }
  statistics(timings0, timings_count, NULL, &rtt0, &rtt0_min, NULL, NULL, NULL);

  double gap0 = measure_gap(rank_i, rank_j, rtt0, 0, messages[0]);

  if (config.verbose && rank == rank_i)
    fprintf(stderr, "rtt(0)=%g, rtt_min(0)=%g, g(0)=%g\n", rtt0, rtt0_min, gap0);

  // Estimate L using rtt0 and gap0
  double L = CALCULATE_L(rtt0, gap0);

  if (config.verbose && rank == rank_i)
    fprintf(stderr, "%8s rtt:%6s o_s: %6s o_r: %6s g:%6s L=%lf, g(0)=%lf\n", "",
            "", "", "", "", L, gap0);
  int iter = 0;

  // store L and gap0 in first row
  results[iter] = (Logp_result){
      .message_size = 0,
      .L = L,
      .rtt = rtt0,
      .o_s = 0,
      .o_r = 0,
      .g = gap0,
  };
  if (config.verbose && rank == rank_i)
    fprintf(stderr, "%7d %.10e %.10e %.10e %.10e %.10e\n", results[iter].message_size,
            results[iter].L,
            results[iter].rtt, results[iter].o_s, results[iter].o_r,
            results[iter].g);
  iter++;

  for (int msgsize = 1; msgsize <= config.max_message_size; msgsize *= 2) {
    synchronize(rank_i, rank_j);

    // Then we want to find the overhead of sending
    // and recieving messages, o_s and o_r
    double rtt = 0.0, rtt_var = 0.0, rtt_min = 0.0, o_s = 0.0, o_s_var = 0.0,
           o_r = 0.0, o_r_var = 0.0;
    measure_overhead(rank_i, rank_j, rtt0, gap0, msgsize, timings0, timings1,
                     timings_count, &o_s, &o_s_var, &o_r, &o_r_var, &rtt,
                     &rtt_min, &rtt_var);

    double g = measure_or_estimate_g(rank_i, rank_j, rtt, rtt0, rtt0_min, gap0, msgsize);

    results[iter] = (Logp_result){
        .message_size = msgsize,
        .L = L, // NOTE: we store the rtt, so TODO is to change to .rtt
        .rtt = rtt,
        .o_s = o_s,
        .o_r = o_r,
        .g = g,
    };

    if (config.verbose && rank == rank_i) {
      fprintf(stderr, "%7d %.10e %.10e %.10e %.10e BANDWIDTH=%g\n", msgsize, results[iter].L,
              o_s, o_r, results[iter].g, 2*(msgsize)/rtt);
    }

    iter++;
  }
  assert(iter == config.num_datapoints);

  int gap_stable = 0;
  if (rank == rank_i)
    gap_stable = has_gap_stabilized(&results[iter - 1], &results[iter - 2]);
  share_int(rank_i, rank_j, &gap_stable);

  // TODO: we should check that the gap has stabilized for large m
  if (!gap_stable) {
    fprintf(stderr, "**WARNING** GAP DIDN'T STABILIZE for pair (%d,%d)\n",
            rank_i, rank_j);
  }

#if USE_ADDITIONAL_RESULTS
  {
    Logp_result *measurements = NULL;
    int mcap = 0;
    for (int i = 2; i < iter; i++) {
      int msize = 0;

      int check = false;
      if (rank == rank_i)
        check = has_nonlinear_behaviour(&results[i], &results[i - 1],
                                        &results[i - 2]);
      share_int(rank_i, rank_j, &check);

      // fprintf(stderr, "exchanged check: %d\n", check);
      check = true;

      if (check) {
        // fprintf(stderr, "**WARNING** FOUND NONLINEAR BEHAVIOUR for pair
        // (%d,%d) at index %d with message_size=%d\n", rank_i, rank_j, i,
        // results[i].message_size);
        measure_until_stabilized(rank_i, rank_j, rtt0, rtt0_min, gap0, timings0,
                                 timings1, timings_count, &results[i],
                                 &results[i - 1], &results[i - 2],
                                 &measurements, &msize, &mcap);

#if 0
        // first two elements are the previous results
        // TODO: save this to data
        if (rank == rank_i) {
          for (int k = 2; k < msize; k++) {
            Logp_result res = measurements[k];
            fprintf(stderr, "%7d %g %g %g %g\t(*)\n", res.message_size, res.L,
                    res.o_s, res.o_r, res.g);
          }
        }
#endif

        if (rank == rank_i) {
          for (int k = 2; k < msize; k++) {
            // Save these measurements
            add_to_additional_results(rank_i, rank_j, measurements[k]);
          }
        }
      }
    }
#endif

    free(measurements);
  }

  fprintf(stderr, "done...");

  free(timings1);
  free(timings0);
}

// let root collect all additional datapoints from ranks
static void collect_additional_datapoints(int root_rank, MPI_Datatype result_type,
                                          Logp_additional_result **out,
                                          int *out_count,
                                          int **out_displacements,
                                          int **out_counts) {
  /* Create type for additional result */
  int blocklengths[3] = {1, 1, 1};
  MPI_Datatype types[3] = {result_type, MPI_INT, MPI_INT};
  MPI_Aint offsets[3] = {offsetof(Logp_additional_result, result),
                         offsetof(Logp_additional_result, from_rank),
                         offsetof(Logp_additional_result, to_rank)};
  MPI_Datatype additional_result_type;
  MPI_Type_create_struct(sizeof(blocklengths) / sizeof(*blocklengths),
                         blocklengths, offsets, types, &additional_result_type);
  MPI_Type_commit(&additional_result_type);

  /* First gather counts and sum of counts */
  int counts[1] = {additional_results.size};

  fprintf(stderr, "%d: counts[0] = %d\n", rank, counts[0]);
  int *all_counts = NULL;
  int *displacements = NULL;
  if (rank == root_rank) {
    all_counts = calloc(size, sizeof(int));
    assert(all_counts != NULL);
    displacements = calloc(size, sizeof(int));
    assert(displacements != NULL);
  }

  // do the gather
  MPI_Gather(counts, 1, MPI_INT, all_counts, 1, MPI_INT, root_rank, MPI_COMM_WORLD);

  int total_count = 0;
  MPI_Reduce(counts, &total_count, 1, MPI_INT, MPI_SUM, root_rank, MPI_COMM_WORLD);
  if (rank == 0)
    fprintf(stderr, "total count=%d\n", total_count);

  /* Then allocate all results and set displacements based on counts */
  Logp_additional_result *all_results = NULL;
  if (rank == 0) {
    all_results = calloc(total_count, sizeof(Logp_additional_result));
    assert(all_results != NULL);

    int sum = 0;
    for (int i = 0; i < size; i++) {
      displacements[i] = sum;
      fprintf(stderr, "displacements[%d]=%d\n", i, displacements[i]);
      sum += all_counts[i];
    }
  }

  /* Then do the gatherv call */
  MPI_Gatherv(additional_results.results, additional_results.size,
              additional_result_type, all_results, all_counts, displacements,
              additional_result_type, root_rank, MPI_COMM_WORLD);

  for (int i = 0; i < size; i++) {
    if (rank == root_rank) {
      fprintf(stderr, "counts[%d] = %d\n", i, all_counts[i]);
    }
  }

  if (rank == 0) {
    for (int i = 0; i < total_count; i++) {
      fprintf(stderr, "result[%d] = {from=%d, to=%d, result=...}\n", i,
              all_results[i].from_rank, all_results[i].to_rank);
    }
  }

  // free type
  MPI_Type_free(&additional_result_type);

  if (rank == 0) {
    *out_displacements = displacements;
    *out_counts = all_counts;
    *out = all_results;
    *out_count = total_count;
  }
}
#define RANK_HOSTNAME(rank) (&hostnames[rank * HOSTNAME_SIZE])

static const Logp_additional_result *
find_next_in_additional(const Logp_additional_result *results, int count,
                        int offset, int *index, int from, int to) {
  for (int i = offset; i < count; i++) {
    if (results[i].from_rank == from && results[i].to_rank == to) {
      *index = i + 1;
      return &results[i];
    }
  }
  *index = count;
  return NULL;
}

static int save_results2(Logp_result *all_results,
                         Logp_additional_result *additional_results,
                         int additional_count, int *additional_displacements,
                         int *additional_counts, char *output,
                         const char *hostnames) {
  char filepath[256] = {0};
  char datetime[128] = {0};
  time_t t = time(NULL);
  struct tm *tm_ptr = localtime(&t);

  snprintf(datetime, sizeof(datetime), "%d%d%d_%d%d", tm_ptr->tm_year + 1900,
           tm_ptr->tm_mon + 1, tm_ptr->tm_mday, tm_ptr->tm_hour,
           tm_ptr->tm_min);

  char dirname[256] = {0};
#if 0
  snprintf(dirname, sizeof(dirname), "%s_%s", output, datetime);
#else
  snprintf(dirname, sizeof(dirname), "%s", output);
#endif

  struct stat st = {0};
  if (stat(dirname, &st) != -1) {
    char temp_dirname[256] = {0};
    snprintf(temp_dirname, sizeof(temp_dirname), "%s_%s_XXXXXX", output,
             datetime);

    char *tmpdir = mkdtemp(temp_dirname);
    assert(tmpdir != NULL);

    fprintf(stderr,
            "Unable to save results to dir: '%s', directory/file already "
            "exists, saving to '%s' instead\n",
            dirname, temp_dirname);

    memcpy(dirname, temp_dirname, sizeof(dirname));
  }
  int err = mkdir(dirname, 0700);
  if (err > 0)
    return err;

  snprintf(filepath, sizeof(filepath), "%s/metadata", dirname);
  FILE *metadata_file = fopen(filepath, "w");
  assert(metadata_file != NULL);

  snprintf(filepath, sizeof(filepath), "%s/results", dirname);
  fprintf(stderr, "Writing to file: '%s'\n", filepath);
  FILE *file = fopen(filepath, "w");
  assert(file != NULL);
  fprintf(file, "VERSION 2\n");

  for (int j = 0; j < size; j++) {
    for (int i = 0; i < j; i++) {

      fprintf(metadata_file, "%d %s %d %s %s\n", i, RANK_HOSTNAME(i), j,
              RANK_HOSTNAME(j), filepath);
      fprintf(file, "START FROM %d to %d\n", i, j);
      if (config.verbose)
        fprintf(stdout, "START FROM %d to %d\n", i, j);

      // Uncomment to include header:
      // fprintf(file, "%d -> %d (%d):\n", i, j, config.num_datapoints);
      // fprintf(file, "msgsize:   L:%6s o_r: %6s o_s: %6s g: %6s\n", "", "",
      // "", "", "");
      Logp_result *datapoints = &all_results[i * size * config.num_datapoints +
                                             j * config.num_datapoints];

      const int adisp = additional_displacements[i];
      const int acount = additional_counts[i];
      const Logp_additional_result *additional = &additional_results[adisp];
      int k = 0, q = 0; // k is index into datapoints, q is into additional
      while (q < acount || k < config.num_datapoints) {
        Logp_result res;

        bool a = false;
        if (q < acount) {
          int next_index = q;
          const Logp_additional_result *next =
              find_next_in_additional(additional, acount, q, &next_index, i, j);

          if (!next) {
            q = next_index;
            continue;
          } else if (k < config.num_datapoints &&
                     datapoints[k].message_size < next->result.message_size) {
            res = datapoints[k];
            k++;
          } else {
            res = next->result;
            a = true;
            q = next_index;
          }
        } else if (k < config.num_datapoints) {
          res = datapoints[k];
          k++;
        } else {
          break;
        }
        fprintf(file, "%7d %g %g %g %g %g %s\n", res.message_size, res.L, res.rtt, res.o_s,
                res.o_r, res.g, a ? "1" : "0");
        if (config.verbose)
          fprintf(stdout, "%7d %g %g %g %g %g %s\n", res.message_size, res.L, res.rtt,
                  res.o_s, res.o_r, res.g, a ? "1" : "0");
      }
    }
  }
  fclose(file);
  fclose(metadata_file);

  if (config.fast) {
    snprintf(filepath, sizeof(filepath), "%s/fast_metadata", dirname);
    FILE *fast_metadata_file = fopen(filepath, "w");
    assert(fast_metadata_file != NULL);
    fprintf(stderr, "Writing fast_metadata to file: '%s'\n", filepath);

    for (int i = 0; i < size; i++) {
      fprintf(fast_metadata_file, "%d selected %d\n", i, config.selected[i]);
    }

    fclose(fast_metadata_file);
  }

  return 0;
}

static int save_results(Logp_result *all_results, char *output,
                        const char *hostnames) {

  char filepath[256] = {0};

  char datetime[128] = {0};
  time_t t = time(NULL);
  struct tm *tm_ptr = localtime(&t);

  snprintf(datetime, sizeof(datetime), "%d%d%d_%d%d", tm_ptr->tm_year + 1900,
           tm_ptr->tm_mon + 1, tm_ptr->tm_mday, tm_ptr->tm_hour,
           tm_ptr->tm_min);

  char dirname[256] = {0};
#if 0
  snprintf(dirname, sizeof(dirname), "%s_%s", output, datetime);
#else
  snprintf(dirname, sizeof(dirname), "%s", output);
#endif

  struct stat st = {0};
  if (stat(dirname, &st) != -1) {
    char temp_dirname[256] = {0};
    snprintf(temp_dirname, sizeof(temp_dirname), "%s_%s_XXXXXX", output,
             datetime);

    char *tmpdir = mkdtemp(temp_dirname);
    assert(tmpdir != NULL);

    fprintf(stderr,
            "Unable to save results to dir: '%s', directory/file already "
            "exists, saving to '%s' instead\n",
            dirname, temp_dirname);

    memcpy(dirname, temp_dirname, sizeof(dirname));
  }
  int err = mkdir(dirname, 0700);
  if (err > 0)
    return err;

  snprintf(filepath, sizeof(filepath), "%s/metadata", dirname);
  FILE *metadata_file = fopen(filepath, "w");
  assert(metadata_file != NULL);

  for (int j = 0; j < size; j++) {
    for (int i = 0; i < j; i++) {
      if (i == j)
        continue;

      snprintf(filepath, sizeof(filepath), "%s/%d_%d.txt", dirname, i, j);
      fprintf(stderr, "Writing to file: '%s'\n", filepath);

      fprintf(metadata_file, "%d %s %d %s %s\n", i, RANK_HOSTNAME(i), j,
              RANK_HOSTNAME(j), filepath);

      FILE *file = fopen(filepath, "w");


      // Uncomment to include header:
      // fprintf(file, "%d -> %d (%d):\n", i, j, config.num_datapoints);
      // fprintf(file, "msgsize:   L:%6s o_r: %6s o_s: %6s g: %6s\n", "", "",
      // "", "", "");
      Logp_result *datapoints = &all_results[i * size * config.num_datapoints +
                                             j * config.num_datapoints];
      for (int k = 0; k < config.num_datapoints; k++) {
        Logp_result res = datapoints[k];
        fprintf(file, "%7d %g %g %g %g %g\n", res.message_size, res.L, res.rtt, res.o_s,
                res.o_r, res.g);
      }
      fclose(file);
    }
  }

  fclose(metadata_file);

  if (config.fast) {
    snprintf(filepath, sizeof(filepath), "%s/fast_metadata", dirname);
    FILE *fast_metadata_file = fopen(filepath, "w");
    assert(fast_metadata_file != NULL);
    fprintf(stderr, "Writing fast_metadata to file: '%s'\n", filepath);

    for (int i = 0; i < size; i++) {
      fprintf(fast_metadata_file, "%d selected %d\n", i, config.selected[i]);
    }

    fclose(fast_metadata_file);
  }

  return 0;
}

char *gather_hostnames(int root_rank) {
  char hostname[HOSTNAME_SIZE] = {0};
  gethostname(hostname, sizeof(hostname) - 1);

  MPI_Datatype hostname_type;
  MPI_Type_contiguous(sizeof(hostname), MPI_CHAR, &hostname_type);
  MPI_Type_commit(&hostname_type);

  char *hostnames = NULL;
  if (rank == root_rank)
    hostnames = calloc(sizeof(hostname), size);

  MPI_Gather(hostname, 1, hostname_type, hostnames, 1, hostname_type, root_rank,
             MPI_COMM_WORLD);

  MPI_Type_free(&hostname_type);

  // if (rank == root_rank)
  //   for (int i = 0; i < size; i++)
  //     fprintf(stderr, "hostname: %d = %s\n", i, HOSTNAME(hostnames, i));

  return hostnames;
}

void test_finalize(char *output) {
  MPI_Barrier(MPI_COMM_WORLD);

  logp_end_time = MPI_Wtime();

  if (rank == 0) {
    printf("Test finalized...\n");
    printf("Logp took %g seconds\n", logp_end_time - logp_start_time);
  }

  // Logp_result as MPI struct type
  int blocklengths[6] = {1, 1, 1, 1, 1, 1};
  MPI_Datatype types[6] = {MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE, MPI_DOUBLE,
                           MPI_INT};
  MPI_Aint offsets[6] = {offsetof(Logp_result, L), offsetof(Logp_result, rtt), offsetof(Logp_result, o_s),
                         offsetof(Logp_result, o_r), offsetof(Logp_result, g),
                         offsetof(Logp_result, message_size)};
  MPI_Datatype result_type;
  MPI_Type_create_struct(6, blocklengths, offsets, types, &result_type);
  MPI_Type_commit(&result_type);

  Logp_result *all_results = NULL;
  if (rank == 0)
    all_results =
        calloc(size * size * config.num_datapoints, sizeof(Logp_result));
  MPI_Gather(node_results, size * config.num_datapoints, result_type,
             all_results, size * config.num_datapoints, result_type, 0,
             MPI_COMM_WORLD);

#if USE_ADDITIONAL_RESULTS
  Logp_additional_result *additional_results = NULL;
  int *additional_displacements = NULL, *additional_counts = NULL;
  int additional_count = 0;
  collect_additional_datapoints(0, result_type, &additional_results,
                                &additional_count, &additional_displacements,
                                &additional_counts);
#endif

  MPI_Type_free(&result_type);

  char *hostnames = gather_hostnames(0);

  if (rank == 0) {
#if USE_ADDITIONAL_RESULTS
    save_results2(all_results, additional_results, additional_count,
                  additional_displacements, additional_counts, output,
                  hostnames);
    free(additional_counts);
    free(additional_displacements);
    free(additional_results);
#else
    save_results(all_results, output, hostnames);
#endif
    free(hostnames);
    free(all_results);
  }

  if (config.fast)
    cluster_leaves_finalize(config.leaves, config.leaves_count);

  free(reqs);
  free(node_results);

  free(messages[1]);
  free(messages[0]);
}
