#!/usr/bin/env python3
import argparse
import sqlite3

import subprocess

from matplotlib import pyplot as plt

import numpy as np

from sys import argv, stdin

from common import *
from broadcast import extract_bcast_lines
from load_into_sqlite import load_from_database
from predict_bcast import extract_logp_data, predict_algo_logp


from matplotlib import pyplot as plt
import seaborn as sns

sns.set_theme()

import pandas as pd


parser = argparse.ArgumentParser(prog="BCast plotter")
parser.add_argument("filename")
parser.add_argument("--title", default=None, type=str)
parser.add_argument("--show-n", default=False, action="store_true")
parser.add_argument("--no-errorbar", action="store_true")
parser.add_argument("--throughput", action="store_true")
parser.add_argument("--format", default="term")
parser.add_argument("--no-label-lines", default=False, action="store_true")
parser.add_argument("--no-interactive", default=False, action="store_true")
parser.add_argument("--outfile", type=str, default=None)
parser.add_argument("--speedup", default=False, action="store_true")
parser.add_argument("--speedup-key", default="mpi:BCastDefault")
parser.add_argument("--x-axis-title", type=str, default=None)
parser.add_argument("--y-axis-title", type=str, default=None)
parser.add_argument("--key-filter", type=str, default=None)
parser.add_argument("--database", type=str, default=None)
parser.add_argument("--database-table", type=str, default="default")

args = parser.parse_args()

key_filter = set(args.key_filter.split(",")) if args.key_filter else None

# source: https://stackoverflow.com/questions/31410043/hiding-lines-after-showing-a-pyplot-figure
class InteractiveLegend(object):
    def __init__(self, legend=None):
        if legend == None:
            legend = plt.gca().get_legend()
        self.legend = legend
        self.fig = legend.axes.figure
        self.lookup_artist, self.lookup_handle = self._build_lookups(legend)
        self._setup_connections()
        self.update()

    def _setup_connections(self):
        # for artist in self.legend.texts + self.legend.legendHandles:
        for artist in self.legend.texts + self.legend.legend_handles:
            artist.set_picker(10)  # 10 points tolerance
        self.fig.canvas.mpl_connect("pick_event", self.on_pick)
        self.fig.canvas.mpl_connect("button_press_event", self.on_click)

    def _build_lookups(self, legend):
        labels = [t.get_text() for t in legend.texts]
        # handles = legend.legendHandles
        handles = legend.legend_handles
        label2handle = dict(zip(labels, handles))
        handle2text = dict(zip(handles, legend.texts))
        lookup_artist = {}
        lookup_handle = {}
        for artist in legend.axes.get_children():
            if artist.get_label() in labels:
                handle = label2handle[artist.get_label()]
                lookup_handle[artist] = handle
                lookup_artist[handle] = artist
                lookup_artist[handle2text[handle]] = artist
        lookup_handle.update(zip(handles, handles))
        lookup_handle.update(zip(legend.texts, handles))
        return lookup_artist, lookup_handle

    def on_pick(self, event):
        handle = event.artist
        if handle in self.lookup_artist:
            artist = self.lookup_artist[handle]
            artist.set_visible(not artist.get_visible())
            self.update()

    def on_click(self, event):
        if event.button == 3:
            visible = False
        elif event.button == 2:
            visible = True
        else:
            return
        for artist in self.lookup_artist.values():
            artist.set_visible(visible)
        self.update()

    def update(self):
        for artist in self.lookup_artist.values():
            handle = self.lookup_handle[artist]
            if artist.get_visible():
                handle.set_visible(True)
            else:
                handle.set_visible(False)
        self.fig.canvas.draw()


def extract(line, keys, floats=None, ints=None):
    out = []
    parts = line.split()
    if not parts:
        return []
    out.append(parts[3])
    out.append(parts[4][1:])
    for part in parts:
        kv = part.split("=")
        if kv[0] in keys:
            v = kv[1]
            if floats and kv[0] in floats:
                v = float(v)
            if ints and kv[0] in ints:
                v = int(v)
            out.append(v)
    return out


def indexin(fields, field):
    for i, f in enumerate(fields):
        if f == field:
            return i
    raise Exception


def plot(
    datas,
    fields,
    xvar,
    yvar,
    keyfunc,
    errvar=None,
    minvar=None,
    maxvar=None,
    throughput=False,
    out_file=None,
    title=None,
    n_field=None,
):
    plt.style.use("ggplot")  # Change/Remove This If you Want

    fig, ax = plt.subplots()

    xvar_ind = indexin(fields, xvar) + 2
    yvar_ind = indexin(fields, yvar) + 2
    errvar_ind = -1
    if errvar:
        errvar_ind = indexin(fields, errvar) + 2
    maxvar_ind = -1
    if maxvar:
        maxvar_ind = indexin(fields, maxvar) + 2
    minvar_ind = -1
    if minvar:
        minvar_ind = indexin(fields, minvar) + 2

    n_ind = -1
    if n_field:
        n_ind = indexin(fields, n_field) + 2

    N = None

    mapping = {}
    for data in datas:
        if not data:
            continue

        if n_ind != -1:
            if N is None:
                N = data[n_ind]
            else:
                assert N == data[n_ind]

        key = keyfunc(data)

        if key_filter and key not in key_filter:
            continue

        xval = data[xvar_ind]
        yval = data[yvar_ind]

        if throughput:
            yval = xval / yval

        errval = None
        if errvar:
            errval = data[errvar_ind]

        if not key in mapping:
            mapping[key] = [[], [], [], [[], []]]

        mapping[key][0].append(xval)
        mapping[key][1].append(yval)

        if errvar:
            mapping[key][2].append(errval)

        if minvar and maxvar:
            minval = data[minvar_ind]
            maxval = data[maxvar_ind]
            mapping[key][3][0].append(minval)
            mapping[key][3][1].append(maxval)

    # print(mapping)

    lines = []

    # plot speedup curve of baseline compared to BEST
    if args.speedup:
        baseline_key = args.speedup_key  # "mpi-BCastDefault"
        # baseline_xs, baseline_ys = mapping[baseline_key][0:2]
        baseline_xs = mapping[baseline_key][0]
        baseline_ys = mapping[baseline_key][1]
        # print()
        # print(baseline_xs)
        # print(baseline_ys)
        speedup_curve = np.zeros(len(baseline_xs))
        for i, x in enumerate(baseline_xs):
            best_value = None
            for key, value in mapping.items():
                if key == baseline_key:
                    continue

                if not value:
                    continue
                # print('kv', value[1])
                if best_value is None or best_value > value[1][i]:
                    if len(value[1]) > i:
                        best_value = value[1][i]
                        # doesn't exist
            if best_value is None:
                speedup_curve[i] = 0
            else:
                speedup_curve[i] = baseline_ys[i] / best_value
        # print("speedup_curve", speedup_curve)
        lines.append(ax.plot(baseline_xs, speedup_curve, label="speedup"))

    for key, value in mapping.items():
        if errvar:
            # yerror = value[2]
            # if len(value[3]) == len(value[1]):
            #     yerror = value[3]
            # print("yerror", yerror, value)
            # lines.append(ax.errorbar(value[0], value[1], yerr=yerror, label=key, fmt="-o"))
            # print("mean", mean)
            # print("errvar", yerror)

            variance = value[2]

            min_max = None
            if value[3]:
                min_max = value[3]

            mean = np.array(value[1])
            stdvar = np.sqrt(np.array(variance))

            lower = mean - stdvar
            upper = mean + stdvar

            sns.set_palette(sns.color_palette("viridis"))
            lines.append(ax.plot(value[0], value[1], label=key))
            lines.append(ax.scatter(value[0], value[1], label=key))
            ax.fill_between(value[0], lower, upper, color="#888888", alpha=0.5)

            # min and max
            # if min_max:
            #     print("MINMAX")
            #     ax.fill_between(
            #         value[0], min_max[0], min_max[1], color="blue", alpha=0.01
            #     )

            # ax.fill_between(trees_grid, test_acc.mean(axis=1) - 2*test_acc.std(axis=1), test_acc.mean(axis=1) + 2*test_acc.std(axis=1), color='#888888', alpha=0.2)
        else:
            lines.append(ax.plot(value[0], value[1], label=key))

    if not args.no_label_lines:
        # DEPENDENCY: matplotlib-label-lines
        from labellines import labelLine, labelLines

        labelLines(ax.get_lines(), align=False, fontsize=14)
    else:
        plt.legend()
        if not args.no_interactive:
            leg = InteractiveLegend()
        else:
            leg = ax.legend()

    if args.x_axis_title:
        ax.set_xlabel(args.x_axis_title)
    else:
        ax.set_xlabel(xvar)
    if args.y_axis_title:
        ax.set_ylabel(args.y_axis_title)
    else:
        ax.set_ylabel(yvar)

    if title is not None:
        if args.show_n and N is not None:
            title = f"{title} N={N}"
        ax.set_title(title)

    if out_file:
        plt.savefig(out_file)
    else:
        plt.show()


lines = []
if args.filename == "-":
    lines = stdin.readlines()
else:
    with open(args.filename) as f:
        lines = f.readlines()

lines = extract_bcast_lines(lines)


# print("LINES", lines)

# parse data
fields = ["N", "msgsize", "mean", "var", "min", "max", "P", "name"]
ints = set(["N", "msgsize", "P"])
floats = set(["mean", "var", "min", "max"])
data = list(map(lambda x: extract(x, set(fields), ints=ints, floats=floats), lines))

msgsizes = set()
for datapoint in data:
    msgsizes.add(datapoint[3])
msgsizes = sorted(msgsizes)

# print(data[0])
# assert False

if args.database:
    con = sqlite3.connect(args.database)
    metadata, ranks, nodes = load_from_database(args.database_table, con)
    logp_data = extract_logp_data(metadata, ranks)
    print("logp_data", logp_data)
    subset = set()
    for i in range(0, ranks):
        subset.add(i)
    prediction = predict_algo_logp(logp_data, "BCastBtree", subset, msgsizes)
    print("msgsizes", msgsizes)
    print("prediction", prediction)

    pred_data = []
    for i in range(len(prediction)):
        pred_data.append(
            [
                "prediction",
                "BCastBTree",
                data[0][2],
                msgsizes[i],
                prediction[i],
                0,
                0,
                0,
                ranks,
                "prediction",
            ]
        )
    print("pred_data")
    data += pred_data


print(data)

if args.no_errorbar:
    plot(
        data,
        fields,
        "msgsize",
        "mean",
        lambda x: f"{x[0]}:{x[1]}",
        throughput=args.throughput,
        out_file=args.outfile,
        title=args.title,
        n_field="N",
    )
else:
    plot(
        data,
        fields,
        "msgsize",
        "mean",
        lambda x: f"{x[0]}:{x[1]}",
        errvar="var",
        minvar="min",
        maxvar="max",
        throughput=args.throughput,
        out_file=args.outfile,
        title=args.title,
        n_field="N",
    )
# print(data)
