#!/bin/python3

import os
import sys
import argparse
import subprocess
import tempfile
import shlex

from string import Template

from create_topo_file import create_topo_file
from run_job_common import sbatch, run_slurm_file, parse_args, run_make_build

SLURM_HEADER_TEMPLATE = """#!/bin/sh
#SBATCH --account=nn9589k
#SBATCH --partition=$partition
#SBATCH --time=$time
#SBATCH --nodes=$nodes
#SBATCH --ntasks-per-node=$tasks
#SBATCH --job-name="$name"
#SBATCH --output="fram/outputs/${name}_%j"
#SBATCH --mail-user="$mail_user"
#SBATCH --mail-type=ALL
#SBATCH --mem=10G
"""

SLURM_SCRIPT = """
WORKDIR=${SLURM_SUBMIT_DIR}
cd ${WORKDIR}
echo "we are running from this directory: $SLURM_SUBMIT_DIR"
echo " the name of the job is: $SLURM_JOB_NAME"
echo "Th job ID is $SLURM_JOB_ID"
echo "The job was run on these nodes: $SLURM_JOB_NODELIST"
echo "Number of nodes: $SLURM_JOB_NUM_NODES"
echo "We are using $SLURM_CPUS_ON_NODE cores"
echo "We are using $SLURM_CPUS_ON_NODE cores per node"
echo "Total of $SLURM_NTASKS cores"
"""

SLURM_SCRIPT_END = """
echo "== Job finished at $(date) =="
"""


def build_slurm_file(variables, script):
    templ = Template(SLURM_HEADER_TEMPLATE)
    header = templ.safe_substitute(variables)
    slurm_file_contents = (
        header + SLURM_SCRIPT + "\n" + script + "\n" + SLURM_SCRIPT_END
    )
    return slurm_file_contents


def tail_output(name, job_id):
    path = f"fram/outputs/{name}_{job_id}"
    subprocess.run(["tail", "--retry", "-f", path])


def wait_for_complete(job_id):
    while True:
        time.sleep(10)
        result = subprocess.run(["sbatch", slurm_file], capture_output=True)
       


args, passed_args = parse_args()

SCRIPT_TEMPLATE = """
# echo '==== START OF machine topology ==='
# module purge
# module load OpenMPI/4.1.4-GCC-11.3.0
# module list
# hostname -A
# lstopo-no-graphics --no-io --no-legend --of txt
# echo '==== END OF machine topology ==='

module purge
module load ${mpi_module}
module list

# make clean
# (cd lib && make clean && make)
# make

echo "Running:" ${mpi_runner} ./lgate -f "${topo_file}" ${passed_args}
time ${mpi_runner} ./lgate -f "${topo_file}" -W $W -N $N ${passed_args}
"""


def select_mpi_module(args):
    if args.use_openmpi:
        return "OpenMPI/4.1.4-GCC-11.3.0"
    return "intel/2020b"


def build_script(args, args_dict):
    group_size = args.group_size
    if group_size == -1:
        group_size = args.tasks

    if args.topo_file:
        topo_file = args.topo_file
    else:
        topo_filename = (
            f"fram/topos/{args.name}_{group_size}_{args.nodes}_{args.tasks}.topo"
        )
        topo_file = create_topo_file(
            args.name, group_size, args.nodes, args.tasks, topo_filename
        )

    mpi_runner = "srun"
    mpi_module = select_mpi_module(args)
    if args.use_openmpi:
        mpi_runner = "mpirun"

    msgsizes = "1,2,4,8,32,64,128,256,512,1024,2048,4096,8192,16384,32768,65536,131072,262144,362144,462144,524288,1048576,2097152"

    script = str(
        Template(SCRIPT_TEMPLATE).safe_substitute(
            mpi_module=mpi_module,
            mpi_runner=mpi_runner,
            topo_file=topo_file,
            N=args_dict["N"],
            W=args_dict["W"],
            MSGSIZES=msgsizes,
            passed_args=passed_args,
        )
    )

    # delete params that are only used here
    del args_dict["W"]
    del args_dict["N"]
    del args_dict["use_openmpi"]
    del args_dict["topo_file"]
    return script


mpi_module = select_mpi_module(args)

args_dict = args.__dict__
script = build_script(args, args_dict)
slurm_file_contents = build_slurm_file(args_dict, script)
print(slurm_file_contents)


if args.build or args.run:
    print("[===] Building")
    run_make_build(mpi_module)

if args.run:
    print("[===] Running")
    run_slurm_file(args.name, slurm_file_contents, no_watch=args.no_watch)
