#!/bin/python3

import os
import sys
import argparse
import subprocess
import tempfile
import shlex

from string import Template

from create_topo_file import create_topo_file

SLURM_HEADER_TEMPLATE = """#!/bin/sh
#SBATCH --partition=$partition
#SBATCH --account=ie-idi
#SBATCH --time=$time
#SBATCH --nodes=$nodes
#SBATCH --nodelist=$nodelist
#SBATCH --ntasks-per-node=$tasks
#SBATCH --mem=$memory
#SBATCH --job-name="$name"
#SBATCH --output="idun/outputs/${name}_%j"
#SBATCH --mail-user="$mail_user"
#SBATCH --mail-type=ALL
"""

SLURM_SCRIPT = """
WORKDIR=${SLURM_SUBMIT_DIR}
cd ${WORKDIR}
echo "we are running from this directory: $SLURM_SUBMIT_DIR"
echo " the name of the job is: $SLURM_JOB_NAME"
echo "Th job ID is $SLURM_JOB_ID"
echo "The job was run on these nodes: $SLURM_JOB_NODELIST"
echo "Number of nodes: $SLURM_JOB_NUM_NODES"
echo "We are using $SLURM_CPUS_ON_NODE cores"
echo "We are using $SLURM_CPUS_ON_NODE cores per node"
echo "Total of $SLURM_NTASKS cores"
"""

SLURM_SCRIPT_END = """
echo "== Job finished at $(date) =="
"""

def build_slurm_file(variables, script):
    templ = Template(SLURM_HEADER_TEMPLATE)
    header = templ.safe_substitute(variables)
    slurm_file_contents = header + SLURM_SCRIPT + "\n" + script + "\n" + SLURM_SCRIPT_END
    return slurm_file_contents

def tail_output(name, job_id):
    path = f"idun/outputs/{name}_{job_id}"
    subprocess.run(["tail", "--retry", "-f", path])

def sbatch(name, slurm_file, watch_output=True):
    print(f'Running: $ sbatch {slurm_file}')
    result = subprocess.run(["sbatch", slurm_file], capture_output=True)
    print('result', result)
    # Submitted batch job 17489530
    output = result.stdout.decode()
    splits = output.split()

    err_output = result.stdout.decode()
    print('OUTPUT', output, err_output)
    if splits[0] == "Submitted" and splits[1] == "batch" and splits[2] == "job":
        job_id = splits[3]
        print('ID', job_id)
        if watch_output:
            tail_output(name, job_id)
    return False


def run_slurm_file(name, slurm_file_content, watch_output=True):
    with tempfile.NamedTemporaryFile() as fp:
        encoded = slurm_file_content.encode("utf-8")
        fp.write(encoded)
        fp.flush()
        sbatch(name, fp.name, watch_output=watch_output)


parser = argparse.ArgumentParser(prog="slurm helper")
# parser.add_argument('working_dir')
parser.add_argument('--run', default=False, action="store_true")
parser.add_argument('--partition', type=str, default="CPUQ")
parser.add_argument('--nodes', type=int, default=1)
parser.add_argument('--tasks', type=int, default=1)
parser.add_argument('--memory', type=int, default=12000)
parser.add_argument('--nodelist', type=str, default="idun-03-06")
parser.add_argument('--name', type=str, default="default")
parser.add_argument('--time', type=str, default="00:00:10")
parser.add_argument('--group-size', type=int, default=-1)
parser.add_argument('--mail-user', type=str, default="jorgerei@stud.ntnu.no")
parser.add_argument('--topo-file', type=str)
parser.add_argument('--use-openmpi', default=False, action="store_true")
parser.add_argument('--no-watch-output', default=False, action="store_true")
parser.add_argument('-N', default=1000, type=int)
parser.add_argument('-W', default=5, type=int)
# parser.add_argument('--mpi-module', type=str, default="intel/2020b")

# use :: as delimiter, arguments after will be sent to script
last_arg = len(sys.argv)
for i, arg in enumerate(sys.argv):
    if arg == "::":
        last_arg = i
        break
parser_args = sys.argv[1:last_arg]
passed_args = shlex.join(sys.argv[last_arg+1:])
print('parser_args', parser_args)
print('passed_args', passed_args)
# print('passed_args', passed_args)
# print('passed_args JOINED:', shlex.join(passed_args))
args = parser.parse_args(parser_args)

SCRIPT_TEMPLATE = """
module purge
module load ${mpi_module}
module list

# Build
make clean
(cd lib && make clean && make)
make

module purge
module load ${mpi_module}
module list

${mpi_runner} ./lgate ${passed_args}
"""

def build_script(args, args_dict):
    group_size = args.group_size
    if group_size == -1:
        group_size = args.tasks

    if args.topo_file:
        topo_file = args.topo_file
    else:
        topo_filename = f"idun/topos/{args.name}_{group_size}_{args.nodes}_{args.tasks}.topo"
        topo_file = create_topo_file(args.name, group_size, args.nodes, args.tasks, topo_filename)

    mpi_module = "intel/2020b"
    mpi_runner = "srun"
    if args.use_openmpi:
        mpi_module = "OpenMPI/4.1.4-GCC-11.3.0"
        mpi_runner = "mpirun"

    msgsizes = "1,2,4,8,32,64,128,256,512,1024,2048,4096,8192,16384,32768,65536,131072,262144,362144,462144,524288,1048576,2097152"
    # x = 100000
    # while x <= 600000:
    #     msgsizes += f"{x},"
    #     x += 100000
    # print('msgsizes', msgsizes)

    script = str(Template(SCRIPT_TEMPLATE).safe_substitute(mpi_module=mpi_module, mpi_runner=mpi_runner, topo_file=topo_file, N=args_dict["N"], W=args_dict["W"], MSGSIZES=msgsizes, passed_args=passed_args))

    # delete params that are only used here 
    del args_dict["W"]
    del args_dict["N"]
    del args_dict["use_openmpi"]
    del args_dict["topo_file"]
    return script

args_dict = args.__dict__
print('args', args_dict)
script = build_script(args, args_dict)
slurm_file_contents = build_slurm_file(args_dict, script)
print(slurm_file_contents)

if args.run:
    print('Running')
    run_slurm_file(args.name, slurm_file_contents, watch_output=False)


