from typing import Optional, List, Tuple, Dict
import os
import sys
import argparse
import subprocess
import csv


from dataclasses import dataclass

from run_many_logp_common import *
from common import read_meta_file

parser = argparse.ArgumentParser(prog="slurm helper")
parser.add_argument("machine")
parser.add_argument("--output", default="output_data", type=str)
parser.add_argument("--msgsize", default=0, type=int)
parser.add_argument("--metric", default="rtt", type=str)
parser.add_argument("--plot", default=False, action="store_true")
parser.add_argument("--param-search", default=False, action="store_true")
parser.add_argument("--table", default=False, action="store_true")
parser.add_argument("--no-run", default=False, action="store_true")
parser.add_argument("--figures-output", default="figures_out", type=str)

args = parser.parse_args()

def bool_to_str(v: bool) -> str:
    if v:
        return "true"
    return "false"


def create_logp_ini(
    identifier: str, reps: int, max_message_size=262144, output_dir=".", additional_measurements=True
):
    contents = f"""verbose = false
repetitions = {reps}
epsilon = 0.01
rtt_factor = 1.2

max_message_size  = {max_message_size}
additional_measurements = {bool_to_str(additional_measurements)}
g_saturation = true
g_saturation_limit = 65536
"""

    path = f"{output_dir}/{identifier}_logp.ini"
    with open(path, "w") as f:
        f.write(contents)
    return path


def run(run_spec: Run, base_dir: str, no_run=False):
    # if machine.name == "arm1":
    #    subprocess.run(["/bin/python3", "run_arm1_job.py"])
    try:
        run_spec.create_base_dir(base_dir)
    except FileExistsError as e:
        pass

    output_logp_dir = run_spec.output_dir(base_dir)
    if os.path.isdir(output_logp_dir):
        print("Skipping run", run_spec, "file already exists")
        return RunResult(
            identifier=run_spec.identifier(),
            output_dir=output_logp_dir,
            exists=True,
            success=False,
            stdout_path=None,
            logp_ini_path=None,
        )

    output_dir = run_spec.output_base_dir(base_dir)
    logp_ini_path = create_logp_ini(
        run_spec.identifier(),
        run_spec.reps,
        run_spec.max_message_size,
        output_dir=output_dir,
	additional_measurements=run_spec.additional_measurements
    )

    topo = run_spec.machine.topology(run_spec.nodes)
    print("topology", topo)

    topo_file = f"{output_dir}/{run_spec.nodes}_{run_spec.machine.cores}.topo"
    with open(topo_file, "w") as f:
        f.write(topo)

    stdout_path = f"{output_dir}/{run_spec.nodes}_{run_spec.machine.cores}_{run_spec.reps}_{run_spec.max_message_size}.out"
    if not no_run:
        if run_spec.machine.slurm:
            result = run_spec.machine.run(
                run_spec, base_dir, topo=topo_file, logp_ini_path=logp_ini_path
            )
        else:
            with open(stdout_path, "w") as f:
                result = run_spec.machine.run(
                    run_spec,
                    base_dir,
                    topo=topo_file,
                    stdout=f,
                    logp_ini_path=logp_ini_path,
                )
    else:
        result = None

    return RunResult(
        identifier=run_spec.identifier(),
        output_dir=run_spec.output_dir(base_dir),
        stdout_path=stdout_path,
        exists=False,
        success=True,
        logp_ini_path=logp_ini_path,
    )


RUNS = [
    # HPC PC
    Run(
        machine=HPCLAB,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=8,
        max_message_size=262144,
    ),
    Run(
        machine=HPCLAB,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=32,
        max_message_size=262144,
    ),
    Run(
        machine=HPCLAB,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=64,
        max_message_size=262144,
    ),
    Run(
        machine=HPCLAB,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
    ),
    Run(
        machine=HPCLAB,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=1024,
        max_message_size=262144,
    ),
    Run(
        machine=HPCLAB,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=8 * 1024,
        max_message_size=262144,
    ),
    # TARO PC
    # Run(
    #     machine=TARO,
    #     expected_clusters=2,
    #     timeout="00:10:00",
    #     nodes=1,
    #     reps=8,
    #     max_message_size=262144,
    # ),
    # Run(
    #     machine=TARO,
    #     expected_clusters=2,
    #     timeout="00:10:00",
    #     nodes=1,
    #     reps=32,
    #     max_message_size=262144,
    # ),
    # Run(
    #     machine=TARO,
    #     expected_clusters=2,
    #     timeout="00:10:00",
    #     nodes=1,
    #     reps=64,
    #     max_message_size=262144,
    # ),
    # Run(
    #     machine=TARO,
    #     expected_clusters=2,
    #     timeout="00:10:00",
    #     nodes=1,
    #     reps=128,
    #     max_message_size=262144,
    # ),
    # Run(
    #     machine=TARO,
    #     expected_clusters=2,
    #     timeout="00:10:00",
    #     nodes=1,
    #     reps=1024,
    #     max_message_size=262144,
    # ),
    # Run(
    #     machine=TARO,
    #     expected_clusters=2,
    #     timeout="00:10:00",
    #     nodes=1,
    #     reps=8 * 1024,
    #     max_message_size=262144,
    # ),
    # ARM1 PC
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=8,
        max_message_size=262144,
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=32,
        max_message_size=262144,
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=64,
        max_message_size=262144,
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=1024,
        max_message_size=262144,
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=8 * 1024,
        max_message_size=262144,
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="core"
    ),
    Run(
        machine=ARM1,
        expected_clusters=2,
        timeout="00:10:00",
        nodes=1,
        reps=256,
        max_message_size=262144,
        tag="core"
    ),
    # FRAM
    Run(machine=FRAM, timeout="00:02:00", nodes=1, reps=8, max_message_size=262144),
    Run(machine=FRAM, timeout="00:02:00", nodes=1, reps=32, max_message_size=262144),
    Run(machine=FRAM, timeout="00:02:00", nodes=1, reps=64, max_message_size=262144),
    Run(machine=FRAM, timeout="00:02:00", nodes=1, reps=128, max_message_size=262144),
    Run(machine=FRAM, timeout="00:05:00", nodes=2, reps=8, max_message_size=262144),
    Run(machine=FRAM, timeout="00:05:00", nodes=2, reps=32, max_message_size=262144),
    Run(machine=FRAM, timeout="00:05:00", nodes=2, reps=64, max_message_size=262144),
    Run(machine=FRAM, timeout="00:05:00", nodes=2, reps=128, max_message_size=262144),
    Run(machine=FRAM, timeout="00:10:00", nodes=4, reps=8, max_message_size=262144),
    Run(machine=FRAM, timeout="00:10:00", nodes=4, reps=32, max_message_size=262144),
    Run(machine=FRAM, timeout="00:10:00", nodes=4, reps=64, max_message_size=262144),
    Run(machine=FRAM, timeout="00:10:00", nodes=4, reps=128, max_message_size=262144),
    ## Run(machine=FRAM, timeout="00:10:00", nodes=1, reps=128, max_message_size=262144, tag="block_cyclic", params={"numa_order": "scatter"}),
    ## Run(machine=FRAM, timeout="00:10:00", nodes=1, reps=256, max_message_size=262144, tag="block_cyclic", params={"numa_order": "scatter"}),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="reversed_order",
    ),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="randomized_order",
    ),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="mpi_barrier",
    ),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="block_cyclic",
    ),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="mpi_barrier_normal_order",
    ),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="mpi_barrier_normal_order2",
    ),
    Run(
        machine=FRAM,
        timeout="00:10:00",
        nodes=1,
        reps=128,
        max_message_size=262144,
        tag="reversed_mpi_barrier_",
    ),
    Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=8, max_message_size=262144),
    Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=32, max_message_size=262144),
    Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=64, max_message_size=262144),
    # FRAM singlecore
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=1,
        reps=8,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=1,
        reps=32,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=1,
        reps=64,
        max_message_size=262144,
    ),
    # singlecore 4
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=4,
        reps=8,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=4,
        reps=32,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=4,
        reps=64,
        max_message_size=262144,
    ),
    # singlecore 16
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=16,
        reps=8,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=16,
        reps=32,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=16,
        reps=64,
        max_message_size=262144,
    ),
    # singlecore 32
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=32,
        reps=8,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=32,
        reps=32,
        max_message_size=262144,
    ),
    Run(
        machine=FRAM_SINGLECORE,
        timeout="00:10:00",
        nodes=32,
        reps=64,
        max_message_size=262144,
    ),
    # IDUN
    Run(machine=IDUN, timeout="00:05:00", nodes=1, reps=8, max_message_size=262144),
    Run(machine=IDUN, timeout="00:05:00", nodes=1, reps=32, max_message_size=262144),
    Run(machine=IDUN, timeout="00:05:00", nodes=1, reps=64, max_message_size=262144),
    Run(machine=IDUN, timeout="00:05:00", nodes=1, reps=128, max_message_size=262144),

    # BETZY
    # Run(machine=BETZY, timeout="00:02:00", nodes=4, reps=8, max_message_size=2, additional_measurements=False),
    Run(machine=BETZY16, timeout="00:10:00", nodes=4, reps=8, max_message_size=262144),
    # Run(machine=BETZY16, timeout="00:05:00", nodes=4, reps=8, max_message_size=262144),
    ## Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=8, max_message_size=262144),
    ## Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=32, max_message_size=262144),
    ## Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=64, max_message_size=262144),
    # Run(machine=FRAM, timeout="00:10:00", nodes=8, reps=128, max_message_size=262144),
    #
    Run(
        machine=BETZY,
        timeout="00:10:00",
        nodes=4,
        reps=8,
        max_message_size=4096,
        tag="test",
    ),
]


def create_clustering_method_list(
    machine: MachineNode,
    res: RunResult,
    ranks,
    nodes,
) -> List[ClusteringMethod]:
    print("MACHINES", nodes, "ranks", ranks)
    machines = extract_unique_machines(nodes)
    smallest_machine = min([len(group) for group in nodes])
    print("MACHINES", machines)
    n_machines = len(machines)
    n_nodes = len(nodes)
    n_sockets = len(nodes) * machine.sockets
    n_ranks = n_nodes * machine.cores
    out = [
        ClusteringMethod(
            name="k_medoids",
            tag="machine",
            distmatrix=True,
            n_clusters=n_machines,
        ),
        ClusteringMethod(
            name="k_medoids",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
        ),
        ClusteringMethod(
            name="k_medoids",
            tag="socket",
            distmatrix=True,
            n_clusters=n_sockets,
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.4, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.4, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.35, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.3, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.25, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.19, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="ssstree",
            tag="socket",
            distmatrix=True,
            parameters={"alpha": 0.23, "eigen_d": 0},
        ),
        ClusteringMethod(
            name="hdbscan",
            tag="socket",
            distmatrix=True,
            parameters={"min_cluster_size": 2, "min_samples": 1},
        ),
        ClusteringMethod(
            name="hdbscan",
            tag="socket",
            distmatrix=True,
            parameters={
                "min_cluster_size": machine.cores // machine.sockets,
                "min_samples": machine.cores // machine.sockets,
            },
        ),
        ClusteringMethod(
            name="hdbscan",
            tag="socket",
            distmatrix=True,
            parameters={
                "min_cluster_size": machine.cores // machine.sockets,
                "min_samples": 2,
            },
        ),
        ClusteringMethod(
            name="hdbscan",
            tag="node",
            distmatrix=True,
            parameters={
                "min_cluster_size": machine.cores - 1,
                "min_samples": machine.cores - 1,
            },
        ),
        ClusteringMethod(
            name="hdbscan",
            tag="machine",
            distmatrix=True,
            parameters={
                "min_cluster_size": smallest_machine,
                "min_samples": smallest_machine,
            },
        ),
        ClusteringMethod(
            name="dbscan",
            tag="node",
            distmatrix=True,
            parameters={"eps": 0.01},
        ),
        ClusteringMethod(
            name="dbscan",
            tag="socket",
            distmatrix=True,
            parameters={"eps": 0.1},
        ),
        ClusteringMethod(
            name="agglomative",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={},
        ),
        ClusteringMethod(
            name="agglomative",
            tag="socket",
            distmatrix=True,
            n_clusters=n_sockets,
            parameters={},
        ),
        ClusteringMethod(
            name="agglomative",
            tag="machine",
            distmatrix=True,
            n_clusters=n_machines,
            parameters={},
        ),
        ClusteringMethod(
            name="spectral",
            tag="socket",
            distmatrix=True,
            n_clusters=n_sockets,
            parameters={},
        ),
        ClusteringMethod(
            name="spectral",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={},
        ),
        ClusteringMethod(
            name="connected_components",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={"threshold": 0.01},
        ),
        ClusteringMethod(
            name="connected_components",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={"threshold": 0.0001},
        ),
        ClusteringMethod(
            name="connected_components",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={"threshold": 0.00000001},
        ),
        ClusteringMethod(
            name="connected_components",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={"threshold": 0.000000001},
        ),
        ClusteringMethod(
            name="connected_components",
            tag="node",
            distmatrix=True,
            n_clusters=n_nodes,
            parameters={"threshold": 0.000001},
        ),
        # ClusteringMethod(
        #     name="affinity_propagation",
        #     tag="node",
        #     n_clusters=n_nodes,
        #     distmatrix=True,
        #     parameters={},
        # ),
    ]
    return out


def _join_parameters(parameters: Dict[str, any]) -> str:
    return ",".join(map(lambda it: f"{it[0]}={str(it[1])}", parameters.items()))


if args.plot:
    from run_many_logp_plot import (
        plot_result,
        plot_result_line,
        plot_machine_results,
        plot_machine_results_with_clusters,
        plot_machine_results_with_clusters_ssstree,
    )

if args.table:
    from run_many_logp_table import (
        create_clustering_results_table,
        create_clustering_results,
        create_clustering_results_table_leveled,
    )

if args.param_search:
    from run_many_logp_table import (
        table_param_search,
    )


def get_existing_runs(machine=None):
    results = []
    for run_spec in RUNS:
        if machine and run_spec.machine.name != machine:
            continue
        res = run(run_spec, args.output, args.no_run)
        if res.success or res.exists:
            results.append((run_spec, res))
    return results


def find_machine(machine_name: str):
    if machine_name in MACHINES:
        machine = MACHINES[machine_name]
        assert machine.name == machine_name
        return machine
    assert False, "Invalid machine"


def main_run():
    print("Writing data to", args.output)
    results = []
    machine = None
    for run_spec in RUNS:
        if args.machine == run_spec.machine.name:
            if not machine:
                machine = run_spec.machine
            print("Running", run_spec)

            res = run(run_spec, args.output, args.no_run)
            print(res)
            if res.success or res.exists:
                results.append((run_spec, res))
            if (res.success or res.exists) and args.plot:
                pass
                # print("Plotting")
                plot_result(res, args.msgsize, args.figures_output, args.metric)
                # plot_result_line(res, args.figures_output)
                #

    if args.plot:
        print("Plotting all")
        # plot_machine_results_with_clusters(machine, results, args.msgsize, args.figures_output)
        plot_machine_results_with_clusters_ssstree(
            machine, results, args.msgsize, args.figures_output
        )

        plot_machine_results(
            machine, results, args.msgsize, args.figures_output, args.metric
        )


def average_accuracies(accuracies):
    sum_machine = 0.0
    sum_node = 0.0
    sum_sockets = 0.0
    for acc in accuracies:
        sum_machine += acc[0]
        sum_node += acc[1]
        sum_sockets += acc[2]

    n = len(accuracies)
    return sum_machine / n, sum_node / n, sum_sockets / n


main_run()

if args.param_search:
    results = get_existing_runs(machine=args.machine)
    machine = find_machine(args.machine)
    param_search_results = table_param_search(
        machine, results, args.msgsize, args.figures_output
    )
    for param_search_result in param_search_results:
        print("RESULT", param_search_result)
    # print("PARAM_SEARCH_REULST", param_search_results)


# TODO: remove this
if args.table:
    from create_logp_distance_figures import predict_rtt

    results = get_existing_runs(machine=args.machine)
    all_results = []
    best_results = (
        {}
    )  # R -> method -> [(machine_accuracy,node_accuracy,socket_accuracy)]

    for (run, result) in results:
        print("Clustering for result", run, result)

        directory = result.output_dir
        metadata, ranks, nodes = read_meta_file(directory)

        A = predict_rtt(metadata, ranks, args.msgsize)

        methods = create_clustering_method_list(run.machine, result, ranks, nodes)

        res_results = create_clustering_results(
            run.machine,
            run,
            result,
            methods,
            metadata,
            ranks,
            nodes,
            A,
            args.msgsize,
        )

        for res in res_results:
            # print("RES", res)
            if res.run.reps not in best_results:
                best_results[res.run.reps] = {}

            if res.method.name not in best_results[res.run.reps]:
                best_results[res.run.reps][res.method.name] = []

            best_results[res.run.reps][res.method.name].append(res)

        all_results += res_results

    with open("results.txt", "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        header = [
            "system",
            "cluster_name",
            "cluster_tag",
            "R",
            "ranks",
            "nodes",
            "machines",
            "clusters",
            "accuracy_machine",
            "accuracy_node",
            "accuracy_socket",
            "parameters",
        ]
        writer.writerow(header)
        for res in all_results:
            writer.writerow(
                [
                    res.machine.name,
                    res.method.name,
                    res.method.tag,
                    res.run.reps,
                    res.n_ranks,
                    res.n_nodes,
                    res.n_machines,
                    res.n_clusters,
                    res.accuracy.machine,
                    res.accuracy.node,
                    res.accuracy.socket,
                    _join_parameters(res.method.parameters),
                ]
            )

    best_accuracies = []
    grouped_accuracies = {}
    for R, method_results in best_results.items():
        for method, method_results in method_results.items():

            sum_accuracies_per_p = {}
            max_accuracies_per_p = {}
            p_counts = {}

            for res in method_results:
                if not res.n_ranks in sum_accuracies_per_p:
                    sum_accuracies_per_p[res.n_ranks] = [0.0, 0.0, 0.0]
                if not res.n_ranks in max_accuracies_per_p:
                    max_accuracies_per_p[res.n_ranks] = [0.0, 0.0, 0.0]

                if res.n_ranks not in p_counts:
                    p_counts[res.n_ranks] = 0
                p_counts[res.n_ranks] += 1
                sum_accuracies_per_p[res.n_ranks][0] += res.accuracy.machine
                sum_accuracies_per_p[res.n_ranks][1] += res.accuracy.node
                sum_accuracies_per_p[res.n_ranks][2] += res.accuracy.socket

                max_accuracies_per_p[res.n_ranks][0] = max(
                    max_accuracies_per_p[res.n_ranks][0], res.accuracy.machine
                )
                max_accuracies_per_p[res.n_ranks][1] = max(
                    max_accuracies_per_p[res.n_ranks][1], res.accuracy.node
                )
                max_accuracies_per_p[res.n_ranks][2] = max(
                    max_accuracies_per_p[res.n_ranks][2], res.accuracy.socket
                )

            for n_ranks, vals in max_accuracies_per_p.items():
                # avg_machine = vals[0] / p_counts[n_ranks]
                # avg_node = vals[1] / p_counts[n_ranks]
                # avg_socket = vals[2] / p_counts[n_ranks]

                max_machine = vals[0]
                max_node = vals[1]
                max_socket = vals[2]
                best_accuracies.append(
                    (
                        method,
                        n_ranks,
                        R,
                        round(max_machine, 3),
                        round(max_node, 3),
                        round(max_socket, 3),
                    )
                )

                key = f"{method}_{n_ranks}"
                if not key in grouped_accuracies:
                    grouped_accuracies[key] = []
                grouped_accuracies[key].append(
                    (round(max_machine, 3), round(max_node, 3), round(max_socket, 3))
                )
            # best_accuracies[R]
    # print("grouped_accuracies", grouped_accuracies)

    import latextable
    from texttable import Texttable
    from clustering import STATIC_METHODS

    table = Texttable()
    table.set_deco(Texttable.HEADER)
    tablerows = []
    tablerows.append(["Method", "Static", "P", "Machine", "Node", "Socket"])
    multicolumn_header = [("", 1), ("", 1), ("", 1), ("Average score", 3)]

    sorted_best_accuracies = sorted(best_accuracies, key=lambda x: x[1])

    seen_keys = set()

    for (method, P, R, max_machine, max_node, max_socket) in sorted_best_accuracies:
        key = f"{method}_{P}"
        if key in seen_keys:
            continue
        seen_keys.add(key)

        static = False
        if method in STATIC_METHODS:
            static = True

        if method == "k_medoids":
            method = "kmedoids"

        avg_machine, avg_node, avg_socket = average_accuracies(grouped_accuracies[key])

        row = [
            method,
            "True" if static else "False",
            P,
            avg_machine,
            avg_node,
            avg_socket,
        ]
        tablerows.append(row)

    print(
        latextable.draw_latex(
            tablerows,
            caption="A comparison of clustering methods for various measurement counts R",
            label=f"table:logp_clustering",
            use_booktabs=True,
            multicolumn_header=multicolumn_header,
        )
    )
