#!/usr/bin/env python3

import os
import sys
import subprocess

from typing import Optional, List, Tuple, Dict
from dataclasses import dataclass, field

from create_topo_file import create_topo


@dataclass
class ClusteringMethod:
    name: str
    tag: str = "default"
    vector: bool = False
    distmatrix: bool = False
    n_clusters: int = 1
    parameters: Dict[str, any] = field(default_factory=lambda: dict())


@dataclass
class MachineNode:
    name: str
    cores: int
    sockets: int
    slurm: bool

    def topology(self, nodes: int):
        # group_size = self.cores//self.sockets
        topo = create_topo(self.name, self.cores, nodes, self.cores)
        print("TOPO", topo)
        return topo

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        raise NotImplementedError


class HPCLabMachineNode(MachineNode):
    def __init__(self):
        self.name = "hpclab"
        self.cores = 4
        self.sockets = 1
        self.slurm = False

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        env = os.environ.copy()
        if logp_ini_path is not None:
            env["LOGP_CONFIG"] = logp_ini_path
        cmd = [
            "mpirun",
            "-np",
            str(self.cores),
            "./lgate",
            "-f",
            topo,
            "-o",
            run_spec.output_dir(base_dir),
        ]
        print("Running", cmd)
        return subprocess.run(cmd, stdout=stdout, env=env)


class TaroMachineNode(MachineNode):
    def __init__(self):
        self.name = "taro"
        self.cores = 4
        self.sockets = 1
        self.slurm = False

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        env = os.environ.copy()
        if logp_ini_path is not None:
            env["LOGP_CONFIG"] = logp_ini_path
        cmd = [
            "mpirun",
            "-np",
            str(self.cores),
            "./lgate",
            "-f",
            topo,
            "-o",
            run_spec.output_dir(base_dir),
        ]
        print("Running", cmd)
        return subprocess.run(cmd, stdout=stdout, env=env)


class ARM1MachineNode(MachineNode):
    def __init__(self):
        self.name = "arm1"
        self.cores = 96
        self.sockets = 2
        self.slurm = False

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        env = os.environ.copy()
        if logp_ini_path is not None:
            env["LOGP_CONFIG"] = logp_ini_path
        cmd = [
            "mpirun",
            "-np",
            str(self.cores),
            "./lgate",
            "-f",
            topo,
            "-o",
            run_spec.output_dir(base_dir),
        ]
        print("Running", cmd)
        return subprocess.run(cmd, stdout=stdout, env=env)


class FRAMMachineNode(MachineNode):
    def __init__(self):
        self.name = "fram"
        self.cores = 32
        self.sockets = 2
        self.slurm = True

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        env = os.environ.copy()
        if logp_ini_path is not None:
            env["LOGP_CONFIG"] = logp_ini_path

        cmd = [
            "python3",
            "run_job_fram.py",
            "--no-watch",
        ]
        if run_spec.args:
            cmd += run_spec.args

        name = f"{self.name}_{run_spec.nodes}"
        if run_spec.tag:
            name += run_spec.tag

        # if run_spec.params and 'numa_order' in run_spec.params:
        #   env["I_MPI_BIND_ORDER"] = run_spec.params["numa_order"]
        #   val = run_spec.params["numa_order"]
        #   name += f"_numa_{val}"

        cmd += [
            "--name",
            name,
            "--time",
            str(run_spec.timeout),
            "--run",
            "--nodes",
            str(run_spec.nodes),
            "--tasks",
            str(run_spec.machine.cores),
            "--topo-file",
            topo,
            "::",
            "-o",
            run_spec.output_dir(base_dir),
        ]

        return subprocess.run(cmd, env=env)


class FRAMSingleCoreMachineNode(FRAMMachineNode):
    def __init__(self):
        self.name = "fram_singlecore"
        self.cores = 1
        self.sockets = 1
        self.slurm = True


class BETZYMachineNode(MachineNode):
    def __init__(self):
        self.name = "betzy"
        self.cores = 128
        self.sockets = 2
        self.slurm = True

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        env = os.environ.copy()
        if logp_ini_path is not None:
            env["LOGP_CONFIG"] = logp_ini_path

        cmd = [
            "python3",
            "run_job_fram.py",
            "--no-watch",
        ]
        if run_spec.args:
            cmd += run_spec.args

        name = f"{self.name}_{run_spec.nodes}"
        if run_spec.tag:
            name += run_spec.tag
        cmd += [
            "--name",
            name,
            "--time",
            str(run_spec.timeout),
            "--run",
            "--nodes",
            str(run_spec.nodes),
            "--tasks",
            str(run_spec.machine.cores),
            "--topo-file",
            topo,
            "::",
            "-o",
            run_spec.output_dir(base_dir),
        ]

        return subprocess.run(cmd, env=env)

class BETZY16CoreMachineNode(BETZYMachineNode):
    def __init__(self):
        self.name = "betzy16"
        self.cores = 16
        self.sockets = 1
        self.slurm = True

class IDUNMachineNode(MachineNode):
    def __init__(self):
        self.name = "idun"
        self.cores = 32
        self.sockets = 2
        self.slurm = True

    def run(self, run_spec, base_dir: str, topo=None, stdout=None, logp_ini_path=None):
        env = os.environ.copy()
        if logp_ini_path is not None:
            env["LOGP_CONFIG"] = logp_ini_path

        cmd = [
            "python3",
            "run_job.py",
            "--no-watch",
        ]
        if run_spec.args:
            cmd += run_spec.args

        name = f"{self.name}_{run_spec.nodes}"
        if run_spec.tag:
            name += run_spec.tag
        cmd += [
            "--name",
            name,
            "--time",
            str(run_spec.timeout),
            "--run",
            "--nodes",
            str(run_spec.nodes),
            "--tasks",
            str(run_spec.machine.cores),
            "--topo-file",
            topo,
            "::",
            "-o",
            run_spec.output_dir(base_dir),
        ]

        return subprocess.run(cmd, env=env)


@dataclass
class Run:
    machine: MachineNode
    timeout: str
    nodes: int
    reps: int
    max_message_size: int
    expected_clusters: int = 1
    additional_measurements: bool = True

    tag: Optional[str] = None
    args: Optional[List[str]] = None
    params: Optional[Dict[str, any]] = None

    def identifier(self):
        id_ = f"{self.machine.name}_{self.nodes}_{self.machine.cores}_{self.max_message_size}_{self.reps}"
        if self.tag:
            id_ += "_" + self.tag
        return id_

    def short_identifier(self):
        id_ = f"{self.machine.name}_{self.nodes}"
        if self.tag:
            id_ += "_" + self.tag
        return id_

    def output_dir(self, base_dir):
        id_ = f"{self.max_message_size}_{self.reps}"
        if self.tag:
            id_ += "_" + self.tag
        return f"{base_dir}/{self.machine.name}/{self.nodes}_{self.machine.cores}/{id_}"

    def output_base_dir(self, base_dir):
        id_ = f"{self.nodes}_{self.machine.cores}"
        return f"{base_dir}/{self.machine.name}"

    def create_base_dir(self, base_dir):
        os.makedirs(self.output_base_dir(base_dir))


@dataclass
class RunResult:
    identifier: str
    output_dir: str
    exists: bool
    success: bool
    stdout_path: Optional[str]
    logp_ini_path: Optional[str]


TARO = TaroMachineNode()
HPCLAB = HPCLabMachineNode()
ARM1 = ARM1MachineNode()
FRAM = FRAMMachineNode()
FRAM_SINGLECORE = FRAMSingleCoreMachineNode()
BETZY = BETZYMachineNode()
IDUN = IDUNMachineNode()
BETZY16 = BETZY16CoreMachineNode()

MACHINES = {
    "hpclab": HPCLAB,
    "arm1": ARM1,
    "fram": FRAM,
    "fram_singlecore": FRAM_SINGLECORE,
    "betzy": BETZY,
    "betzy16": BETZY16,
    "idun": IDUN,
}


def group_by_nodes(results: List[Tuple[Run, RunResult]]):
    by_nodes = {}
    for (run, result) in results:
        if not run.nodes in by_nodes:
            by_nodes[run.nodes] = []
        by_nodes[run.nodes].append((run, result))
    return by_nodes


# used to extract unique nodes from hostnames
def extract_unique_machines(nodes):
    hosts = nodes.keys()
    # split by '-' and drop last part
    machines = set()
    mapping = {}
    for host, ranks in nodes.items():
        machine = "-".join(host.split("-")[0:-1])
        machines.add(machine)
        if not machine in mapping:
            mapping[machine] = set()
        mapping[machine].update(ranks)
    return machines, mapping
