#!/usr/bin/env python3

from random import randint
import random

from matplotlib import pyplot as plt

import seaborn as sns

sns.set_theme()

import numpy as np
from common import read_meta_file
from create_logp_distance_figures import (
    predict_rtt,
    predict_rtt_many,
    predict_latency,
    predict_gap,
    predict_bandwidth,
    predict_overhead,
)
from logp_model_tool import logp_estimate_parameters

from run_many_logp_common import *
from clustering import predict_clusters_distmatrix

SSSTREE_ALPHA = 0.35


def remove_outliers(A):
    cap = np.percentile(A, 95)
    np.putmask(A, A > np.percentile(A, 95), cap)

def get_prediction_function(metric: str):
    predfunc = None
    if metric == "rtt":
        predfunc = predict_rtt
    elif metric == "latency":
        predfunc = predict_latency
    elif metric == "gap":
        predfunc = predict_gap
    elif metric == "bandwidth":
        predfunc = predict_bandwidth
    elif metric == "overhead":
        predfunc = predict_overhead
    else:
        assert (
            metric == "rtt"
            or metric == "latency"
            or metric == "gap"
            or metric == "bandwidth"
            or metric == "overhead"
        )
    return predfunc

def plot_result(result: RunResult, msgsize: int, out_dir: str, metric: str):

    directory = result.output_dir
    metadata, ranks, nodes = read_meta_file(directory)
    predfunc = get_prediction_function(metric)
    A = predfunc(metadata, ranks, msgsize)

    remove_outliers(A)

    ## Generate a mask for the upper triangle
    # mask = np.triu(np.ones_like(A, dtype=bool))

    # Generate a custom diverging colormap
    # cmap = sns.diverging_palette(230, 20, as_cmap=True)
    #
    cmap = sns.color_palette("viridis", as_cmap=True)
    plt.figure()
    sns.heatmap(
        A,
        # mask=mask,
        center=0,
        square=True,
        cbar=True,
        xticklabels=False,
        yticklabels=False,
        cmap=cmap,
    )
    plt.savefig(f"{out_dir}/{result.identifier}_{msgsize}_{metric}")
    plt.close()
    # plt.show()


def plot_result_line(result: RunResult, out_dir: str):
    from matplotlib import pyplot as plt

    directory = result.output_dir
    metadata, ranks, nodes = read_meta_file(directory)

    from_, to, _, mat = metadata[0]
    msg_sizes = mat[:, 0]
    rtt = np.zeros(len(msg_sizes))
    for i, msg_size in enumerate(msg_sizes):
        rtt[i] = mat[i, 2]  # RTT

    # plt.plot(msg_sizes, rtt, label="Measured RTT")
    xs = np.linspace(msg_sizes.min(), msg_sizes.max(), num=1000)
    # rtt_xs, rtt_ys, predictions = predict_rtt_many(mat, ys)
    # rtt_xs, rtt_ys_hi, rtt_ys_lo = predict_rtt_many(mat, ys)
    # # plt.plot(ys, predictions, label="Predicted RTT")
    # plt.plot(rtt_xs, rtt_ys_lo, label="Predicted RTT low")
    # plt.plot(rtt_xs, rtt_ys_hi, label="Predicted RTT high")
    # plt.grid(False)

    xs, (Ls, oss, ors, gs), (xs2, Ls2, oss2, ors2, gs2) = logp_estimate_parameters(
        mat, xs
    )
    plt.plot(xs, Ls, label="L")
    plt.plot(xs, oss, label="os")
    plt.plot(xs, ors, label="or")
    plt.plot(xs, gs, label="g")

    # plt.plot(xs2, Ls2, label="L real")
    # plt.plot(xs2, oss2, label="os real")
    # plt.plot(xs2, ors2, label="or real")
    # plt.plot(xs2, gs2, label="g real")

    # rtt_xs, rtt_ys_hi, rtt_ys_lo = predict_rtt_many(mat, ys)
    # plt.savefig(f"{out_dir}/{result.identifier}")
    plt.legend()
    plt.show()


def plot_machine_results(
    machine: MachineNode,
    results: List[Tuple[Run, RunResult]],
    msgsize: int,
    out_dir: str,
    metric="rtt",
):
    predfunc = get_prediction_function(metric)

    # ASSUMES all results are for same spec except for reps
    groups = group_by_nodes(results)

    import pandas as pd

    # N = 10
    # var_mat = np.zeros((len(groups), N))
    # j = 0
    # data = {}
    # for nodecount, results in groups.items():
    #    for spec, res in results:
    #        directory = res.output_dir
    #        metadata, ranks, nodes = read_meta_file(directory)
    #        A = predfunc(metadata, ranks, msgsize)
    #        variance = A.reshape(ranks * ranks).var()
    #        # i = int(np.log2(spec.reps)) - 1
    #        if not nodecount in data:
    #            data[nodecount] = {}
    #        data[nodecount][spec.reps] = variance
    #    j += 1
    # df = pd.DataFrame(data)
    # print(df)
    # print("SHOWING HEAT MAP")
    # plt.figure()
    # s = sns.heatmap(
    #    df,
    #    cmap="crest",
    #    xticklabels=True,
    #    yticklabels=True,
    # )
    # s.set(xlabel="Nodes", ylabel="R")
    # plt.legend()
    # plt.show()

    # # create heatmap plot of variance
    for nodecount, results in groups.items():
        print("NODECOUNT", nodecount)
        variance_data = {}
        VAR_X = []
        VAR_Y = []
        # X: repetitions
        # Y: variance
        for spec, res in results:
            directory = res.output_dir
            metadata, ranks, nodes = read_meta_file(directory)
            A = predfunc(metadata, ranks, msgsize)
            variance = A.reshape(ranks * ranks).var()
            VAR_X.append(spec.reps)
            VAR_Y.append(variance)
        f = plt.figure()
        s = sns.lineplot(x=VAR_X, y=VAR_Y)
        s.set(xlabel="R", ylabel="Var")
        s.set(title=f"{nodecount} nodes @ {machine.name}")
        f.savefig(
            f"{out_dir}/{machine.name}_{nodecount}_{metric}_{msgsize}_variance.png"
        )
        # plt.show()

    # # create plot of variance
    # for nodecount, results in groups.items():
    #     variance_data = {}
    #     VAR_X = []
    #     VAR_Y = []
    #     # X: repetitions
    #     # Y: variance
    #     for spec, res in results:
    #         directory = res.output_dir
    #         metadata, ranks, nodes = read_meta_file(directory)
    #         A = predfunc(metadata, ranks, msgsize)
    #         variance = A.reshape(ranks * ranks).var()
    #         VAR_X.append(spec.reps)
    #         VAR_Y.append(variance)
    #     sns.lineplot(x=VAR_X, y=VAR_Y)
    #     plt.show()

    for nodecount, results in groups.items():
        print("Nodes", nodecount)
        n = len(results)

        rows = n // 2
        if rows == 0:
            rows = 1

        fig, axs = plt.subplots(
            2,
            rows,
            sharey=True,
            sharex=True,
            constrained_layout=True,
            dpi=600,
            squeeze=False,
        )
        cbar_ax = fig.add_axes([0.91, 0.3, 0.03, 0.4])
        # fig.tight_layout()
        for j in range(2):
            for i in range(n // 2):
                spec, result = results[j * (n // 2) + i]
                directory = result.output_dir
                metadata, ranks, nodes = read_meta_file(directory)
                A = predfunc(metadata, ranks, msgsize)

                cap99 = np.percentile(A, 99)
                print('cap99', cap99)
                outlier_count = (A > cap99).sum()
                remove_outliers(A)
                print(directory, 'has', outlier_count/(A.shape[0]*A.shape[1]), 'outliers')

                # mask = np.triu(np.ones_like(A, dtype=bool))
                cmap = sns.color_palette("viridis", as_cmap=True)
                sns.heatmap(
                    A,
                    # mask=mask,
                    center=0,
                    square=True,
                    ax=axs[j][i],
                    cbar=i == 0,
                    xticklabels=False,
                    yticklabels=False,
                    cmap=cmap,
                    cbar_ax=None if i else cbar_ax,
                )
                axs[j][i].set_title(f"R={spec.reps}")

                # im = axs[j][i].imshow(A)
                # im.axes.get_xaxis().set_visible(False)
                # im.axes.get_yaxis().set_visible(False)
                axs[j][i].grid(False)
        fig.suptitle(
            f"{machine.name.upper()} PLogP {metric} distance matrix with m={msgsize} P={spec.nodes * machine.cores}"
        )
        fig.savefig(f"{out_dir}/{machine.name}_{nodecount}_{metric}_{msgsize}.png")


def random_cluster_color():
    r = randint(0, 255)
    g = randint(0, 255)
    b = randint(0, 255)
    rand_color = (r, g, b)
    return rand_color


def extract_cluster_apriori(
    machine: MachineNode, run_spec: Run, metadata, ranks, nodes
):
    unique_machines = extract_unique_machines(nodes)
    predicted_cluster_counts = [
        machine.sockets * run_spec.nodes,
        run_spec.nodes,
        len(unique_machines),
    ]
    return predicted_cluster_counts


def get_colors_for_groups(groups):
    colors = {}
    random.seed(10)
    for k in groups.keys():
        colors[min(groups[k])] = random_cluster_color()
    return colors


def plot_machine_results_with_clusters_ssstree(
    machine: MachineNode,
    results: List[Tuple[Run, RunResult]],
    msgsize: int,
    out_dir: str,
):
    # ASSUMES all results are for same spec except for reps
    from matplotlib import pyplot as plt

    alphas = [0.005, 0.025, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45]
    clustering_method = "ssstree"

    groups = group_by_nodes(results)
    for nodecount, results in groups.items():
        print("Doing", nodecount)

        # print("Nodes", nodecount)
        n = len(results)
        fig, axs = plt.subplots(len(alphas), n)
        for i in range(n):
            spec, result = results[i]
            directory = result.output_dir
            metadata, ranks, nodes = read_meta_file(directory)

            axs[0][i].set_title(f"R={spec.reps}")

            mat = predict_rtt(metadata, ranks, msgsize)

            for j, alpha in enumerate(alphas):
                axs[j][0].text(-50, 10, f"alpha={alpha}", color="k")

            for j, alpha in enumerate(alphas):

                # cluser
                A = mat.copy()
                groups, pred = predict_clusters_distmatrix(
                    clustering_method,
                    A,
                    n_clusters=len(groups),
                    parameters={"alpha": alpha},
                )
                colors = get_colors_for_groups(groups)

                data = np.zeros((ranks, ranks, 3), dtype=np.uint8)

                for ii in range(ranks):
                    for jj in range(ii + 1, ranks):
                        if pred[ii] == pred[jj]:
                            data[ii, jj] = colors[pred[ii]]
                            data[jj, ii] = colors[pred[ii]]

                im = axs[j][i].imshow(data)
                im.axes.get_xaxis().set_visible(False)
                im.axes.get_yaxis().set_visible(False)
                axs[j][i].grid(False)
        fig.suptitle(
            f"{machine.name.upper()} PLogP distance matrix with clusters m={msgsize} P={spec.nodes * machine.cores}"
        )
        # plt.show()
        fig.savefig(
            f"{out_dir}/{machine.name}_{nodecount}_clusters_{clustering_method}.png"
        )


def plot_machine_results_with_clusters(
    machine: MachineNode,
    results: List[Tuple[Run, RunResult]],
    msgsize: int,
    out_dir: str,
    clustering_method="ssstree",
):
    # ASSUMES all results are for same spec except for reps
    from matplotlib import pyplot as plt

    groups = group_by_nodes(results)
    for nodecount, results in groups.items():
        print("Doing", nodecount)

        # print("Nodes", nodecount)
        n = len(results)
        fig, axs = plt.subplots(3, n)
        for i in range(n):
            spec, result = results[i]
            directory = result.output_dir
            metadata, ranks, nodes = read_meta_file(directory)

            known_cluster_counts = extract_cluster_apriori(
                machine, spec, metadata, ranks, nodes
            )
            # print("known cluster counts", known_cluster_counts)

            mat = predict_rtt(metadata, ranks, msgsize)

            axs[0][0].text(-50, 10, "Socket", color="k")
            axs[1][0].text(-50, 10, "Node", color="k")
            axs[2][0].text(-50, 10, "Machine", color="k")

            for j in range(3):

                # cluser
                A = mat.copy()
                groups, pred = predict_clusters_distmatrix(
                    clustering_method, A, n_clusters=known_cluster_counts[j]
                )
                colors = get_colors_for_groups(groups)

                data = np.zeros((ranks, ranks, 3), dtype=np.uint8)

                for ii in range(ranks):
                    for jj in range(ii + 1, ranks):
                        if pred[ii] == pred[jj]:
                            data[ii, jj] = colors[pred[ii]]
                            data[jj, ii] = colors[pred[ii]]

                axs[j][i].set_title(f"R={spec.reps}")
                im = axs[j][i].imshow(data)
                im.axes.get_xaxis().set_visible(False)
                im.axes.get_yaxis().set_visible(False)
                axs[j][i].grid(False)
        fig.suptitle(
            f"{machine.name.upper()} PLogP distance matrix with clusters m={msgsize} P={spec.nodes * machine.cores}"
        )
        # plt.show()
        fig.savefig(
            f"{out_dir}/{machine.name}_{nodecount}_clusters_{clustering_method}.png"
        )
