from random import randint
import random

from typing import Dict, Set

import numpy as np
from common import read_meta_file
from create_logp_distance_figures import predict_rtt

from run_many_logp_common import *
from clustering import (
    predict_clusters_distmatrix,
    DISTMATRIX_CLUSTER_METHODS,
    hdbscan_clustering_levels,
)

from collections import Counter
from dataclasses import dataclass


from clustering import run_cluster_search

import seaborn as sns

sns.set_theme()


@dataclass
class ClusteringAccuracy:
    machine: float
    node: float
    socket: float


def calculate_accuracy(
    prediction: Dict[int, Set[int]], correct_clusters: Dict[int, int], n_ranks: int
):
    # print("P", prediction)
    # print("C", correct_clusters)

    true_positives = 0
    true_negatives = 0

    for rank in range(0, n_ranks):
        if prediction[rank] == correct_clusters[rank]:
            true_positives += 1
        else:
            true_negatives += 1

    return true_positives / (true_positives + true_negatives)


def _create_correct_mapping(mapping):
    correct_mapping = {}
    for _, group in mapping.items():
        id_ = min(group)
        for elem in group:
            correct_mapping[elem] = id_
    return correct_mapping


def clustering_accuracy(
    machine: MachineNode,
    run: Run,
    cluster_prediction: Dict[int, Set[int]],
    rank_cluster_prediction: Dict[int, Set[int]],
    nodes_mapping: any,
):
    # print(prediction)
    # print(nodes)
    print(run.nodes, machine.cores, machine.sockets)

    n_ranks = run.nodes * machine.cores

    # Machine
    machines, machine_mapping = extract_unique_machines(nodes_mapping)
    correct_machine_mapping = _create_correct_mapping(machine_mapping)
    print("MACHINES", machines)
    machine_accuracy = calculate_accuracy(
        rank_cluster_prediction, correct_machine_mapping, n_ranks
    )

    # Node accuracy
    correct_node_clusters = _create_correct_mapping(nodes_mapping)
    node_accuracy = calculate_accuracy(
        rank_cluster_prediction, correct_node_clusters, n_ranks
    )

    # Socket accuracy
    socket_correct = 0
    socket_wrong = 0
    # check that every node is divided into machine.sockets clusters
    # because we don't want to make this sensitive to assignment algorithm (round-robin, block etc.)

    cores_per_socket = machine.cores // machine.sockets
    for node, ranks in nodes_mapping.items():
        # groups = set()
        groups = Counter()
        for rank in ranks:
            groups.update([rank_cluster_prediction[rank]])
            # print(rank, rank_cluster_prediction[rank])
        maching_counts = [count == cores_per_socket for count in groups.values()]
        if len(groups) == machine.sockets and all(maching_counts):
            socket_correct += 1
        else:
            socket_wrong += 1
        # print("NODE", node, ranks, "GROUPS", groups)
        print("groups", groups, cores_per_socket)

    socket_accuracy = socket_correct / (socket_correct + socket_wrong)

    return machines, ClusteringAccuracy(
        machine_accuracy, node_accuracy, socket_accuracy
    )


def _find_method_accuracy(
    machine: MachineNode,
    run: Run,
    run_result: RunResult,
    ranks: any,
    nodes: any,
    A: any,
    msgsize: int,
    method: ClusteringMethod,
):
    assert method.distmatrix
    prediction, ranks_prediction = predict_clusters_distmatrix(
        method.name, A, method.n_clusters, parameters=method.parameters
    )
    machines, accuracy = clustering_accuracy(
        machine, run, prediction, ranks_prediction, nodes
    )
    return machines, prediction, accuracy


@dataclass
class ClusteringResult:
    method: str
    n_ranks: int
    n_nodes: int
    n_machines: int
    n_clusters: int
    accuracy: ClusteringAccuracy

    machine: MachineNode
    run: Run


def create_clustering_results(
    machine: MachineNode,
    run: Run,
    result: RunResult,
    clustering_methods: List[ClusteringMethod],
    metadata,
    ranks,
    nodes,
    A,  # distance matrix
    msgsize: int,
):
    out = []
    for method in clustering_methods:
        print("METHOD", method)
        machines, prediction, accuracy = _find_method_accuracy(
            machine, run, result, ranks, nodes, A, msgsize, method
        )
        n_clusters = len(prediction)
        res = ClusteringResult(
            method=method,
            n_ranks=ranks,
            n_nodes=len(nodes),
            n_clusters=n_clusters,
            n_machines=len(machines),
            machine=machine,
            run=run,
            accuracy=accuracy,
        )
        out.append(res)
    return out


def create_clustering_results_table(
    machine: MachineNode,
    results: List[Tuple[Run, RunResult]],
    msgsize: int,
    clustering_methods: List[ClusteringMethod],
):
    out = []

    groups = group_by_nodes(results)
    print(groups)
    for nodecount, results in groups.items():
        for run, res in results:
            directory = res.output_dir
            metadata, ranks, nodes = read_meta_file(directory)
            A = predict_rtt(metadata, ranks, msgsize)

            from matplotlib import pyplot as plt

            plt.imshow(A)
            plt.show()

            for method in clustering_methods:
                print("METHOD", method)
                machines, prediction, accuracy = _find_method_accuracy(
                    machine, run, res, ranks, nodes, A, msgsize, method
                )
                n_clusters = len(prediction)
                print("N clusters", n_clusters)
                result = ClusteringResult(
                    method=method,
                    n_ranks=ranks,
                    n_nodes=len(nodes),
                    n_clusters=n_clusters,
                    n_machines=len(machines),
                    machine=machine,
                    run=run,
                    accuracy=accuracy,
                )
                out.append(result)
    return out


def create_clustering_results_table_leveled(
    machine: MachineNode,
    results: List[Tuple[Run, RunResult]],
    msgsize: int,
):
    out = []

    groups = group_by_nodes(results)
    print(groups)
    for nodecount, results in groups.items():
        for run, res in results:
            directory = res.output_dir
            metadata, ranks, nodes = read_meta_file(directory)
            A = predict_rtt(metadata, ranks, msgsize)

            levels = hdbscan_clustering_levels(A)
            print("LEVELS", levels)

    return out


@dataclass
class AccuracyResult:
    r: int
    p: int
    accuracy: ClusteringAccuracy


def _plot_timings(machine_name, timing_results, out_dir: str):
    print(machine_name, timing_results)
    pass


def _plot_param_search(machine_name, accuracy_results, out_dir: str):
    from matplotlib import pyplot as plt
    import pandas as pd

    datas = []

    print("ACCURACY_RESULTS", accuracy_results)

    rs = set()
    ps = set()
    methods = set()
    # we want to plot a grouped bar plot
    # x-axis is the P
    mapping = {}  # mapping[p][r] is a dictionary of accuracies
    for ar in accuracy_results:
        rs.add(ar.r)
        ps.add(ar.p)
        if ar.p not in mapping:
            mapping[ar.p] = {}
        if ar.r not in mapping[ar.p]:
            mapping[ar.p][ar.r] = {}

        for name, acc in ar.accuracy.items():
            mapping[ar.p][ar.r][name] = acc
            methods.add(name)

    print("MAPPING", mapping)
    methods = sorted(methods)
    rs = sorted(rs)
    ps = sorted(ps)

    for p in ps:
        datas = {}
        datas["Method"] = []
        datas["Machine accuracy"] = []
        datas["Node accuracy"] = []
        datas["Socket accuracy"] = []
        for r in rs:
            for m in methods:
                datas[f"Method"].append(f"{m} R={r}")
                datas[f"Machine accuracy"].append(mapping[p][r][m].machine)
                datas[f"Node accuracy"].append(mapping[p][r][m].node)
                datas[f"Socket accuracy"].append(mapping[p][r][m].machine)

        df = pd.DataFrame(datas)
        print(df)
        plt.close()
        ax = df.plot(
            x="Method",
            kind="barh",
            y=["Machine accuracy", "Node accuracy", "Socket accuracy"],
            stacked=True,
            figsize=(6, 8)
            # title=f"Clustering accuracy for {machine_name.upper()} P={p}",
        )
        sns.move_legend(
            ax,
            "lower center",
            bbox_to_anchor=(0, 1),
            ncol=3,
            # title="Accuracy type",
            frameon=False,
        )

        # plt.show()

        plt.tight_layout(h_pad=0.4, w_pad=0.5)
        # plt.show()
        out_file = f"{out_dir}/clustering_accuracy_{machine_name}_P{p}.png"
        plt.savefig(out_file, dpi=300)
        plt.close()

    # df.plot(x="P", kind="bar", stacked=False, title="Accuracies")
    # plt.show()
    # datas = []

    # fig, ax = plt.subplots(layout="constrained")

    # for attribute, measurements in method_ps.items():
    #     print("attribute", attribute)
    #     for s in ps:
    #         if not s in measurements:
    #             continue
    #         measurement = measurements[s].socket
    #         print("measurement", measurement)

    #         offset = width * multiplier
    #         rects = ax.bar(x + offset, measurement, width, label=attribute)
    #         ax.bar_label(rects, padding=3)
    #         multiplier += 1

    # # Add some text for labels, title and custom x-axis tick labels, etc.
    # ax.set_ylabel("Length (mm)")
    # ax.set_title("Accuracy by R")
    # ax.set_xticks(x + width, method_ps.keys())
    # ax.legend(loc="upper left", ncols=3)
    # ax.set_ylim(0, 250)

    # plt.show()


def table_param_search(
    machine: MachineNode, results: List[RunResult], msgsize, out_dir: str
):
    accuracy_results = []
    timings = []
    for (run, result) in results:
        directory = result.output_dir
        metadata, ranks, nodes = read_meta_file(directory)
        A = predict_rtt(metadata, ranks, msgsize)

        accuracy_result = {}
        cluster_results, cluster_timings = run_cluster_search(
            A, run.nodes, run.machine.cores
        )
        for name, cluster_res in cluster_results.items():
            if not name in accuracy_result:
                accuracy_result[name] = ClusteringAccuracy(0.0, 0.0, 0.0)

            for (prediction, ranks_prediction) in cluster_res:
                machines, accuracy = clustering_accuracy(
                    machine, run, prediction, ranks_prediction, nodes
                )
                # print("NAME", name, accuracy)
                accuracy_result[name].socket = max(
                    accuracy_result[name].socket, accuracy.socket
                )
                accuracy_result[name].node = max(
                    accuracy_result[name].node, accuracy.node
                )
                accuracy_result[name].machine = max(
                    accuracy_result[name].machine, accuracy.machine
                )

        accuracy_results.append(
            AccuracyResult(
                r=run.reps, p=run.nodes * machine.cores, accuracy=accuracy_result
            )
        )
        timings.append(cluster_timings)

    sns.set_palette(sns.color_palette("Set2"))
    _plot_param_search(machine.name, accuracy_results, out_dir)
    _plot_timings(machine.name, timings, out_dir)

    return accuracy_results
