#!/usr/bin/env python3
from typing import *


import os
import argparse

from dataclasses import dataclass
from matplotlib import pyplot as plt
import graphviz

# import matplotlib
# from sklearn import preprocessing
# import igraph as ig

import numpy as np

from common import *

ALPHA = 0.35
MAX_DIST = 999999999
MAX_NUM_CHILDREN = 16

COLOR1 = "#D4CC47"
COLOR2 = "#7C4D8B"


def euclid_distance(point1, point2):
    return np.linalg.norm(point1 - point2)


# https://medium.com/@BrendanArtley/matplotlib-color-gradients-21374910584b
def hex_to_RGB(hex_str):
    """#FFFFFF -> [255,255,255]"""
    # Pass 16 to the integer function for change of base
    return [int(hex_str[i : i + 2], 16) for i in range(1, 6, 2)]


def get_color_gradient(c1, c2, n):
    """
    Given two hex colors, returns a color gradient
    with n colors.
    """
    if n == 1:
        return [c1]
    assert n > 1
    c1_rgb = np.array(hex_to_RGB(c1)) / 255
    c2_rgb = np.array(hex_to_RGB(c2)) / 255
    mix_pcts = [x / (n - 1) for x in range(n)]
    rgb_colors = [((1 - mix) * c1_rgb + (mix * c2_rgb)) for mix in mix_pcts]
    return [
        "#" + "".join([format(int(round(val * 255)), "02x") for val in item])
        for item in rgb_colors
    ]


@dataclass
class Node:
    id: int
    children: List["Node"]
    radius: float
    weights: any = None
    index: Optional[int] = None
    parent: Optional["Node"] = None

    # for clusters
    center: Optional[any] = None

    def weight_or_center(self):
        if self.is_cluster():
            return self.children[0].weights
        return self.weights

    def is_cluster(self):
        return not self.is_leaf()

    def is_leaf(self):
        return (
            len(self.children) == 0 and self.weights is not None and self.radius == 0.0
        )

    def calculate_new_radius(self, new_child: "Node"):
        dist = distance(self.center, new_child.weights)
        if dist > self.radius:
            return dist
        return self.radius


def print_node(node: Node, depth=0):
    prefix = " " * 2 * depth
    # print(
    #     f"{prefix}Node(id={node.id}, index={node.index}, radius={node.radius}, weights={node.weights}, is_leaf={node.is_leaf()})"
    # )
    print(
        f"{prefix}Node(id={node.id}, index={node.index}, radius={node.radius}, weights=..., is_leaf={node.is_leaf()})"
    )
    for ch in node.children:
        print_node(ch, depth=depth + 1)


class SSSTree:
    ndims: int
    dataset: List[any]

    max_num_children: int
    id_counter: int

    sss_tree: Node
    height: int

    def __init__(
        self, dataset: List[any], ndims: int, distance, alpha=ALPHA, max_depth=32
    ):
        self.dataset = dataset
        self.ndims = ndims
        self.id_counter = 0
        self.distance = distance
        self.alpha = alpha
        self.max_depth = max_depth
        self.max_num_children = MAX_NUM_CHILDREN

        base_tree = self.create_base_tree(dataset)
        # print("BASE_TREE")
        # print_node(base_tree)

        self.height = 0
        sss_tree = self.create_sss_tree(base_tree, None)

        # print("SSSTree")
        # print_node(sss_tree)

        self.sss_tree = sss_tree

    def root_radius(self, root: Node, children: List[Node]):
        radius = 0.0
        for ch in children[1:]:
            dist = self.distance(children[0].weights, ch.weights)
            radius = max(radius, dist)
        return radius

    def find_global_closest_cluster(
        self, clusters: List[Node], node: Node, pred=None
    ) -> [Node, int, float]:
        assert len(clusters) > 0

        found_index = -1
        found_cluster = clusters[0]
        found_dist = np.inf

        for i, cluster in enumerate(clusters):
            # print("cluster", len(cluster.children))
            for j, cnode in enumerate(cluster.children):
                #print("cnode", cnode)
                dist = self.distance(cnode.weight_or_center(), node.weight_or_center())

                if pred is not None and not pred(cluster):
                    continue

                if dist < found_dist:
                    found_dist = dist
                    found_cluster = cluster
                    found_index = i
        return found_cluster, found_index, found_dist

    def find_closest_cluster(
        self, clusters: List[Node], node: Node, pred=None
    ) -> [Node, int, float]:
        assert len(clusters) > 0

        found_index = 0
        found_cluster = clusters[0]
        found_dist = self.distance(
            clusters[0].weight_or_center(), node.weight_or_center()
        )

        for i, cluster in enumerate(clusters[1:]):
            dist = self.distance(cluster.center, node.weights)

            if pred is not None and not pred(cluster):
                continue

            if dist < found_dist:
                found_dist = dist
                found_cluster = cluster
                found_index = i + 1
        return found_cluster, found_index, found_dist

    def make_cluster(self, node: Node, parent: Optional[Node]) -> Node:
        node_id = self.id_counter
        self.id_counter += 1
        cluster = Node(
            id=node_id, children=[node], radius=0.0, parent=parent, center=node.weights
        )
        node.parent = cluster
        return cluster

    def add_to_cluster(self, cluster: Node, node: Node, parent: Optional[Node]) -> Node:
        if cluster.is_cluster():
            cluster.children.append(node)
            return cluster

        # print('splitting cluster', cluster)
        new_cluster = self.make_cluster(cluster, parent)
        new_cluster.children.append(node)
        node.parent = new_cluster
        return new_cluster

    def create_base_tree(self, weights: List[any]):
        assert len(weights) >= 1

        root_id = self.id_counter
        self.id_counter += 1

        root_node = Node(
            id=root_id,
            index=0,
            children=[],
            radius=0.0,
            weights=weights[0],
        )

        root = self.make_cluster(root_node, None)

        for i, weight in enumerate(weights[1:]):
            node_id = self.id_counter
            self.id_counter += 1
            self.add_to_cluster(
                root,
                Node(
                    id=node_id,
                    index=i + 1,
                    children=[],
                    radius=0.0,
                    weights=weight,
                ),
                root,
            )
        return root

    def create_sss_tree(self, node: Node, parent: Node | None, depth=0):
        self.height = max(self.height, depth)

        if len(node.children) == 0 or depth >= self.max_depth:
            return node
        elif len(node.children) < self.max_num_children:
            node.radius = self.root_radius(node, node.children)
            return node
        else:
            # add new cluster
            if parent is not None:
                node.parent = parent

            node.radius = self.root_radius(node, node.children)
            # print("radius of", node.id, node.radius)

            root = node.children[0]
            clusters = [root]

            M = 2 * node.radius

            for ch in node.children[1:]:
                # print("root", root, "ch", ch)

                add_new = True
                add_to = None
                add_to_ind = -1

                # iterate over clusters
                for i, ch2 in enumerate(clusters):
                    dist = self.distance(node.weight_or_center(), ch.weight_or_center())
                    if dist < M * self.alpha:
                        add_new = False
                        add_to = ch2
                        add_to_ind = i
                        break

                cluster, index, dist = self.find_closest_cluster(clusters, ch)

                if dist >= M * self.alpha:
                    cluster = self.make_cluster(ch, node)
                    # print("creating new cluster:", ch.index)
                    # print_node(cluster)

                    # create new cluster
                    clusters.append(cluster)

                else:
                    # print("add to cluster:")
                    cluster = self.add_to_cluster(cluster, ch, node)
                    clusters[index] = cluster

            node.children = []
            node.weights = None

            for ch in clusters:
                # if depth < 2:
                #     node.children.append(
                #         self.create_sss_tree(ch, node, depth=depth + 1)
                #     )
                # else:
                node.children.append(ch)
            return node

    def clusters(self, node: Optional[Node] = None, min_cluster_size=2, move=False):
        clusters = {}
        cluster_centers = []

        def leaves_rec(node: Node, parent: Optional[Node]):
            nonlocal clusters

            if node.is_leaf():
                if not parent.id in clusters:
                    clusters[parent.id] = []
                    cluster_centers.append(parent)
                clusters[parent.id].append(node)
            else:
                for ch in node.children:
                    leaves_rec(ch, node)

        if node is None:
            node = self.sss_tree
        assert node
        leaves_rec(node, None)

        # if not move:
        #     return clusters

        # Move outliners to closest cluster
        new_clusters = {}
        nodes = []
        removed = set()
        for key, cluster in clusters.items():
            if len(cluster) < min_cluster_size:
                for node in cluster:
                    nodes.append(node)
                removed.add(key)
            else:
                new_clusters[key] = cluster

        # move nodes in this cluster to the closest cluster
        for node in nodes:
            closest_center, _, _ = self.find_closest_cluster(
                cluster_centers, node, pred=lambda cluster: cluster.id not in removed
            )
            if closest_center.id not in new_clusters:
                new_clusters[closest_center.id] = []
            # print("moved", node.index, "to", closest_center.id)
            new_clusters[closest_center.id].append(node)

        return new_clusters

    def leaves(self, node: Optional[Node] = None):
        def leaves_rec(node: Node):
            if node.is_leaf():
                yield node
            else:
                for ch in node.children:
                    yield from leaves_rec(ch)

        if node is None:
            node = self.sss_tree
        assert node
        yield from leaves_rec(node)

    def leaf_clusters(self):
        seen = set()

        def leaf_clusters_rec(node: Node, parent: Optional[Node]):
            if node.is_leaf():
                if parent:
                    if not parent.id in seen:
                        seen.add(parent.id)
                        yield parent
            else:
                for ch in node.children:
                    yield from leaf_clusters_rec(ch, node)

        assert self.sss_tree
        yield from leaf_clusters_rec(self.sss_tree, None)

    def search(self, query: any, distance_func=None, max_distance=0.0001):
        # print("searching query", query)

        if distance_func is None:
            distance_func = distance

        cluster_comparisons = 0
        node_comparisons = 0

        best = None
        best_parent = None
        best_dist = MAX_DIST

        def search_rec(node: Node, parent: Optional[Node]):
            nonlocal cluster_comparisons, node_comparisons
            nonlocal best, best_parent, best_dist

            if node.is_leaf():
                node_comparisons += 1
                dist = distance_func(node.weights, query)
                if dist <= max_distance:
                    if dist < best_dist:
                        best_dist = dist
                        best = node
                        best_parent = parent
                    return True, node, parent
                return False, None, parent

            cluster_comparisons += 1
            dist = distance_func(node.center, query)
            M = 2 * node.radius
            if dist > M * self.alpha:
                # skip
                # print("skipping", node.id)
                return False, None, parent

            for ch in node.children:
                found, value, par = search_rec(ch, node)
                if found:
                    return True, value, par

            return False, None, parent

        search_rec(self.sss_tree, None)
        # print('search used', cluster_comparisons, node_comparisons)
        return best is not None, best, best_parent, best_dist

    def plot(self, graph=None, node=None, is_root=True):
        if not graph:
            graph = graphviz.Digraph(
                comment="SSSTree", filename="plot.gv", engine="sfdp"
            )
        if not node:
            node = self.sss_tree

        if node.is_leaf():
            graph.node(f"{node.index}")
        else:
            for child in node.children:
                for child_b in node.children:
                    if child_b.is_leaf() and child_b.id != child.id:
                        graph.edge(f"{child.index}", f"{child_b.index}")

                with graph.subgraph(name=f"Cluster {node.id}") as subgraph:
                    self.plot(graph=subgraph, node=child, is_root=False)

        if is_root:
            graph.view()

    def plot_sss_tree(self):
        fig, ax = plt.subplots(1, figsize=(8, 8))

        clusters = self.clusters()
        # print(clusters)

        for i, cluster in enumerate(clusters.values()):
            if len(cluster) == 0:
                continue

            xpoints = [node.index for node in cluster]
            ypoints = [cluster[0].index for node in cluster]

            plt.scatter(xpoints, ypoints, label=f"Cluster {i} n={len(cluster)}")
        plt.legend()
        plt.show()


def read_meta_file(directory):
    dir_prefix = os.path.dirname(directory)
    if dir_prefix != "":
        dir_prefix += "/"

    filepath = f"{directory}/metadata"
    assert os.path.isfile(filepath)
    with open(filepath, "r") as f:
        lines = f.readlines()

    out = []
    max_rank = 0
    for line in lines:
        split = line.split()
        from_ = int(split[0])
        to = int(split[2])
        file = split[-1]

        max_rank = max(max_rank, from_)
        max_rank = max(max_rank, to)

        mat = read_mat(dir_prefix + file)
        out.append((from_, to, file, mat))
    return out, max_rank + 1


def make_diagonal_random(A):
    # removed = A[~np.eye(A.shape[0], dtype=bool)].reshape(A.shape[0], int(A.shape[0])-1, -1)

    # mean = A.mean()
    mean = A[A < float("inf")].mean()
    for i in range(A.shape[0]):
        A[i][i] = mean


def make_symmetric(mat):
    n = mat.shape[0]
    for i in range(n):
        for j in range(n):
            mat[i][j] = mat[j][i]
            # mat[j][i] = 0


def read_mat(filename):
    with open(filename, "r") as f:
        lines = f.readlines()
        rows = []
        for line in lines:
            row = list(map(float, line.split()))
            rows.append(row)
        return np.array(rows)


def calc_eigen_n(mat, values, vectors, d: int):
    n = mat.shape[0]
    B = np.zeros_like(mat)
    for i in range(n):
        for j in range(n):
            if j < d:
                B[i][j] = vectors[i][j] * values[j]

    C = np.zeros_like(B)

    for i in range(n):
        for j in range(n):
            for k in range(n):
                C[i][j] += B[i][k] * vectors[j][k]

    return C


def load_data2(directory: str, apply_eigen=False, eigen_n=1):
    metadata, ranks = read_meta_file(directory)

    first_mat = metadata[0][3]

    n_msgsizes = first_mat.shape[0]
    n_fields = first_mat.shape[1] - 1

    # step 1: create 'ranks x ranks' matrices for all dimensions we care about
    # TODO: don't care about first column
    matrices = np.zeros((ranks, ranks, n_msgsizes, n_fields))

    for i, (from_, to, _, mat) in enumerate(metadata):
        for row in range(n_msgsizes):
            for col in range(n_fields):
                # skip first column
                matrices[from_][to][row][col] = mat[row][col + 1]

    # OPTIONAL: apply eigen decomposition
    if apply_eigen:
        for msgsize in range(n_msgsizes):
            for field in range(n_fields):
                # mat = matrices[from_][to][msgsize][
                mat = matrices[:, :, msgsize, field]

                # make symmetric
                mat = np.maximum(mat, mat.transpose())

                # fix up diagonal
                # make_diagonal_random(mat)

                # run eigenvalue decomposition
                values, vectors = linalg.eigh(mat)
                eigen_res = calc_eigen_n(mat, values, vectors, eigen_n)

                # print('Plotting eigen for msgsize', msgsize, 'field', field)

                # plt.subplot(1, 2, 1)
                # plt.imshow(mat)
                # plt.title("Normal")
                # plt.subplot(1, 2, 2)
                # plt.title(f"Eigen n={eigen_n}")
                # plt.imshow(eigen_res)
                # plt.colorbar()
                # plt.show()

                # assign back
                matrices[:, :, msgsize, field] = eigen_res

    # step 2: flatten into weights
    dataset = np.zeros((ranks, ranks * n_msgsizes * n_fields))
    for i in range(ranks):
        weights = matrices[i].flatten()
        dataset[i] = weights

    return dataset, ranks


def load_data(directory: str, ndims=2, eigen=False, eigen_n=1):
    metadata, ranks = read_meta_file(directory)

    if eigen:
        return run_eigen_on_data(metadata, ranks, ndims=ndims, eigen_n=eigen_n), ranks

    # list of [from,to,...data...]
    dataset = np.zeros((len(metadata), 2 + ndims))
    for i, (from_, to, _, mat) in enumerate(metadata):
        dataset[i][0] = from_
        dataset[i][1] = to
        for dim in range(2, ndims + 1):
            dataset[i][dim] = mat[-1][dim]
            # dataset[i][dim] = mat[0][dim]

    return dataset, ranks


def rank_distance(weights1, weights2):
    return np.abs(weights1[0] - weights2[0])


def extract_ranks(weights):
    return int(weights[0]), int(weights[1])


def find_group(ssstree, query, distance_func, min_group_size, max_distance=3.0):
    max_dist = 0.0001
    while max_dist < max_distance:
        # print('find_group', max_dist)
        found, node, parent, dist = ssstree.search(
            query, distance_func, max_distance=max_dist
        )
        unique_ranks = set()
        leaves = ssstree.leaves(parent)
        for leaf in leaves:
            from_, to_ = extract_ranks(leaf.weights)
            unique_ranks.add(from_)

        # print('group', unique_ranks)

        group_size = len(unique_ranks)
        # print('Group_size', group_size)
        if group_size < min_group_size:
            max_dist *= 2
            continue
        else:
            # print('Found group of size', group_size)
            return found, node, parent, dist
    return False, None, None, MAX_DIST

    return clusters


def create_topo(clusters, out_file):
    # print('Topology:')
    with open(out_file, "w") as f:
        print("cluster {", file=f)
        for cluster in clusters.values():
            if len(cluster) == 0:
                continue
            print("\tcluster {", file=f)
            print("\t\tgroup ", end="", file=f)
            print(set(map(lambda node: node.index, cluster)), file=f)
            print("\t}", file=f)
        print("}", file=f)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="SSSTree clustering")
    parser.add_argument("directory")
    parser.add_argument("--print-tree", default=False, action="store_true")
    parser.add_argument("--plot", default=False, action="store_true")
    parser.add_argument("--apply-eigen", default=False, action="store_true")
    parser.add_argument("--eigen-n", default=1, type=int)
    parser.add_argument("--topo-out", default="/dev/stdout", type=str)
    parser.add_argument("--show-clusters", default=False, action="store_true")
    parser.add_argument("--topo", default=False, action="store_true")
    parser.add_argument("--min-cluster-size", default=2, type=int)
    parser.add_argument("--verbose", default=False, action="store_true")
    args = parser.parse_args()

    dataset, n_ranks = load_data2(
        args.directory, apply_eigen=args.apply_eigen, eigen_n=args.eigen_n
    )

    if args.verbose:
        print(
            "Loaded dataset",
            f"dimensions={dataset.shape}",
            f"n_ranks={n_ranks}",
            f"alpha={ALPHA}",
            f"max_num_children={MAX_NUM_CHILDREN}",
        )

    ssstree = SSSTree(dataset, euclid_distance, ndims=n_ranks)

    if args.print_tree:
        print_node(ssstree.sss_tree)

    if args.show_clusters:
        clusters = ssstree.clusters(min_cluster_size=args.min_cluster_size)
        print(f"Clusters (min_cluster_size={args.min_cluster_size}):")
        for cluster in clusters.values():
            cluster_set = set(map(lambda node: node.index, cluster))
            print(cluster_set, "size:", len(cluster_set))

    if args.topo:
        clusters = ssstree.clusters(min_cluster_size=args.min_cluster_size)
        create_topo(clusters, args.topo_out)

    if args.plot:
        ssstree.plot()
        # ssstree.plot_sss_tree()
