#include "topo.h"

#include <assert.h>
#include <ctype.h>
#include <string.h>

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>

#define MAX(a, b) ((a) < (b) ? (b) : (a))

enum TokenType {
  None,
  BracketOpen,
  BracketClose,
  Dash,
  Comma,
  Number,
  String,
  EndOfFile
};

const char *TokenStrs[] = {"None",  "BracketOpen", "BracketClose", "Dash",
                           "Comma", "Number",      "String",       "EndOfFile"};

typedef struct {
  int start_index;
  int length;

  union {
    char *str_value;
    int int_value;
  };
  bool static_;
  enum TokenType type;
} Token;

Token CommaToken = {.type = Comma, .static_ = true, .length = 1};
Token BracketOpenToken = {.type = BracketOpen, .static_ = true, .length = 1};
Token BracketCloseToken = {.type = BracketClose, .static_ = true, .length = 1};
Token EndOfFileToken = {.type = EndOfFile, .static_ = true, .length = 0};
Token ClusterToken = {
    .type = String, .static_ = true, .str_value = "cluster", .length = 7};
Token GroupToken = {
    .type = String, .static_ = true, .str_value = "group", .length = 5};
Token DashToken = {.type = Dash, .static_ = true, .length = 1};

typedef struct {
  char *src;
  int size;
  int offset;

  int id_counter;

  Token *tokens;
  int token_len;
  int tokeni;
} State;

#define Token_OK(token) (token.type != None && token.type != EndOfFile)

static Token next_token(State *s) {
  if (s->tokeni >= s->token_len)
    return (Token){.type = EndOfFile, .static_ = true};
  return s->tokens[s->tokeni++];
}

static Token peek_token(State *s) {
  if (s->tokeni >= s->token_len)
    return (Token){.type = EndOfFile};
  return s->tokens[s->tokeni];
}

static Token lex_next_token(State *s) {
  if (s->offset >= s->size)
    return (Token){.type = EndOfFile};

  while (s->offset < s->size) {
    char c = s->src[s->offset];
    if (c == ',') {
      s->offset++;
      return CommaToken;
    } else if (c == '{') {
      s->offset++;
      return BracketOpenToken;
    } else if (c == '}') {
      s->offset++;
      return BracketCloseToken;

    } else if (c == '-') {
      s->offset++;
      return DashToken;

    } else if (isdigit(c)) {
      int number = c - '0';
      int i = 1;
      while (s->offset + i < s->size && isdigit(s->src[s->offset + i])) {
        number *= 10;
        number += s->src[s->offset + i] - '0';
        i++;
      }
      s->offset += i;
      return (Token){
          .type = Number, .static_ = true, .int_value = number, .length = i};

    } else if (isalpha(c)) {
      int i = 0;
      while (s->offset + i < s->size && isalpha(s->src[s->offset + i])) {
        i++;
      }
      char *str = calloc(i + 1, sizeof(char));
      assert(str != NULL);
      memcpy(str, &s->src[s->offset], i);

      s->offset += i;
      return (Token){
          .type = String, .static_ = false, .str_value = str, .length = i};

    } else if (isspace(c)) {
      s->offset++;
    } else {
      break;
    }
  }

  return (Token){.type = None};
}

static bool is_token(Token a, Token b) {
  if (a.type != b.type) {
    return false;
  }

  if (a.type == String && strcmp(a.str_value, b.str_value) != 0) {
    return false;
  }

  if (a.type == Number && a.int_value != b.int_value)
    return false;

  return true;
}

static bool next_is_token(State *s, Token token) {
  Token next = peek_token(s);
  return is_token(token, next);
}

static int expect_token(State *s, Token token) {
  Token current_token = next_token(s);
  if (token.type != current_token.type) {
    fprintf(stderr, "expected %s got %s\n", TokenStrs[token.type],
            TokenStrs[current_token.type]);
    exit(1);
  }
  if (token.type == String &&
      strcmp(current_token.str_value, token.str_value) != 0) {
    fprintf(stderr, "expected String %s got %s\n", token.str_value,
            current_token.str_value);
    exit(1);
  }
  if (token.type == Number && current_token.int_value != token.int_value) {
    fprintf(stderr, "expected Number %d got %d\n", token.int_value,
            current_token.int_value);
    exit(1);
  }
  return 0;
}

void init_cluster(Cluster *cluster) {
  cluster->id = -1;
  cluster->leaf = false;
  cluster->parent = NULL;
  cluster->leader_rank = -1;
  cluster->group = NULL;
  cluster->sub_clusters = NULL;
  cluster->n_sub_clusters = 0;
}

static int parse_cluster_single(State *s, Cluster *parent,
                                Cluster *cluster_out);

static int parse_cluster_list(State *s, Cluster *parent,
                              Cluster **cluster_list_out,
                              size_t *cluster_count_out) {
  size_t clusters_cap = 100;
  Cluster *cluster_list = malloc(sizeof(Cluster) * clusters_cap);
  assert(cluster_list != NULL);

  int cluster_count = 0;

  while (true) {
    if (next_is_token(s, ClusterToken)) {
      if (cluster_count + 1 >= clusters_cap) {
        cluster_list = realloc(cluster_list, sizeof(Cluster) * clusters_cap);
        assert(cluster_list != NULL);
      }

      Cluster *cluster = &cluster_list[cluster_count];
      init_cluster(cluster);

      int ret = parse_cluster_single(s, parent, cluster);
      if (ret != 0)
        return ret;

      cluster_count++;
    } else {
      break;
    }
  }
  *cluster_list_out = cluster_list;
  *cluster_count_out = cluster_count;
  return 0;
}

static int parse_number_list(State *s, int **members_out,
                             int *members_len_out) {
  int members_cap = 1;
  int members_len = 0;
  int *members = calloc(members_cap, sizeof(int));
  assert(members != NULL);

  expect_token(s, BracketOpenToken);

  if (next_is_token(s, BracketCloseToken)) {
    // empty list
    return 0;
  }

  do {
    Token next = next_token(s);
    assert(Token_OK(next));

    if (is_token(next, BracketCloseToken)) {
      break;
    }

    if (next.type != Number) {
      fprintf(stderr, "expected %s got %s\n", TokenStrs[Number],
              TokenStrs[next.type]);
      assert(next.type == Number);
      return 1;
    }

    if (next_is_token(s, DashToken)) {
      Token range_start = next;
      expect_token(s, DashToken);
      Token range_end = next_token(s);
      if (range_end.type != Number) {
        fprintf(stderr, "expected number to end range, got %s\n",
                TokenStrs[range_end.type]);
        assert(range_end.type == Number);
      }

      int start = range_start.int_value;
      int end = range_end.int_value;

      if (members_len + (end - start) + 1 >= members_cap) {
        members_cap *= 2;
        members_cap = MAX(members_cap, members_len + (end - start) + 1);
        members = realloc(members, members_cap * sizeof(int));
        assert(members != NULL);
      }

      for (int i = start; i <= end; i++) {
        members[members_len] = i;
        members_len++;
      }
    } else {
      if (members_len + 1 >= members_cap) {
        members_cap *= 2;
        members = realloc(members, members_cap * sizeof(int));
        assert(members != NULL);
      }
      members[members_len] = next.int_value;
      members_len++;
    }

    if (next_is_token(s, CommaToken))
      expect_token(s, CommaToken);
  } while (true);

  *members_out = members;
  *members_len_out = members_len;

  return 0;
}

static int parse_cluster_single(State *s, Cluster *parent,
                                Cluster *cluster_out) {
  init_cluster(cluster_out);

  cluster_out->parent = parent;

  expect_token(s, ClusterToken);
  expect_token(s, BracketOpenToken);

  if (next_is_token(s, GroupToken)) {
    expect_token(s, GroupToken);

    int *members = NULL;
    int members_len = 0;
    if (parse_number_list(s, &members, &members_len) != 0) {
      return 1;
    }

    assert(members_len > 0);

    cluster_out->leaf = true;
    cluster_out->leader_rank = members[0];
    NodeGroup *group = malloc(sizeof(NodeGroup));
    assert(group != NULL);
    group->ranks = members;
    group->n_ranks = members_len;
    cluster_out->id = s->id_counter++;

    cluster_out->group = group;

    expect_token(s, BracketCloseToken);
    return 0;
  } else {

    size_t clusters_size = 0;
    Cluster *cluster_list = NULL;

    int ret = parse_cluster_list(s, cluster_out, &cluster_list, &clusters_size);

    assert(clusters_size > 0);

    cluster_out->n_sub_clusters = clusters_size;
    cluster_out->sub_clusters = cluster_list;
    cluster_out->leaf = false;
    cluster_out->id = s->id_counter++;
    cluster_out->leader_rank = cluster_list[0].leader_rank;

    expect_token(s, BracketCloseToken);
    return ret;
  }
}

void print_cluster(FILE *file, Cluster *cluster, int depth) {
  fprintf(file, "%*scluster #%d leader=%d  (sub-clusters=%d) parent=%d{\n",
          depth, "", cluster->id, cluster->leader_rank, cluster->n_sub_clusters,
          cluster->parent ? cluster->parent->id : -1);
  if (cluster->leaf) {
    fprintf(file, "%*s  group {", depth, "");
    for (int i = 0; i < cluster->group->n_ranks; i++) {
      fprintf(file, "%d", cluster->group->ranks[i]);
      if (i + 1 < cluster->group->n_ranks)
        fprintf(file, ",");
    }
    fprintf(file, "}\n");
  } else {
    for (int i = 0; i < cluster->n_sub_clusters; i++) {
      print_cluster(file, &cluster->sub_clusters[i], depth + 1);
    }
  }
  fprintf(file, "%*s}\n", depth, "");
}

void print_topology(FILE *file, Topology *topology) {
  fprintf(file, "Topology:\n");
  print_cluster(file, &topology->cluster, 0);
}

int tokenize(State *s) {
  int token_cap = 100;
  int token_len = 0;
  Token *tokens = malloc(token_cap * sizeof(Token));
  assert(tokens != NULL);

  do {
    Token next = lex_next_token(s);
    if (!Token_OK(next))
      break;

    if (token_len + 1 >= token_cap) {
      token_cap *= 2;
      tokens = realloc(tokens, token_cap * sizeof(Token));
      assert(tokens != NULL);
    }

    tokens[token_len] = next;
    token_len++;
  } while (true);

  s->tokens = tokens;
  s->token_len = token_len;

  return 0;
}

int parse_topology(char *src, Topology *out) {
  assert(src != NULL);
  State s = {
      .src = src,
      .size = strlen(src),
      .offset = 0,
      .id_counter = 0,
  };

  tokenize(&s);

  // for (int i = 0; i < s.token_len; i++) {
  //   if (s.tokens[i].type == Number) {
  //     fprintf(stderr, "Token: Number %d\n", s.tokens[i].int_value);
  //   } else if (s.tokens[i].type == String) {
  //     fprintf(stderr, "Token: String %s\n", s.tokens[i].str_value);
  //   } else {
  //     fprintf(stderr, "Token: %s\n", TokenStrs[s.tokens[i].type]);
  //   }
  // }

  parse_cluster_single(&s, NULL, &out->cluster);

  // Free tokens
  for (int i = 0; i < s.token_len; i++) {
    if (s.tokens[i].type == String && !s.tokens[i].static_) {
      free(s.tokens[i].str_value);
    }
  }
  free(s.tokens);
  return 0;
}

static void cluster_leaves_helper(Cluster *cluster, NodeGroup **leaves,
                                  int *leaves_cap, int *leaves_count) {
  if (!cluster->leaf) {
    for (int i = 0; i < cluster->n_sub_clusters; i++)
      cluster_leaves_helper(&cluster->sub_clusters[i], leaves, leaves_cap,
                            leaves_count);
  } else {
    if (*leaves_count + 1 >= *leaves_cap) {
      *leaves_cap = *leaves_cap * 2;
      *leaves = realloc(*leaves, sizeof(NodeGroup) * *leaves_cap);
      assert(*leaves != NULL);
    }

    NodeGroup new_group = *cluster->group;
    new_group.n_ranks = cluster->group->n_ranks;
    new_group.ranks = malloc(sizeof(int) * cluster->group->n_ranks);
    memcpy(new_group.ranks, cluster->group->ranks,
           sizeof(int) * cluster->group->n_ranks);
    assert(new_group.ranks != NULL);

    (*leaves)[*leaves_count] = new_group;
    (*leaves_count)++;
  }
}

void cluster_leaves(Cluster *cluster, NodeGroup **leaves_out,
                    int *leaves_count_out) {
  int leaves_count = 0;
  int leaves_cap = 1;
  NodeGroup *leaves = malloc(sizeof(NodeGroup) * leaves_cap);
  assert(leaves != NULL);
  cluster_leaves_helper(cluster, &leaves, &leaves_cap, &leaves_count);

  *leaves_out = leaves;
  *leaves_count_out = leaves_count;
}

void cluster_leaves_finalize(NodeGroup *leaves, int leaves_count) {
  for (int i = 0; i < leaves_count; i++) {
    free(leaves[i].ranks);
  }
  free(leaves);
}

void cluster_finalize(Cluster *cluster) {
  if (cluster->leaf) {
    free(cluster->group->ranks);
    free(cluster->group);
    cluster->group = NULL;
  } else {
    for (int i = 0; i < cluster->n_sub_clusters; i++) {
      cluster_finalize(&cluster->sub_clusters[i]);
    }
    free(cluster->sub_clusters);
  }
}

void topology_finalize(Topology *topology) {
  cluster_finalize(&topology->cluster);
}
